Package: semFromKeys
Type: Package
Title: Run 'lavaan' Models from Keys Lists
Version: 0.5.2
Authors@R: c(
    person(
      "Timothy F.", "Bainbridge",
      email = "tfbainbridge@gmail.com",
      role = c("aut", "cre", "cph")
    )
  )
Description: Specifying 'lavaan' models manually can be time consuming when
    multiple similar models are required. The 'semFromKeys' package streamlines
    the process of running 'lavaan' models by generating model code from simple
    keys lists and running entire collections of models at once.
    The package was inspired by the process used in the code for
    Bainbridge, T. F., Ludeke, S. G., & Smillie, L. D. (2022)
    <doi:10.1037/pspp0000395>.
    The package also optionally checks that identical models have not been run
    on the same data, which saves time when code needs to be run again.
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
Config/roxygen2/version: 8.1.0
Depends: R (>= 3.6.0)
Imports: stringr, lavaan, openssl, withr, tools, matrixcalc, Matrix,
        stats, stringr
Suggests: testthat (>= 3.0.0), here, rstudioapi
Config/testthat/edition: 3
URL: https://github.com/timbainbridge/semFromKeys
BugReports: https://github.com/timbainbridge/semFromKeys/issues
NeedsCompilation: no
Packaged: 2026-08-22 06:13:46 UTC; tim
Author: Timothy F. Bainbridge [aut, cre, cph]
Maintainer: Timothy F. Bainbridge <tfbainbridge@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-22 06:40:02 UTC
Built: R 4.5.2; ; 2026-08-22 08:58:19 UTC; unix

# Generated by roxygen2: do not edit by hand

export(bifactor.from.keys)
export(cache.clean)
export(cache.setup)
export(cfa.from.keys)
export(efa.from.keys)
export(esem.from.keys)
export(esem.from.mods)
export(sem.check)
export(sem.cor)
export(sem.path)
importFrom(Matrix,nearPD)
importFrom(lavaan,
  fitMeasures,
  parameterEstimates,
  sam,
  sem,
  standardizedSolution,
  summary
)
importFrom(matrixcalc,is.positive.definite)
importFrom(openssl,md5)
importFrom(stringr,
  str_replace_all,
  str_split
)
importFrom(withr,with_options)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# semFromKeys

<!-- badges: start -->

[![R-CMD-check](https://github.com/timbainbridge/semFromKeys/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/timbainbridge/semFromKeys/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The ‘semFromKeys’ package was designed to streamline running ‘lavaan’
models with similar structures using keys lists to generate model code
instead of writing out the code for models manually. For confirmatory
factor analyses (CFAs) and bifactor models, the code creates and runs a
series of models based on keys indicating each of the factors in the
models. For exploratory factor analyses (EFAs) keys list are used to
create a target rotation for a single EFA. For latent variable
correlations, the model takes fitted CFA models and runs a series of
models computing correlations between latent variables and, optionally,
single items. For exploratory structural equation models (ESEM), the
code takes a fitted EFA model and fitted CFA and/or bifactor models and
runs an ESEM for each CFA or bifactor model input. In the ESEM, the EFA
factors predict a series of latent variables in separate models using
Burt’s (1976) 2-stage procedure to prevent interpretational confounding.
The ESEM models were designed to run analyses equivalent to that of
Bainbridge, Ludeke, and Smillie (2022).

Although the package might be of most use to those running ESEM similar
to those of Bainbridge and colleagues (2022), it could also be very
helpful to anyone wanting to create a correlation matrix based on latent
variables rather than sum scores or to estimate a CFA measurement model
for each scale in a sample to either check measurement characteristics
before proceeding with further analyses or to simply compute measurement
model based reliability statistics.

For sets of models that take a long time to run, code has been included
to allow the first run to save outputs that can be checked against in
subsequent runs. If nothing has changed, then the previous outputs are
returned, saving the time (and energy) of running them again. To get
this feature to work, the R version has to be 4.0 or later and a cache
directory will have to be set with the `cache.setup()` function, which,
by default, configures a cache directory in the users’ cache as
determined by the operating system. It can alternatively be set as a
subdirectory within the current project or, if not using a project, the
current working directory. Once the cache is set, `save_out = TRUE` can
be included in function calls to save the relevant outputs, and
`check = TRUE` can be included to look for previous outputs and only run
models where something has changed. The feature means that small changes
in data cleaning or model code need not necessitate re-running
time-consuming models if only a small number have changed.

Given that the package enables creating files in a cache directory, the
`cache.clean()` function has also been included to help clean up files.
To comply with CRAN policies, the cache directory is set as a temporary
environment variable, so it has to be set each time the global
environment is cleared.

## Installation

You can install the development version of ‘semFromKeys’ from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("timbainbridge/semFromKeys")
```

You can install the stable CRAN version with:

``` r
install.packages("semFromKeys")
```

## Example

The following example generates keys, runs CFAs and an EFA using these
keys, computes correlations between CFA latent variables, and uses
outputs from these to run ESEMs.

### CFAs

In this case, keys can be created from names in the dataset but they can
also be created with simple code to generate a list.

``` r
library(semFromKeys)
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
```

The lists should look something like this:

``` r
keys
#> $grit_c
#> [1] "grit_c_1" "grit_c_2" "grit_c_3" "grit_c_4" "grit_c_5" "grit_c_6"
#> 
#> $grit_p
#> [1] "grit_p_1" "grit_p_2" "grit_p_3" "grit_p_4" "grit_p_5" "grit_p_6"
#> 
#> $hope_a
#> [1] "hope_a_1" "hope_a_2" "hope_a_3" "hope_a_4"
#> 
#> $hope_p
#> [1] "hope_p_1" "hope_p_2" "hope_p_3" "hope_p_4"
```

Once keys are created, the CFAs can be run. The function produces
messages of progress. These can help identify which models produced
errors or warnings or to keep track of progress for collections of
models with long run times.

``` r
cfa_fit <- cfa.from.keys(keys, BFIGritHope, fit_save = TRUE)
#> Fitting models
#> 1 / 4   grit_c
#> 2 / 4   grit_p
#> 3 / 4   hope_a
#> 4 / 4   hope_p
#> Generating parameter estimates
#> 1 / 4   grit_c
#> 2 / 4   grit_p
#> 3 / 4   hope_a
#> 4 / 4   hope_p
#> Generating model fit statistics
#> 1 / 4   grit_c
#> 2 / 4   grit_p
#> 3 / 4   hope_a
#> 4 / 4   hope_p
```

Results can be examined; for example, standard ‘lavaan’ summaries:

``` r
lavaan::summary(cfa_fit$fit$grit_c)
#> lavaan 0.7-2 ended normally after 12 iterations
#> 
#>   Estimator                                         ML
#>   Optimization method                           NLMINB
#>   Number of model parameters                        18
#> 
#>   Number of observations                           388
#>   Number of missing patterns                         1
#> 
#> Model Test User Model:
#>                                                       
#>   Test statistic                                64.001
#>   Degrees of freedom                                 9
#>   P-value (Chi-square)                           0.000
#> 
#> Parameter Estimates:
#> 
#>   Standard errors                             Standard
#>   Information                                 Observed
#>   Observed information based on                Hessian
#> 
#> Latent Variables:
#>                    Estimate  Std.Err  z-value  P(>|z|)
#>   grit_c =~                                           
#>     grit_c_1          0.904    0.053   17.204    0.000
#>     grit_c_2          0.832    0.057   14.560    0.000
#>     grit_c_3          0.685    0.059   11.656    0.000
#>     grit_c_4          0.794    0.056   14.089    0.000
#>     grit_c_5          0.879    0.057   15.317    0.000
#>     grit_c_6          0.762    0.063   12.173    0.000
#> 
#> Intercepts:
#>                    Estimate  Std.Err  z-value  P(>|z|)
#>    .grit_c_1          3.235    0.058   55.482    0.000
#>    .grit_c_2          2.856    0.061   47.163    0.000
#>    .grit_c_3          3.088    0.059   52.185    0.000
#>    .grit_c_4          3.219    0.059   54.439    0.000
#>    .grit_c_5          2.938    0.062   47.644    0.000
#>    .grit_c_6          3.376    0.064   52.736    0.000
#> 
#> Variances:
#>                    Estimate  Std.Err  z-value  P(>|z|)
#>    .grit_c_1          0.501    0.052    9.722    0.000
#>    .grit_c_2          0.731    0.064   11.400    0.000
#>    .grit_c_3          0.889    0.072   12.431    0.000
#>    .grit_c_4          0.726    0.063   11.539    0.000
#>    .grit_c_5          0.704    0.064   11.070    0.000
#>    .grit_c_6          1.010    0.081   12.452    0.000
#>     grit_c            1.000
```

And selected fit measures:

``` r
cfa_fit$fit_measures[, c("cfi", "rmsea")]
#>              cfi      rmsea
#> grit_c 0.9328014 0.12550173
#> grit_p 0.9143581 0.12213711
#> hope_a 0.9796273 0.12148480
#> hope_p 0.9978190 0.03467277
```

These models can be used to examine the measurement characteristics of
the scales or to calculate latent variable model-based reliability
scores (e.g., with
`sapply(cfa_fit$fit, function(x) semTools::compRelSEM(x)[[1]])` for
composite reliability, Jöreskog, 1971).

Bifactor models can be run with a similar, albeit more complex, method.
See `?bifactor.from.keys` for details.

### EFAs

As for CFAs, an EFA can be run from a keys list. In this case, the keys
list indicates factor that items are expected to load on rather than
separate models. These are used to generate a target rotation to help
ensure the EFA matches expectations.

``` r
keys_e0 <- paste0("bfi_", c("e", "a", "c", "n", "o"))
keys_e <- sapply(
  keys_e0,
  function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))],
  simplify = FALSE
)
```

After the keys list has been created, the model can be run similarly to
the CFAs. When running the model, fit measures can be restricted to
speed up estimation if not all are required (as for ‘lavaan’s’
`lavaan::fitMeasures()` function).

``` r
efa_fit <- efa.from.keys(
  keys_e, BFIGritHope, check = FALSE, fit_save = TRUE,
  fit_measures = c("chisq", "df", "pvalue", "bic")
)
#> Fitting models
#> 1 / 1   efa
#> Generating parameter estimates
#> 1 / 1   efa
#> Generating model fit statistics
#> 1 / 1   efa
```

EFA results can be examined in a similar way to the CFAs.

``` r
# Not run due to length
# lavaan::summary(efa_fit$fit$efa)  # Standard lavaan summary
efa_fit$fit_measures                # Fit measures
#>        chisq   df pvalue      bic
#> efa 4808.621 1480      0 62887.71
```

### Correlations

The fitted CFA models can also be used to calculate latent variable
correlations. In this example, Burt’s 2-stage procedure is used by
setting `nagy = FALSE` to save time, but in many cases Nagy and
colleagues’ (2017) method will be superior and is the default. See
`?sem.cor()` for further details.

``` r
latent_cors <- sem.cor(BFIGritHope, cfa_fit$fit, nagy = FALSE)
#> Fitting models
#> 1 / 6   grit_c.grit_p
#> 2 / 6   grit_c.hope_a
#> 3 / 6   grit_c.hope_p
#> 4 / 6   grit_p.hope_a
#> 5 / 6   grit_p.hope_p
#> 6 / 6   hope_a.hope_p
#> Generating parameter estimates
#> 1 / 6   grit_c.grit_p
#> 2 / 6   grit_c.hope_a
#> 3 / 6   grit_c.hope_p
#> 4 / 6   grit_p.hope_a
#> 5 / 6   grit_p.hope_p
#> 6 / 6   hope_a.hope_p
latent_cors$cor_mat
#>           grit_c    grit_p    hope_a    hope_p
#> grit_c 1.0000000 0.4962176 0.3724859 0.3002280
#> grit_p 0.4962176 1.0000000 0.8993154 0.8325711
#> hope_a 0.3724859 0.8993154 1.0000000 0.9276052
#> hope_p 0.3002280 0.8325711 0.9276052 1.0000000
```

### ESEM

Finally, outputs from CFA, bifactor, and EFA models can be used as
inputs into ESEMs where the scales of the CFAs and bifactor models are
regressed on the EFA factors. In this example, bifactor models are not
included.

``` r
esem_fit <- esem.from.mods(
  BFIGritHope, efa_fit$fit$efa, cfa_fit$fit, fit_save = FALSE
)
#> Fitting models
#> 1 / 4   grit_c
#> 2 / 4   grit_p
#> 3 / 4   hope_a
#> 4 / 4   hope_p
#> Generating parameter estimates
#> 1 / 4   grit_c
#> 2 / 4   grit_p
#> 3 / 4   hope_a
#> 4 / 4   hope_p
```

The function provides standard ‘lavaan’ outputs, as well as r-squared
values and regression parameters.

``` r
# Not run due to length
# lavaan::summary(esem_fit$fit$grit_c)  # Standard lavaan summary
round(esem_fit$r2, 3)
#>           R2    se ci.lower ci.upper
#> grit_c 0.508 0.041    0.428    0.588
#> grit_p 0.731 0.035    0.663    0.799
#> hope_a 0.782 0.030    0.724    0.840
#> hope_p 0.610 0.040    0.532    0.688
```

``` r
esem_fit$b$grit_c
#>       rhs est.std    se      z pvalue ci.lower ci.upper
#> 388 bfi_e  -0.155 0.046 -3.369  0.001   -0.244   -0.065
#> 389 bfi_a   0.077 0.048  1.607  0.108   -0.017    0.171
#> 390 bfi_c   0.432 0.048  8.995  0.000    0.338    0.526
#> 391 bfi_n  -0.367 0.048 -7.606  0.000   -0.461   -0.272
#> 392 bfi_o   0.100 0.048  2.112  0.035    0.007    0.193
```

To take advantage of functions’ time-saving `check = TRUE` for
subsequent running of code, a cache directory will need to be set. To
see how to do this, see `?cache.setup`.

## References

Bainbridge, T. F., Ludeke, S. G., & Smillie, L. D. (2022). Evaluating
the Big Five as an organizing framework for commonly used psychological
trait scales. Journal of Personality and Social Psychology, 122(4),
749-777. <https://doi.org/10.1037/pspp0000395>.

Burt, R. S. (1976). Interpretational confounding of unobserved variables
in Structural Equation Models. Sociological Methods & Research, 5(1),
3-52. <https://doi.org/10.1177/004912417600500101>.

Jöreskog, K. G. (1971). Statistical Analysis of Sets of Congeneric
Tests. Psychometrika, 36(2), 109-133.
<https://doi.org/10.1007/BF02291393>.

Nagy, G., Brunner, M., Lüdtke, O., and Greiff, S. (2017). Extension
Procedures for Confirmatory Factor Analysis. Journal of Experimental
Education, 85(4), 574-596.
<https://doi.org/10.1080/00220973.2016.1260524>.

<!-- You'll need to render `README.Rmd` regularly, to keep `README.md` up-to-date. `devtools::build_readme()` is handy for this. -->

#' Runs lavaan models after checking for code or data changes.
#'
#' @description
#' sem.check takes a model list, keys lists, and data, and
#' produces outputs from a series of lavaan models.
#' The function is primarily used as a function to be called by other function
#' within the package but it can also be run independently.
#'
#' For each model, the code optionally checks for previously saved model code,
#' a hash of data, and important parameter inputs.
#' If they exist and match values for the current code and data,
#' the model is not run, the previous output is loaded instead.
#' If either the code has changed, the hash of the data has changed, or
#' important parameter inputs have change, or any of these do not exist,
#' then the models are run as normal.
#'
#' @param mods A named list of lavaan models to run.
#' @param data
#' A dataframe or object coercible to a dataframe.
#' Data must include all observed variables used in any of the models.
#' @param keys_s A named keys list matching the names and length of mod.
#' @param keys_e A named keys list of the factors in an ESEM to be included.
#' @param std
#' Logical.
#' `TRUE` indicates that standardised parameter estimates should be saved;
#' `FALSE` indicates that unstandardised parameters estimates should be saved.
#' @param fit_save
#' Logical.
#' `TRUE` indicates model fit measures should be included in the output;
#' `FALSE` indicates model fit measures should not be included in the output.
#' `FALSE` may be desirable when fit measures are of little interest and when
#' they may take a long time to estimate.
#' Fit can still be examined for individual models with,
#' `lavaan::fitMeasures(fit$fit$[model name])`.
#' @param fit_measures
#' A vector of fit measures to save or 'all' to select all fit measures,
#' as per the `fit.measures` parameter from lavaan's [lavaan::fitMeasures()]
#' function.
#' Defaults to 'all'. Irrelevant if `fit_save = FALSE`.
#' @param target
#' A matrix indicating a rotation target,
#' as used in the `rotation.args` argument in lavaan (see [lavaan::efa()].
#' If `NULL` (default),
#' target is not specified and lavaan uses default behaviour.
#' Irrelevant when the model does not include an EFA or ESEM.
#' @param name
#' A string indicating a subdirectory where model outputs will be saved when
#' `save_out = TRUE` and checked against when `check = TRUE`.
#' Defaults to "sem".
#' The name should be unique for each set of models, or outputs from calls with
#' the same name will be overwritten.
#' @param orthogonal
#' Logical.
#' Sets the `orthogonal` parameter, as per lavaan (see [lavaan::lavOptions()]).
#' `TRUE` indicates that unspecified latent variable correlations should be
#' fixed at 0;
#' `FALSE` indicates that unspecified latent variable correlations should be
#' freely estimated.
#' Defaults to `FALSE`.
#' @param miss
#' A string.
#' Sets the `missing` parameter, as per lavaan (see [lavaan::lavOptions()]).
#' Defaults to 'ML'.
#' @param est
#' A string.
#' Sets the `estimator` parameter, as per lavaan (see [lavaan::lavOptions()]).
#' The default ('default') uses the lavaan default for the model being run.
#' @param std.lv
#' Logical.
#' Sets the `std.lv` parameter, as per lavaan (see [lavaan::lavOptions()]).
#' `TRUE` indicates that factor variances should be fixed to 1.
#' `FALSE` indicates that loadings of the first items of factors should be fixed
#' to 1.
#' Defaults to `FALSE`.
#' @param check
#' Logical.
#' `TRUE` indicates that current inputs should be compared to previous inputs
#' if they exist and that the model should not be rerun if nothing has changed;
#' `FALSE` indicates that these checks should not be made and the model should
#' be run regardless of the existence of previous inputs.
#' @param save_out
#' Logical.
#' `TRUE` indicates that model code, a hash of the data, important input
#' parameter values, and output will be saved;
#' `FALSE` indicates that nothing will be saved.
#' Selecting `save_out = TRUE` enables the function to not rerun models next
#' time if `check = TRUE` the next time the code is run and nothing has changed
#' in the meantime.
#'
#' @return
#' Returns a list of length 2 (if `fit_save = FALSE`) or
#' 3 (if `fit_save = TRUE`).
#' The elements of the list are: a list of lavaan model output objects;
#' a list of parameter estimates from the models (standardized if `std = TRUE`);
#' and, if `fit_save = TRUE`, a matrix of fit measures for each model.
#'
#' @details
#' The function is largely intended to be used as a helper function to upstream
#' functions,
#' including [cfa.from.keys()], [bifactor.from.keys()], [efa.from.keys()], and
#' [esem.from.mods()].
#' Although it is recommended to use the appropriate upstream function whenever
#' possible,
#' there are not (currently) options to do so when customised lavaan models are
#' required;
#' for example, when allowing two items' residuals to correlate in a CFA.
#' `sem.check()` can be used in these cases (see example).
#'
#' Matching the philosophy of the package, the function is designed to run for
#' multiple models with a similar design. If you are using the function for a
#' single model, transform inputs into lists as appropriate or simply use lavaan
#' without the assistance of semFromKeys.
#'
#' The function includes functionality designed to save time re-running code
#' when lots of slow models are included.
#' To do this, when `save_out = TRUE` and a cache directory has been set,
#' the model will save various inputs and outputs from the function call, and,
#' when `check = TRUE`, the model will look for any previously saved outputs
#' from earlier model runs in the same cache directory and
#' only run again if nothing has changed.
#' In cases where something has changed,
#' then the function will re-run the models where something has changed
#' (but it will not for those where nothing has changed).
#' Changes to arguments that influence all lavaan model runs (e.g., miss or est)
#' will trigger all models to be re-run.
#'
#' For either `save_out = TRUE` or `check = TRUE`,
#' the function will look for a cache directory set and created by the
#' [cache.setup()] function.
#' If a cache directory has not been set for the current session,
#' then the function will exit with an error suggesting that either
#' [cache.setup()] be run or `save_out` and `check` set to `FALSE`.
#'
#' When the cache directory is found and output from previous runs are detected,
#' the comparisons performed are for:
#' * model code;
#' * hashes of the data (using [openssl::md5()]);
#' * values of the `miss`, `est`, `std`, `std.lv`, and `orthogonal`parameters;
#' * the class of model objects (i.e., class lavaan); and,
#' * the class of parameter estimates (i.e., class lavaan.data.frame).
#'
#' For most applications, the checking feature can be safely ignored by not
#' saving outputs (i.e., `fit_save = FALSE`)
#' and not checking for past saves (i.e., `check = FALSE`, both default),
#' but may be beneficial in cases with lots of models or very slow models.
#' However, the functionality can be safely used for faster runs too.
#'
#' @seealso
#' [cfa.from.keys()], [efa.from.keys()], [bifactor.from.keys()], and
#' [esem.from.mods()]---all these function depend upon `sem.check()` to work;
#' [lavaan::sem()], which is used to estimate the models;
#' [lavaan::parameterEstimates()], which is used to estimate parameter values;
#' and [lavaan::fitMeasures()], which is used to estimate fit statistics when
#' `fit_save = TRUE`.
#'
#' @importFrom stringr str_replace_all
#' @importFrom lavaan sem
#' @importFrom lavaan standardizedSolution
#' @importFrom lavaan parameterEstimates
#' @importFrom lavaan fitMeasures
#' @importFrom lavaan summary
#' @importFrom openssl md5
#' @importFrom withr with_options
#' @export
#'
#' @examples
#' # Create CFA keys
#' keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#' keys <- sapply(
#'   keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#' )
#' # Create model code
#' mods <- mapply(
#'   x = keys, y = names(keys), SIMPLIFY = FALSE,
#'   FUN = function(x, y) paste(y, "=~", paste(x, collapse = " + "))
#' )
#' # Edit model code to add correlated residuals
#' mods[[1]] <- paste0(mods[[1]], "\ngrit_c_1 ~~ grit_c_2")
#' # Estimate the models with sem.check()
#' cfa_fit <- sem.check(mods, BFIGritHope, keys, name = "cfa", check = FALSE)

sem.check <- function(
    mods, data, keys_s = NULL, keys_e = NULL,
    fit_save = FALSE, fit_measures = "all",
    miss = "ML", est = "default", std.lv = FALSE, std = TRUE,
    orthogonal = FALSE, target = NULL,
    name = "sem", check = FALSE, save_out = FALSE
) {
  if (!is.logical(fit_save)) {
    stop("'fit_save' is not logical. It should be 'TRUE' or 'FALSE'.")
  }
  if (fit_save & !is.character(fit_measures)) {
    stop("'fit_measures' is not a character vector.")
  }
  if (!is.logical(check)) {
    stop("'check' is not logical. It should be 'TRUE' or 'FALSE'.")
  }
  if (!is.logical(save_out)) {
    stop("'save_out' is not logical. It should be 'TRUE' or 'FALSE'.")
  }
  if (save_out | check) {
    found <- FALSE
    for (i in 1:5) {
      env <- parent.frame(i)
      if (exists(".cache_env", envir = env)) {
        found <- TRUE
        break
      }
    }
    if (!found) {
      stop(
        paste(
          "A cache directory is not configured so cannot be cleaned.",
          "Use the 'cache.setup()' function to configure a directory to clean."
        )
      )
    }
    cache_dir <- get("cache_dir", envir = get(".cache_env", envir = env))
    if (!is.character(name)) {
      stop("'name' is not a character string.")
    }
  }
  if (!is.logical(std)) {
    stop("'std' is not logical. It should be 'TRUE' or 'FALSE'.")
  }
  if (!is.logical(orthogonal)) {
    stop("'orthogonal' is not logical. It should be 'TRUE' or 'FALSE'.")
  }
  if (!is.logical(std.lv)) {
    stop("'std.lv' is not logical. It should be 'TRUE' or 'FALSE'.")
  }
  if (!is.list(mods)) {
    stop(
      paste(
        "'mods' is not a list.",
        "If you are trying to run a single model, you will have to make a",
        "length 1 list",
        "(e.g., 'mods = list(mod_name = mod)')"
      )
    )
  }
  data <- as.data.frame(data)
  if (is.null(keys_s) & is.null(keys_e)) {
    stop("At least one of 'keys_s' or 'keys_e' must be specified.")
  }
  if (!is.list(keys_s) & !is.null(keys_s)) {
    stop(
      paste(
        "'keys_s' is not a list.",
        "Please ensure that keys_s is specified correctly.",
        "See the function's help for further information."
      )
    )
  }
  if (!is.null(keys_s)) {
    if (length(mods) != length(keys_s)) {
      stop(
        paste(
          "'keys_s' and 'mods' are not the same length.",
          "If 'keys_s' is supposed to be specified, then it should be the same",
          "length as 'mods'."
        )
      )
    }
    if (sum(table(names(keys_s)) > 1) > 0) {
      stop(
        paste(
          "At least two elements of 'keys_s' share the same name.",
          "Please ensure that all model names are unique."
        )
      )
    }
  }
  if (!is.list(keys_e) & !is.null(keys_e)) {
    stop(
      paste(
        "'keys_e' is not a list.",
        "Please ensure that keys_s is specified correctly.",
        "See the function's help for further information."
      )
    )
  }
  if (!is.null(keys_e)) {
    if (sum(table(names(keys_e)) > 1) > 0) {
      stop(
        paste(
          "At least two elements of 'keys_e' share the same name.",
          "Please ensure that all model names are unique."
        )
      )
    }
    if (sum(sapply(keys_e, function(x) length(x) < 2)) > 0) {
      short_factors <-
        names(keys_e)[sapply(keys_e, function(x) length(x) == 1)]
      warning(
        paste(
          paste0("    ", short_factors, collapse = "    \n"),
          "\n\nThe above factors have only 1 item.",
          "This is not necessarily a problem",
          "(assuming no more serious errors have occured)",
          "but it is unusual for EFAs and may not be what you intended to do."
        )
      )
    }
  }
  if (sum(!(unlist(keys_s)) %in% colnames(data)) > 0) {
    una_items <- unlist(keys_s)[!(unlist(keys_s)) %in% colnames(data)]
    stop(
      paste0(
        "The following items are in 'keys_s' but they are not in 'data':",
        "\n      ",
        paste0(una_items, collapse = "\n      "),
        paste0(
          "\n\nEnsure that the column names of 'data' and keys list item ",
          "names match and that 'data' is a dataframe or coercible to a ",
          "dataframe.\n",
          "If using bifactor.from.keys, ensure that you have not swapped keys ",
          "inadvertently (e.g., keys_g for keys_b)."
        )
      )
    )
  }
  if (sum(!(unlist(keys_e)) %in% colnames(data)) > 0) {
    una_items <- unlist(keys_e)[!(unlist(keys_e)) %in% colnames(data)]
    message(paste0("  ", una_items, collapse = "  \n"))
    stop(
      paste(
        "The above items are in 'keys_e' but they are not in 'data'.",
        "Ensure that data is a data frame (or coercible into a data frame)",
        "and that column names of 'data' and keys list item names match."
      )
    )
  }
  if (save_out) {
    if (!dir.exists(file.path(cache_dir, name))) {
      dir.create(file.path(cache_dir, name))
    }
  }
  # Tell user which model set is running
  if (check) {
    # Load hashes, prior models, and critical parameters
    m0 <-
      if (file.exists(file.path(cache_dir, name, paste0(name, "_mod.rds")))) {
        tmp <- readRDS(file.path(cache_dir, name, paste0(name, "_mod.rds")))
        # Remove spaces
        lapply(tmp, function(x) gsub(" ", "", x))
      } else FALSE
    params0 <-
      if (
        file.exists(file.path(cache_dir, name, paste0(name, "_params.rds")))
      ) {
        readRDS(file.path(cache_dir, name, paste0(name, "_params.rds")))
      } else FALSE
    hash_d0 <-
      if (file.exists(file.path(cache_dir, name, paste0(name, "_hash.rds")))) {
        readRDS(file.path(cache_dir, name, paste0(name, "_hash.rds")))
      } else FALSE
    # Create hashes
    hash_d <- sapply(
      names(mods),
      function(x) {
        md5(paste(data[c(keys_s[[x]], unlist(keys_e))], collapse = ""))
      }
    )
    # Compare to previous hashes
    # if includes possibility of hash_d0 <- c(FALSE, [hash]) for some reason.
    hash_d_test <- if (length(hash_d0) != 1 | hash_d0[[1]] != FALSE) {
      sapply(
        names(mods),
        function(x) {
          if (x %in% names(hash_d0)) hash_d[[x]] == hash_d0[[x]] else FALSE
        }
      )
    } else FALSE
    # Remove spaces for comparison
    mods0 <- lapply(mods, function(x) gsub(" ", "", x))
    m_test <- if (length(m0) != 1 | m0[[1]] != FALSE) {
      sapply(
        names(mods0),
        function(x) {
          if (x %in% names(m0)) {
            # Only check to 4-decimal places.
            str_replace_all(
              mods0[[x]],
              "([0-9]+\\.[0-9]+)",
              ~format(round(as.numeric(.x), 4), nsmall = 4)
            ) ==
              str_replace_all(
                m0[[x]],
                "([0-9]\\.[0-9]+)",
                ~format(round(as.numeric(.x), 4), nsmall = 4)
              )
          } else FALSE
        }
      )
    } else FALSE
    params <- c(
      miss = miss, est = est, std = std, std.lv = std.lv,
      orthogonal = orthogonal
    )
    param_test <- if (length(params0) == length(params)) {
      ifelse(sum(params0 != params) > 0, FALSE, TRUE)
    } else FALSE
    # Load old object if it exists
    fit0 <-
      if (file.exists(file.path(cache_dir, name, paste0(name, "_fit.rds")))) {
        readRDS(file.path(cache_dir, name, paste0(name, "_fit.rds")))
      } else FALSE
    if (std == TRUE) {
      if (
        file.exists(file.path(cache_dir, name, paste0(name, "_par_std.rds")))
      ) {
        par0 <- readRDS(
          file.path(cache_dir, name, paste0(name, "_par_std.rds"))
        )
      } else {
        par0 <- FALSE
      }
    } else {
      if (file.exists(file.path(cache_dir, name, paste0(name, "_par.rds")))) {
        par0 <- readRDS(file.path(cache_dir, name, paste0(name, "_par.rds")))
      } else {
        par0 <- FALSE
      }
    }
    if (file.exists(file.path(cache_dir, name, paste0(name, "_fit_m.rds")))) {
      fit_m0 <- readRDS(file.path(cache_dir, name, paste0(name, "_fit_m.rds")))
    } else {
      fit_m0 <- FALSE
    }
    # If hashes are correct but the fitted object is moved or deleted, then the
    # fit0 object will be FALSE. To ensure this doesn't break everything, check
    # that the fit0 object is a lavaan object. If not, run again.
    fit_type <- sapply(
      names(mods),
      function(x) {
        ifelse(x %in% names(fit0), inherits(fit0[[x]], "lavaan"), FALSE)
      }
    )
    par_type <- sapply(
      names(mods),
      function(x) {
        ifelse(
          x %in% names(par0),
          inherits(par0[[x]], "lavaan.data.frame")[1],
          FALSE
        )
      }
    )
  } else {
    m_test <- FALSE
    hash_d_test <- FALSE
    param_test <- FALSE
    fit_type <- FALSE
    par_type <- FALSE
    fit_m0 <- FALSE
  }
  # Run
  # Warnings required to be printed immediately so people know which model
  # caused the issue -> use with_options.
  message("Fitting models")
  fit <- with_options(
    list(warn = 1),
    mapply(
      function(m1, hash_d1, n_mod, mods1, ft, n) {
        if (hash_d1 & m1 & ft & param_test) {
          fit0[[n_mod]]
        } else {
          if (length(mods) <= 1000) {
            # Print progress for every model
            message(paste(n, "/", length(mods), " ", n_mod))
          } else {
            # Print progress for every 10th model (starting from 1)
            if (n %in% round((0:(length(mods)/10)*10) + 1, 0)) {
              message(paste(n, "/", length(mods), " ", n_mod))
            }
          }
          if (is.null(target)) {
            if (est == "default") {
              sem(
                model   = mods1,
                data    = data[c(keys_s[[n_mod]], unlist(keys_e))],
                missing = miss,
                std.lv  = std.lv,
                orthogonal = orthogonal
              )
            } else {
              sem(
                model   = mods1,
                data    = data[c(keys_s[[n_mod]], unlist(keys_e))],
                missing = miss,
                estimator = est,
                std.lv  = std.lv,
                orthogonal = orthogonal
              )
            }
          } else {
            if (est == "default") {
              sem(
                model   = mods1,
                data    = data[c(keys_s[[n_mod]], unlist(keys_e))],
                missing = miss,
                std.lv  = std.lv,
                rotation = "target",
                rotation.args = list(
                  rstarts = 30, row.weights = "none", algorithm = "gpa",
                  std.ov = TRUE, target = target, orthogonal = orthogonal
                )
              )
            } else {
              sem(
                model   = mods1,
                data    = data[c(keys_s[[n_mod]], unlist(keys_e))],
                missing = miss,
                estimator = est,
                std.lv  = std.lv,
                rotation = "target",
                rotation.args = list(
                  rstarts = 30, row.weights = "none", algorithm = "gpa",
                  std.ov = TRUE, target = target, orthogonal = orthogonal
                )
              )
            }
          }
        }
      },
      mods1 = mods,
      m1 = m_test,
      hash_d1 = hash_d_test,
      n_mod = names(mods),
      ft = fit_type,
      n = seq_along(mods),
      SIMPLIFY = FALSE
    )
  )
  message("Generating parameter estimates")
  par <- mapply(
    function(m1, hash_d1, n_mod, fit1, pt, n) {
      if (hash_d1 & m1 & pt & param_test) {
        out <- par0[[n_mod]]
      } else {
        if (length(fit) <= 1000) {
          # Print progress for every model
          message(paste(n, "/", length(fit), " ", n_mod))
        } else {
          # Print progress for every 10th model (starting from 1)
          if (n %in% round((0:(length(fit)/10)*10) + 1, 0)) {
            message(paste0(n, "/", length(fit), " ", n_mod))
          }
        }
        if (std) {
          standardizedSolution(fit1)
        } else {
          parameterEstimates(fit1, standardized = TRUE)
        }
      }
    },
    fit1 = fit,
    m1 = m_test,
    hash_d1 = hash_d_test,
    n_mod = names(fit),
    pt = par_type,
    n = seq_along(fit),
    SIMPLIFY = FALSE
  )
  if (fit_save) {
    message("Generating model fit statistics")
    fit_m1 <- mapply(
      function(m1, hash_d1, n_mod, fit1, n) {
        # If the model's the same and fit measures exist...
        if (hash_d1 & m1 & (n_mod %in% rownames(fit_m0)) & param_test) {
          # If fit_measures == "all"
          if (fit_measures[1] == "all") {
            # If existing fit_m0 includes all fit stats...
            if (length(fit_m0[n_mod, ]) >= 55) {
              # Return existing
              fit_m0[n_mod, ]
            } else {
              # Else, recalculate
              message(paste(n, "/", length(fit), " ", n_mod))
              fitMeasures(fit1)
            }
          } else {
            # If fit_m0 includes all required fit measures
            if (
              sum(fit_measures %in% names(fit_m0[n_mod, ])) ==
              length(fit_measures)
            ) {
              # Return existing (with only required fit measures)
              fit_m0[n_mod, fit_measures]
            } else {
              # Else, recalculate with required fit measures
              message(paste(n, "/", length(fit), " ", n_mod))
              fitMeasures(fit1, fit.measures = fit_measures)
            }
          }
        } else {
          # If the model has changed, recalculate with required fit measures
          message(paste(n, "/", length(fit), " ", n_mod))
          if (fit_measures[1] == "all") {
            fitMeasures(fit1)
          } else {
            fitMeasures(fit1, fit.measures = fit_measures)
          }
        }
      },
      fit1 = fit,
      m1 = m_test,
      hash_d1 = hash_d_test,
      n_mod = names(fit),
      n = seq_along(fit),
      SIMPLIFY = FALSE
    )
    if (length(unlist(fit_m1)) == 0) {
      warning(
        "All the fit measures you have specified are not recognised by lavaan."
      )
    } else {
      fit_m1 <- do.call(rbind, args = fit_m1)
      if (ncol(fit_m1) != length(fit_measures) & fit_measures[1] != "all") {
        warning(
          paste(
            "The following fit measures that you specified are not recognised",
            "by lavaan:\n     ",
            paste(
              fit_measures[!fit_measures %in% colnames(fit_m1)],
              collapse = "\n      "
            ),
            collapse = ""
          )
        )
      }
    }
  }
  # Save models (so they can be read in next time)
  if (save_out) {
    saveRDS(fit, file.path(cache_dir, name, paste0(name, "_fit.rds")))
    if (std) {
      saveRDS(par, file.path(cache_dir, name, paste0(name, "_par_std.rds")))
    } else {
      saveRDS(par, file.path(cache_dir, name, paste0(name, "_par.rds")))
    }
    if (fit_save) {
      saveRDS(fit_m1, file.path(cache_dir, name, paste0(name, "_fit_m.rds")))
    }
    # Save mod + hash + params
    saveRDS(mods, file.path(cache_dir, name, paste0(name, "_mod.rds")))
    saveRDS(hash_d, file.path(cache_dir, name, paste0(name, "_hash.rds")))
    saveRDS(params, file.path(cache_dir, name, paste0(name, "_params.rds")))
  }
  # Return
  if (std) {
    if (fit_save) {
      return(list(fit = fit, par_std = par, fit_measures = fit_m1))
    } else {
      return(list(fit = fit, par_std = par))
    }
  } else {
    if (fit_save) {
      return(list(fit = fit, par = par, fit_measures = fit_m1))
    } else {
      return(list(fit = fit, par = par))
    }
  }
}

test_that(
  "'std' is not logical",
  {
    mods <- mapply(
      x = keys, y = names(keys), SIMPLIFY = FALSE,
      FUN = function(x, y) paste(y, "=~", paste(x, collapse = " + "))
    )
    expect_error(
      sem.check(mods, BFIGritHope, keys, std = 42),
      "'std' is not logical"
    )
  }
)
test_that(
  "'mods' is not a list",
  {
    mods <- mapply(
      x = keys, y = names(keys), SIMPLIFY = FALSE,
      FUN = function(x, y) paste(y, "=~", paste(x, collapse = " + "))
    )
    expect_error(
      sem.check(mods$grit_c, BFIGritHope, keys),
      "'mods' is not a list"
    )
  }
)
test_that(
  "'keys_s' and 'mods' are a different length",
  {
    mods <- mapply(
      x = keys, y = names(keys), SIMPLIFY = FALSE,
      FUN = function(x, y) paste(y, "=~", paste(x, collapse = " + "))
    )
    keys_l <- keys[-4]
    expect_error(
      sem.check(mods, BFIGritHope, keys_l),
      "'keys_s' and 'mods' are not the same length"
    )
  }
)

#' Runs an EFA model based on items in a keys list.
#'
#' `efa.from.keys` runs a exploratory factor analysis (EFA) in lavaan with a
#' rotation targeted based on a keys list.
#'
#' @inheritParams sem.check
#' @param keys
#' A named list of keys. Names must be factor names, elements must be
#' vectors of items that should be targeted to load on the factor.
#' @param data
#' A dataframe or object coercible to a dataframe.
#' Data must include all observed variables in any of the keys.
#' @param name
#' A string indicating a subdirectory where model outputs will be saved when
#' `save_out = TRUE` and checked against when `check = TRUE`.
#' Defaults to 'efa'.
#' Irrelevant if both `save_out = FALSE` and `check = FALSE`.
#' The name should be unique for each set of models, or outputs from calls with
#' the same name will be overwritten.
#' @param std.lv
#' Sets the `std.lv` param, as per lavaan (see [lavaan::lavOptions()]).
#' Defaults to `TRUE`.
#'
#' @return
#' Returns a list of lists.
#' The elements are a list of lavaan bifactor model output objects;
#' a list of parameter estimates from the models (standardized if `std = TRUE`);
#' and, if `fit_save = TRUE`, a matrix of fit measures for each model.
#'
#' @details
#' The function was designed to streamline running exploratory structural
#' equation models (ESEM) using Burt's (1976) 2-stage procedure to prevent
#' interpretational confounding in the context of ESEM. However, it can also be
#' used to easily run a targeted EFA with only a keys list to avoid having to
#' manually specify the target and model.
#'
#' The function was designed for use with established multidimensional scales,
#' such that a target is always reasonable.
#' The function does not currently support untargeted rotations.
#'
#' The model relies on [sem.check()] for the back-end of running the models.
#' This enables saving inputs and outputs from model runs
#' (with `save_out = TRUE`) and checking to see if anything has changed from
#' prior runs before running again (with `check = TRUE`).
#' The functionality was included for a number of very slow models or a lot of
#' faster models, such that time spent rerunning them would be onerous.
#' For further details on how this works, see the [sem.check()] function
#' documentation.
#'
#' @seealso
#' [sem.check()], which `efa.from.keys()` uses for all the back-end, and
#' [lavaan::sem()], which is used to estimate the models.
#'
#' @references
#' Burt, R. S. (1976).
#' Interpretational confounding of unobserved variables in Structural Equation
#' Models. Sociological Methods & Research, 5(1), 3-52.
#' https://doi.org/10.1177/004912417600500101.
#'
#' @importFrom lavaan summary
#' @export
#'
#' @examples
#' # Create EFA keys
#' # Using only 3 factors to save time
#' keys_e0 <- paste0("bfi_", c("e", "a", "c"))
#' # Using less than all items to save time
#' # (This results in a less than ideal solution but it shouldn't matter for an
#' # example)
#' keys_e <- sapply(
#'   keys_e0,
#'   function(x) {
#'     names(BFIGritHope)[grep(paste0(x, "\\d_[1-2]"), names(BFIGritHope))]
#'   },
#'   simplify = FALSE
#' )
#' # Run model
#' efa_fit <- efa.from.keys(keys_e, BFIGritHope, check = FALSE, fit_save = TRUE)
#' # Examine results
#' summary(efa_fit$fit$efa)                   # Standard lavaan summary
#' efa_fit$fit_measures[, c("cfi", "rmsea")]  # Fit measures

efa.from.keys <- function(
    keys, data, orthogonal = FALSE, fit_save = TRUE, fit_measures = "all",
    std.lv = TRUE, miss = "ML", est = "default",
    name = "efa", check = FALSE, save_out = FALSE
) {
  target <- sapply(keys, function(y) ifelse(!unlist(keys) %in% y, 0, NA))
  mod <- list(
    paste(
      paste0(
        paste0('efa("', name, '")*', names(keys), collapse = " + "),
        " =~\n",
        paste(unlist(keys), collapse = " + ")
      )
    )
  )
  names(mod) <- name
  sem.check(
    mod,
    data,
    name = name,
    keys_s = NULL,
    keys_e = keys,
    std = FALSE,  # For use in 2-stage procedure, must use non-standardised.
    fit_save = fit_save,
    fit_measures = fit_measures,
    orthogonal = orthogonal,
    std.lv = std.lv,
    target = target,
    miss = miss,
    est = est,
    check = check,
    save_out = save_out
  )
}

test_that(
  "Test normal behaviour when fit_save = FALSE",
  {
    efa_fit <- efa.from.keys(keys_e, BFIGritHope, check = FALSE, fit_save = FALSE)
    expect_equal(length(efa_fit), 2)
    expect_equal(length(efa_fit$fit), 1)
    expect_equal(length(efa_fit$par), 1)
    expect_equal(sum(sapply(efa_fit$fit, function(x) class(x) != "lavaan")), 0)
  }
)
test_that(
  "Test normal behaviour when orthogonal = TRUE",
  {
    efa_fit <- efa.from.keys(
      keys_e, BFIGritHope, check = FALSE, fit_save = FALSE, orthogonal = TRUE
    )
    expect_equal(length(efa_fit), 2)
    expect_equal(length(efa_fit$fit), 1)
    expect_equal(length(efa_fit$par), 1)
    expect_equal(sum(sapply(efa_fit$fit, function(x) class(x) != "lavaan")), 0)
    # Check orthogonal
    expect_equal(
      sum(
        abs(
          efa_fit$par$efa$est[
            grepl(
              paste0("^", names(keys_e), "$", collapse = "|"),
              efa_fit$par$efa$lhs
            ) &
              efa_fit$par$efa$op == "~~" &
              efa_fit$par$efa$lhs != efa_fit$par$efa$rhs
          ] > 10e-10
        )
      ),
      0
    )
  }
)
test_that(
  "Test normal behaviour when fit_save = TRUE",
  {
    efa_fit <- efa.from.keys(
      keys_e, BFIGritHope, check = FALSE, fit_save = TRUE
    )
    expect_equal(length(efa_fit), 3)
    expect_equal(length(efa_fit$fit), 1)
    expect_equal(length(efa_fit$par), 1)
    expect_equal(sum(sapply(efa_fit$fit, function(x) class(x) != "lavaan")), 0)
  }
)

# keys_e goes to keys_e in sem.check, not keys_s, so needs to be tested
test_that(
  "Misspecified keys",
  {
    keys_mistake <- keys_e
    keys_mistake$bfi_e[1] <- "mistake"
    expect_error(
      efa.from.keys(keys_mistake, BFIGritHope, check = FALSE, fit_save = FALSE),
      "items are in 'keys_e' but they are not in 'data'"
    )
    expect_error(
      efa.from.keys(keys_e$bfi_e, BFIGritHope, check = FALSE, fit_save = FALSE),
      "'keys_e' is not a list"
    )
  }
)
test_that(
  "Keys with the same name",
  {
    keys_e_nm <- keys_e
    names(keys_e_nm)[1:2] <- "bfi_e"
    expect_error(
      efa.from.keys(keys_e_nm, BFIGritHope, check = FALSE, fit_save = FALSE),
      "two elements of 'keys_e' share the same name"
    )
  }
)
test_that(
  "Key of length 1 warning",
  {
    keys_l1 <- keys_e
    keys_l1$bfi_e <- keys_e$bfi_e[1]
    expect_warning(
      efa.from.keys(keys_l1, BFIGritHope, check = FALSE, fit_save = FALSE),
      "factors have only 1 item"
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after changes to orthogonal",
  {
    cache_dir <- cache.setup("tests/testthat")
    check_fit <- efa.from.keys(
      keys_e, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      efa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        orthogonal = TRUE
      ),
      "1 / \\d"
    )
  }
)
test_that(
  "Non-logical orthogonal setting",
  {
    expect_error(
      efa.from.keys(
        keys_e, BFIGritHope, check = FALSE, fit_save = FALSE, orthogonal = 42
      )
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after changes to est",
  {
    cache.setup("tests/testthat")
    check_fit <- efa.from.keys(
      keys_e, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      efa.from.keys(
        keys_e, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        est = "MLR"
      ),
      "1 / \\d"
    )
  }
)

#' Cleans selected files from the current cache directory
#'
#' Functions in the package include options to write files to a cache directory,
#' set up with the [cache.setup()] function.
#' Obsoletely files might be created if the `name` parameter in a function call
#' is changed or if a function call is no longer being used.
#' The `cache.clean()` function provides a way to find files that have not be
#' modified in the last `older_than` days and lists the files for the user to
#' agree to delete or not.
#'
#' @param older_than
#' A positive number indicating number of days (fractions allowed).
#' Files with older modification times than this will be deleted
#' after confirmation if `interactive = TRUE`.
#' @param interactive
#' Logical.
#' `TRUE` indicates that confirmation will be required before files are deleted.
#' `FALSE` indicates that files will be deleted without requiring confirmation.
#' It is recommended to use `TRUE` except for carefully checked automated
#' processes.
#' Defaults to `TRUE`.
#'
#' @return `NULL` (invisibly). Primarily called to clean up the cache directory.
#'
#' @details
#' The function is interactive if run in an interactive session with
#' `interctive = TRUE`.
#' In addition to deleting old files, the function will also optionally delete
#' empty directories and, if the top level cache directory is also empty,
#' then it will also optionally delete the cache directory and unset it.
#' The latter will only be done in interactive sessions with
#' `interactive = TRUE` to avoid breaking non-interactive code.
#'
#' One way to clean only unused files is to run all current code and then
#' run `cache.clean()` with `older_than` set to something greater than the
#' number of days it takes for the code to run and less than when the code was
#' previously run. For example, if your code takes 5 minutes to run and you
#' previously ran it 2 days ago, you could run all your code and, when it has
#' finished, use `cache.clean(older_than = 1)`.
#'
#' `semFromKeys` has no way to know what directories you have specified as the
#' cache in the past and cannot clean up unknown former cache directories.
#' Therefore, it is recommended that you use either the default location or
#' the same location for all models within the same project.
#' Either of these will enable easy detection, so that unneeded cached files
#' can be found and deleted.
#'
#' @seealso [cache.setup()]
#'
#' @export
#'
#' @examples
#' \dontrun{
#'   # Setup a cache directory
#'   cache.setup()
#'   # Now code with check = TRUE or save_out = TRUE will work, e.g.,
#'   # Create CFA keys
#'   keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#'   keys <- sapply(
#'     keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#'   )
#'   # Run models
#'   cfa_fit <- cfa.from.keys(
#'     keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
#'   )
#'   # Check that they are not estimated again.
#'   cfa_fit <- cfa.from.keys(
#'     keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
#'   )
#'
#'   # The cache.clean lines are commented out so you do not inadvertently
#'   # delete any models created with your own code in the same cache directory.
#'
#'   # cache.clean(30)  # Delete files not modified in the last 30 days.
#'   # cache.clean(60/86400)  # Delete files not modified in the last minute.
#'   # Delete all files matching those created by the package in the current
#'   # cache directory and subdirectories without confirmation.
#'   # cache.clean(0, interactive = FALSE)
#' }

cache.clean <- function(older_than = NULL, interactive = TRUE) {
  if (is.null(older_than)) {
    stop(
      "Please specify a value for 'older_than'. To delete all files, set to 0."
    )
  }
  found <- FALSE
  for (i in 1:5) {
    env <- parent.frame(i)
    if (exists(".cache_env", envir = env)) {
      found <- TRUE
      break
    }
  }
  if (!found) {
    stop(
      paste(
        "A cache directory is not configured so cannot be cleaned.",
        "Use the 'cache.setup()' function to configure a directory to clean."
      )
    )
  }
  cache_dir <- get("cache_dir", envir = get(".cache_env", envir = env))
  files <- file.info(
    list.files(cache_dir, full.names = TRUE, recursive = TRUE)
  )
  cutoff <- Sys.time() - (older_than * 86400)  # convert days to seconds
  files_del <- rownames(files)[
    files$mtime < cutoff &
      # Don't delete files that the package doesn't save.
      grepl("_(fit(|_m)|par(|_std|ams)|mod|hash).rds$", rownames(files))
  ]
  if (length(files_del) == 0) {
    message("No files to delete.")
  } else {
    # Prompt user if interactive
    if (!interactive() & interactive) {
      stop(
        paste(
          "Running 'cache.clean' in a non-interactive session with",
          "'interactive = TRUE' does not delete anything to prevent unexpected",
          "data loss.",
          "If you want to clean the cache in a non non-interactive session,",
          "please set 'interactive = TRUE'."
        )
      )
    }
    if (interactive) {
      message(
        paste0(
          "About to delete the following ", length(files_del),
          " file(s), older than ", older_than, " from\n '",
          cache_dir, "':\n\n  ",
          paste0(
            sub(paste0(cache_dir, ".*/"), "", files_del), collapse = "\n  "
          )
        )
      )
      response <- readline("Continue? (y/n): ")
      if (tolower(substr(response, 1, 1)) != "y") {
        message("Cancelled.")
        return(invisible(NULL))
      }
    }
    # Delete files
    unlink(files_del)
    message(paste("Deleted", length(files_del), "file(s)."))
  }
  dirs <- list.dirs(cache_dir, full.names = TRUE, recursive = TRUE)
  dirs_empty <- sapply(
    dirs,
    function(x) {
      tmp <- list.files(x, full.names = TRUE, recursive = TRUE)
      length(tmp) == 0
    }
  )
  dirs_del <- dirs[dirs_empty]
  dirs_del <- dirs_del[-1]  # Don't delete parent (i.e., cache_dir)
  if (length(dirs_del) > 0) {
    if (!interactive() & interactive) {
      stop(
        paste(
          "Running 'cache.clean' in a non-interactive session with",
          "'interactive = TRUE' does not delete anything to prevent unexpected",
          "data loss.",
          "If you want to clean the cache in a non non-interactive session,",
          "please set 'interactive = TRUE'."
        )
      )
    }
    if (interactive) {
      message(
        paste0(
          "\nAbout to delete the following ", length(dirs_del) - 1,
          " empty directories from\n '",
          cache_dir, "':\n\n  ",
          paste0(
            sub(paste0(cache_dir, ".*/"), "", dirs_del[-1]), collapse = "\n  "
          )
        )
      )
      response <- readline("Continue? (y/n): ")
      if (tolower(substr(response, 1, 1)) != "y") {
        message("Cancelled.")
        return(invisible(NULL))
      }
    }
    # Delete empty directories
    unlink(dirs_del)
  }
  if (length(list.files(cache_dir, full.names = TRUE, recursive = TRUE)) > 0) {
    return(invisible(NULL))
  }
  if (!interactive() & interactive) {
    stop(
      paste(
        "Running 'cache.clean' in a non-interactive session with",
        "'interactive = TRUE' does not delete anything to prevent unexpected",
        "data loss.",
        "If you want to clean the cache in a non non-interactive session,",
        "please set 'interactive = TRUE'."
      )
    )
  }
  if (interactive) {
    message(paste0("The top cache directory, '", cache_dir, "', is empty."))
    response <- readline(
      paste(
        "Do you want to delete it?",
        "Doing so will unset the cache directory (y/n): "
      )
    )
    if (tolower(substr(response, 1, 1)) != "y") {
      message("Cancelled.")
      return(invisible(NULL))
    }
  }
  unlink(cache_dir)
  rm(.cache_env, envir = env)
  message(
    paste0(
      "'", cache_dir, "' has been deleted.\n",
      "To use 'check = TRUE' or 'save_out = TRUE',",
      "'cache.setup()' will have to be reinitiated."
    )
  )
  return(invisible(NULL))
}

test_that(
  "Test normal behaviour with fit_save = FALSE",
  {
    cfa_fit <- cfa.from.keys(keys, BFIGritHope, fit_save = FALSE)
    expect_equal(length(cfa_fit), 2)
    expect_equal(length(cfa_fit$fit), length(keys))
    expect_equal(length(cfa_fit$par), length(keys))
    expect_equal(sum(sapply(cfa_fit$fit, function(x) class(x) != "lavaan")), 0)
  }
)
test_that(
  "Test normal behaviour with fit_save = TRUE",
  {
    cfa_fit <- cfa.from.keys(keys, BFIGritHope, fit_save = TRUE)
    expect_equal(length(cfa_fit), 3)
    expect_equal(length(cfa_fit$fit), length(keys))
    expect_equal(length(cfa_fit$par), length(keys))
    expect_equal(sum(sapply(cfa_fit$fit, function(x) class(x) != "lavaan")), 0)
  }
)
test_that(
  "Test keys not specified",
  {
    expect_error(
      cfa.from.keys(keys = NULL, BFIGritHope, fit_save = FALSE),
      "one of 'keys_s' or 'keys_e'"
    )
    expect_error(
      cfa.from.keys(data = BFIGritHope, fit_save = FALSE),
      'argument "keys" is missing'
    )
  }
)
test_that(
  "No data provided",
  {
    expect_error(
      cfa.from.keys(keys, fit_save = FALSE),
      'argument "data" is missing'
    )
  }
)
test_that(
  "Various things 'not logical'",
  {
    expect_error(
      cfa.from.keys(keys, BFIGritHope, fit_save = 42),
      "'fit_save' is not logical"
    )
    expect_error(
      cfa.from.keys(keys, BFIGritHope, check = 42, fit_save = FALSE),
      "'check' is not logical"
    )
    expect_error(
      cfa.from.keys(keys, BFIGritHope, fit_save = FALSE, save_out = 42),
      "'save_out' is not logical"
    )
    expect_error(
      cfa.from.keys(keys, BFIGritHope, fit_save = FALSE, std.lv = 42),
      "'std.lv' is not logical"
    )
  }
)
test_that(
  "'fit_measures' is not a character vector",
  {
    expect_error(
      cfa.from.keys(keys, BFIGritHope, fit_save = TRUE, fit_measures = 42),
      "'fit_measures' is not a character vector"
    )
  }
)
test_that(
  "Incorrect fit measures",
  {
    expect_warning(
      cfa.from.keys(
        keys, BFIGritHope, check = FALSE, fit_save = TRUE,
        fit_measures = c("NotAFitMeasure", "AlsoNotAFitMeasure")
      ),
      "All the fit measures"
    )
    expect_warning(
      cfa.from.keys(
        keys, BFIGritHope, check = FALSE, fit_save = TRUE,
        fit_measures = c("NotAFitMeasure", "AlsoNotAFitMeasure", "cfi")
      ),
      "fit measures that you specified are not recognised"
    )
  }
)
test_that(
  "Mistakes in keys",
  {
    keys_mistake <- keys
    keys_mistake$grit_c[1] <- "mistake"
    expect_error(
      cfa.from.keys(keys_mistake, BFIGritHope, fit_save = FALSE),
      "items are in 'keys_s' but they are not in 'data'"
    )
  }
)
test_that(
  "Non-list keys list",
  {
    expect_error(
      cfa.from.keys(keys$grit_c, BFIGritHope, fit_save = FALSE),
      "'keys_s' is not a list"
    )
  }
)
test_that(
  "Short keys",
  {
    keys_l2 <- keys
    keys_l2$grit_c <- keys$grit_c[1:2]
    expect_warning(
      cfa.from.keys(keys_l2, BFIGritHope, fit_save = FALSE),
      "only length 2"
    )
    keys_l1 <- keys
    keys_l1$grit_c <- keys$grit_c[1]
    expect_error(
      cfa.from.keys(keys_l1, BFIGritHope, fit_save = FALSE),
      "only length 1"
    )
  }
)
test_that(
  "Keys with the same name",
  {
    keys_nm <- keys
    names(keys_nm)[1:2] <- "grit"
    expect_error(
      cfa.from.keys(keys_nm, BFIGritHope, fit_save = FALSE),
      "two elements of 'keys_s' share the same name"
    )
  }
)
test_that(
  "Unrecognised missing specification",
  {
    expect_error(
      cfa.from.keys(keys, BFIGritHope, fit_save = FALSE, miss = "banana"),
      "invalid value in missing"
    )
  }
)
test_that(
  "Caching not set up",
  {
    expect_error(
      cfa.from.keys(keys, BFIGritHope, check = TRUE, fit_save = FALSE),
      "A cache directory is not configured"
    )
    expect_error(
      cfa.from.keys(keys, BFIGritHope, save_out = TRUE, fit_save = FALSE),
      "A cache directory is not configured"
    )
  }
)
test_that(
  "'name' not a string (but a number) when required.",
  {
    cache.setup("tests/testthat/cache", interactive = FALSE)
    expect_error(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = FALSE,
        name = 42
      ),
      "'name' is not a character"
    )
  }
)
test_that(
  "'name' not a string (but NULL) when required.",
  {
    cache.setup("tests/testthat/cache", interactive = FALSE)
    expect_error(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = FALSE,
        name = NULL
      ),
      "'name' is not a character"
    )
  }
)
test_that(
  "'name' is not a string when not required",
  {
    expect_no_error(
      cfa.from.keys(
        keys, BFIGritHope, check = FALSE, save_out = FALSE, fit_save = FALSE,
        name = 42
      )
    )
    expect_no_error(
      cfa.from.keys(
        keys, BFIGritHope, check = FALSE, save_out = FALSE, fit_save = FALSE,
        name = NULL
      )
    )
  }
)
test_that(
  "Data incorrectly specified",
  {
    expect_error(
      cfa.from.keys(keys, BFIGritHope[NULL, ], fit_save = FALSE),
      "some variables have no values"
    )
    expect_error(
      cfa.from.keys(keys, data = "cfa", fit_save = FALSE),
      "items are in 'keys_s' but they are not in 'data'"
    )
  }
)
test_that(
  "Test 'save_out = TRUE' file creation and 'check = TRUE' correctly loading",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    name <- "cfa"
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
      name = name
    )
    expect_all_true(
      c(
        file.exists(file.path(cache_dir, name, paste0(name, "_fit.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_par.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_fit_m.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_mod.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_hash.rds")))
      )
    )
    check_fit2 <- expect_no_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        name = name
      ),
      message = "\\d / \\d"
    )
    expect_identical(check_fit, check_fit2)
  }
)
test_that(
  "Cache location invalid",
  {
    expect_error(
      cache.setup(c("tests/testthat/cache", "tests/testhtat/invalid")),
      "the condition has length > 1"
    )
    expect_error(
      cache.setup(42), "'location' is not a length 1 character vector"
    )
  }
)
test_that(
  "Test partial running on 'check = TRUE' after changes to a model",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    name <- "cfa"
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
      name = name
    )
    # Change a model
    cfa_mod <- readRDS(file.path(cache_dir, name, paste0(name, "_mod.rds")))
    cfa_mod[3] <- sub(" \\+ hope_a_4", "", cfa_mod[3])
    saveRDS(cfa_mod, file.path(cache_dir, name, paste0(name, "_mod.rds")))
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        name = name
      ),
      "3 / \\d"
    )
    saveRDS(cfa_mod, file.path(cache_dir, name, paste0(name, "_mod.rds")))
    expect_no_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        name = name
      ),
      message = "([1-2]|4) / \\d"
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after changing to full 'fit_measures' set",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
      fit_measures = c("chisq", "cfi", "rmsea")
    )
    check_fit2 <- expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE
      ),
      "1 / \\d"
    )
    expect_all_true(
      c(ncol(check_fit$fit_measures) == 3, ncol(check_fit2$fit_measures) >= 55)
    )
  }
)
# This is treated slightly differently to the above in sem.check
# due to distinction between 'fit_measures = "all"' and a subset.
test_that(
  "Test running on 'check = TRUE' after adding fit measures (not to full set)",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
      fit_measures = "cfi"
    )
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
        fit_measures = c("cfi", "rmsea")
      ),
      "1 / \\d"
    )
  }
)
test_that(
  "Test partial running on 'check = TRUE' after changes to a data hash",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    name <- "cfa"
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
      name = name
    )
    # Change a model
    cfa_hash <- readRDS(file.path(cache_dir, name, paste0(name, "_hash.rds")))
    cfa_hash[3] <- "helloworld123"
    saveRDS(cfa_hash, file.path(cache_dir, name, paste0(name, "_hash.rds")))
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        name = name
      ),
      "3 / \\d"
    )
    saveRDS(cfa_hash, file.path(cache_dir, name, paste0(name, "_hash.rds")))
    expect_no_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        name = name
      ),
      message = "([1-2]|4) / \\d"
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after changes to miss",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        miss = "pairwise"
      ),
      "1 / \\d"
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after changes to est",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        est = "MLR"
      ),
      "1 / \\d"
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after changes to std.lv",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        std.lv = FALSE
      ),
      "1 / \\d"
    )
  }
)
test_that(
  "Test running on 'check = TRUE' after selecting subset of 'fit_measures'",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    check_fit <- cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    check_fit2 <- expect_no_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE,
        fit_measures = c("chisq", "cfi", "rmsea")
      ),
      message = "\\d / \\d"
    )
    expect_all_true(
      c(ncol(check_fit$fit_measures) >= 55, ncol(check_fit2$fit_measures) == 3)
    )
  }
)
test_that(
  "Test adding a model between saved runs",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    cfa.from.keys(
      keys[-1], BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
      ),
      "1 / \\d"
    )
    cfa.from.keys(
      keys[-4], BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE
    )
    expect_message(
      cfa.from.keys(
        keys, BFIGritHope, check = TRUE, save_out = FALSE, fit_save = TRUE
      ),
      "4 / \\d"
    )
  }
)
test_that(
  "Test deleting of cache files",
  {
    cache_dir <- cache.setup("tests/testthat/cache", interactive = FALSE)
    name <- "cfa"
    cfa.from.keys(
      keys, BFIGritHope, check = TRUE, save_out = TRUE, fit_save = TRUE,
      name = name
    )
    expect_all_true(
      c(
        file.exists(file.path(cache_dir, name, paste0(name, "_fit.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_par.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_fit_m.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_mod.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_hash.rds")))
      )
    )
    cache.clean(0, interactive = FALSE)
    expect_equal(
      length(list.files(cache_dir, full.names = TRUE, recursive = TRUE)), 0
    )
  }
)

test_that(
  "Cache directory not set",
  {
    expect_error(
      cache.clean(0, interactive = FALSE), "cache directory is not configured"
    )
  }
)
test_that(
  "Test 'older_than' not set with cache.setup",
  {
    cache.setup("tests/testthat/cache", interactive = FALSE)
    expect_error(
      cache.clean(interactive = FALSE), "specify a value for 'older_than'"
    )
  }
)
test_that(
  "No files to delete",
  {
    cache.setup("tests/testthat/cache", interactive = FALSE)
    expect_message(cache.clean(0, interactive = FALSE), "No files to delete")
  }
)

#' Sets up a cache directory for storing model inputs and outputs
#'
#' Using `save_out = TRUE` or `check = TRUE` from
#' [cfa.from.keys()], [bifactor.from.keys()], [efa.from.keys()], or
#' [esem.from.mods()] requires a cache directory to save model objects to or
#' to check against.
#' By default, `cache.setup` uses the system's default cache location,
#' otherwise a location within the current working directory or project can be
#' selected.
#' The function needs to be run once after the environment is cleared whenever
#' a cache directory is required.
#'
#' @param location
#' The cache location to save model objects to to enable checking.
#' If `location = "user"` (default, recommended if unsure),
#' then the system's default cache location is used.
#' If any other string is used a folder will be created within either the
#' [getwd()] directory or,
#' if available, within the directory identified by [here::here()].
#' If using an RStudio project, these will create the directory in the
#' project directory.
#' @param interactive
#' Logical.
#' `TRUE` indicates that confirmation will be required before a cache directory
#' is set *if* the default location is not used.
#' `FALSE` indicates that a cache directory will be set without confirmation.
#' Irrelevant if `location = "user"`.
#' It is recommended to use `TRUE`.
#' Defaults to `TRUE`.
#'
#' @return
#' The cache directory path (invisibly).
#' Primarily called to set up the cache configuration.
#'
#' @details
#' Saving outputs to save time running identical models more than once requires
#' files to be saved on your computer.
#' To avoid writing to your computer without your permission,
#' you are required to run this function first to ensure that you know *that*
#' files are being saved and *where* files are being saved.
#' The function temporarily sets a hidden environment variable '.cache_env'
#' (with, if empty,
#' `assign(".cache_env", new.env(parent = emptyenv()), envir = parent.frame(1))`
#' and with
#' `assign("cache_dir", cache_dir, envir = get(".cache_env", envir = parent.frame(1)))`,
#' if not), which will be removed whenever the environment is cleared.
#'
#' By default, the function creates the cache directory in the standard place
#' for the operating system being used, appended by `semFromKeys/[projectname]`,
#' if a project is being used and the `rstudioapi` is available.
#' If a project is not being used or the `rstudioapi` package is not available,
#' then the relative path with be `semFromKeys`.
#'
#' In the later case, if two function calls include the same `name` assignment,
#' then outputs from one will overwrite the other. Therefore, it is recommended
#' to either set the cache directory to something other than the default or use
#' the default within a project with the `rstudioapi` package installed.
#' Obviously this latter option will likely not work outside of RStudio,
#' so, in that case, it is recommended to use a custom location.
#'
#' To ensure that R knows what the cache directory is,
#' a hidden environment variable containing the information---`.cache_env`---is
#' saved within the working environment.
#' Other functions from the package will look for and use `.cache_env` to
#' identify the cache directory.
#' As a result, whenever the working environment is cleared
#' (e.g., upon closing RStudio) or whenever the `.cache_env` is otherwise
#' removed,
#' the `cache.setup` function will need to be re-run before the cache
#' functionality will work,
#' regardless of the status of the any recently used cache directories.
#'
#' The function relies on code that was unavailable in versions of R prior to
#' version 4.0,
#' so anyone using an earlier version of R will either have to update R or forgo
#' the caching functionality.
#'
#' `semFromKeys` has no way to know what directories you have specified as the
#' cache in the past and cannot clean up unknown former cache directories.
#' Therefore, it is recommended that you use either the default location or
#' the same location for all models within the same project.
#' Either of these will enable easy detection, so that unneeded cached files
#' can be found and deleted.
#'
#' Functions that directly or indirectly might require a cache directory are:
#' [cfa.from.keys()], [bifactor.from.keys()], [efa.from.keys()],
#' [esem.from.mods()], and [sem.check()].
#'
#' @seealso
#' [cfa.from.keys()], [bifactor.from.keys()], [efa.from.keys()],
#' [esem.from.mods()], and [sem.check()] for dependent functions;
#' [tools::R_user_dir()] for default cache directories;
#' [here::here()] for project-based cache directory setting; and
#' [cache.clean()] for a function to clean cache.
#'
#' @export
#'
#' @examples
#' \dontrun{
#'   # Setup a cache directory
#'   cache.setup()
#'   # Now code with check = TRUE or save_out = TRUE will work, e.g.,
#'   # Create CFA keys
#'   keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#'   keys <- sapply(
#'     keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#'   )
#'   # Run models
#'   cfa_fit <- cfa.from.keys(
#'     keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
#'   )
#'   # Check that models are not run again.
#'   cfa_fit <- cfa.from.keys(
#'     keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
#'   )
#' }

cache.setup <- function(location = "user", interactive = TRUE) {
  if (getRversion() < "4.0") {
    stop(
      paste0(
        "Setting up cache requires R >= 4.0. ",
        "Your system is currently running version ", getRversion(), ". ",
        "Please update your R version or use 'check = FALSE' and ",
        "'save_out = FALSE' for all function calls with these options."
      )
    )
  }
  if (location == "user") {
    if (!requireNamespace("rstudioapi")) {
      cache_dir <- tools::R_user_dir("semFromKeys", which = "cache")
    } else {
      if (rstudioapi::isAvailable()) {
        project <- rstudioapi::getActiveProject()
        if (is.null(project)) {
          cache_dir <- tools::R_user_dir("semFromKeys", which = "cache")
        } else {
          project <- sub(".*/", "", project)
          cache_dir <- tools::R_user_dir(
            paste0("semFromKeys/", project), which = "cache"
          )
        }
      } else {
        cache_dir <- tools::R_user_dir("semFromKeys", which = "cache")
      }
    }
  } else {
    if (!is.character(location)) {
      stop("'location' is not a length 1 character vector")
    }
    if (interactive & interactive()) {
      if (requireNamespace("here")) {
        message(
          paste0(
            "The cache director will be set as: '",
            paste0(here::here(), "/", location), "'."
          )
        )
        response <- readline("Continue? (y/n): ")
      } else {
        message(
          paste0(
            "The cache director will be set as: ",
            paste0(getwd(), "/", location), "."
          )
        )
        response <- readline("Continue? (y/n): ")
      }
      if (tolower(substr(response, 1, 1)) != "y") {
        message("Cancelled. Cache director not set.")
        return(invisible(NULL))
      }
    }
    if (requireNamespace("here")) {
      cache_dir <- paste0(here::here(), "/", location)
    } else {
      cache_dir <- paste0(getwd(), "/", location)
    }
  }
  if (!dir.exists(cache_dir)) dir.create(cache_dir, recursive = TRUE)
  if (!exists(".cache_env", mode = "environment", envir = parent.frame(1))) {
    assign(".cache_env", new.env(parent = emptyenv()), envir = parent.frame(1))
  }
  assign(
    "cache_dir", cache_dir, envir = get(".cache_env", envir = parent.frame(1))
  )
  message(paste0("The cache directory is set as: '", cache_dir, "'."))
  return(invisible(cache_dir))
}

test_that(
  "Test normal behaviour: EFA and CFA only",
  {
    esem_fit <- esem.from.mods(
      efa_fit, cfa_fit, data = BFIGritHope, fit_save = FALSE
    )
    expect_equal(length(esem_fit), 4)
    expect_equal(length(esem_fit$fit), length(keys))
    expect_equal(length(esem_fit$par), length(keys))
    expect_equal(
      sum(sapply(esem_fit$fit, function(x) !inherits(x, "lavaan"))), 0
    )
    expect_equal(length(esem_fit$b), length(keys))
    expect_equal(nrow(esem_fit$r2), length(keys))
  }
)
test_that(
  "Test normal behaviour: EFA and bifactor only",
  {
    esem_fit <- esem.from.mods(
      efa_fit, bif_fit = bif_fit, data = BFIGritHope, fit_save = FALSE
    )
    expect_equal(length(esem_fit), 4)
    expect_equal(length(esem_fit$fit), length(keys_g))
    expect_equal(length(esem_fit$par), length(keys_g))
    expect_equal(
      sum(sapply(esem_fit$fit, function(x) !inherits(x, "lavaan"))), 0
    )
    expect_equal(length(esem_fit$b), length(keys_g))
    expect_equal(nrow(esem_fit$r2), length(keys_g))
  }
)
test_that(
  "Test normal behaviour: EFA and both CFA and bifactor with fit_save = TRUE",
  {
    esem_fit <- esem.from.mods(
      BFIGritHope, efa_fit, cfa_fit = cfa_fit, bif_fit = bif_fit,
      fit_save = TRUE
    )
    expect_equal(length(esem_fit), 5)
    expect_equal(length(esem_fit$fit), length(keys_g) + length(keys))
    expect_equal(length(esem_fit$par), length(keys_g) + length(keys))
    expect_equal(
      sum(sapply(esem_fit$fit, function(x) !inherits(x, "lavaan"))), 0
    )
    expect_equal(nrow(esem_fit$fit_measures), length(keys_g) + length(keys))
    expect_equal(length(esem_fit$b), length(keys_g) + length(keys))
    expect_equal(nrow(esem_fit$r2), length(keys_g) + length(keys))
  }
)
test_that(
  "Neither cfa_fit nor bif_fit set",
  {
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit),
      "one of 'cfa_fit' and 'bif_fit' must be specified"
    )
  }
)
test_that(
  "Non-lavaan object inputs",
  {
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit, cfa_fit = list(a = 1:2, b = 3:4)),
      "'cfa_fit' are not objects of type lavaan"
    )
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit, bif_fit = list(a = 1:2, b = 3:4)),
      "'bif_fit' are not objects of type lavaan"
    )
    expect_error(
      esem.from.mods(
        BFIGritHope, efa_fit = list(a = 1:2, b = 3:4), cfa_fit = cfa_fit
      ),
      "'efa_fit' is not an object of type lavaan"
    )
  }
)
test_that(
  "efa_fit is NULL",
  {
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit = NULL, cfa_fit),
      "'efa_fit' is NULL"
    )
  }
)
test_that(
  "Multi-factor CFA",
  {
    cfa2_mod <- paste0(
      mapply(
        x = keys[1:2], y = names(keys[1:2]),
        FUN = function(x, y) paste(y, "=~", paste(x, collapse = " + "))
      ),
      collapse = "\n"
    )
    cfa2_fit <- list(grit = lavaan::cfa(cfa2_mod, BFIGritHope))
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit, cfa2_fit),
      "CFA containing more than one latent variable has been found"
    )
  }
)
test_that(
  "Factors with the same names",
  {
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit, cfa_fit[c(1, 1:4)]),
      "two different models in 'cfa_fit' have factors with the same name"
    )
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit, bif_fit = bif_fit[c(1, 1:2)]),
      "different models in 'bif_fit' have general factors with the same name"
    )
  }
)
test_that(
  "CFA factor with the same name as a bifactor general factor",
  {
    keys2 <- keys
    names(keys2)[1] <- "grit"
    cfa_fit2 <- cfa.from.keys(keys2, BFIGritHope, fit_save = FALSE)$fit
    expect_error(
      esem.from.mods(BFIGritHope, efa_fit, cfa_fit2, bif_fit),
      "models in 'cfa_fit' have identically named factor"
    )
  }
)
# Item in a group factor but not in the general factor should be sorted during
# bifactor creation, not here.
test_that(
  "No error from items in group factor not in general factor",
  {
    keys_g2 <- keys_g
    keys_g2$grit <- keys_g2$grit[-1]
    bif_fit2 <- suppressWarnings(  # In preamble, not important for the test.
      bifactor.from.keys(
        keys_g2, keys_b, keys, BFIGritHope, fit_save = FALSE
      )$fit
    )
    expect_no_error(esem.from.mods(BFIGritHope, efa_fit, bif_fit = bif_fit2))
  }
)
test_that(
  "Test 'save_out = TRUE' file creation and 'check = TRUE' correctly loading",
  {
    cache_dir <- cache.setup("tests/testthat", interactive = FALSE)
    name <- "esem"
    check_fit <- esem.from.mods(
      BFIGritHope, efa_fit, cfa_fit, check = TRUE, save_out = TRUE,
      fit_save = TRUE, name = name
    )
    expect_all_true(
      c(
        file.exists(file.path(cache_dir, name, paste0(name, "_fit.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_par_std.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_fit_m.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_mod.rds"))),
        file.exists(file.path(cache_dir, name, paste0(name, "_hash.rds")))
      )
    )
    check_fit2 <- expect_no_message(
      esem.from.mods(
        BFIGritHope, efa_fit, cfa_fit, check = TRUE, save_out = TRUE,
        fit_save = TRUE, name = name
      ),
      message = "\\d / \\d"
    )
    expect_identical(check_fit, check_fit2)
  }
)

#' Runs CFA models for multiple scales based on items in a keys list.
#'
#' `cfa.from.keys` runs a confirmatory factor analysis (CFA) model for each
#' element of a keys list. The keys list must be a named list of scales, where
#' each element is an item from the corresponding scale. The function is
#' designed to streamline running CFA models for all scales in a sample and to
#' input model outputs into downstream functions.
#'
#' @inheritParams sem.check
#' @param keys
#' A named list of keys.
#' Names should be scale names,
#' elements should a list of items included in each scale.
#' @param data
#' A dataframe or object coercible to a dataframe.
#' Data must include all observed variables in any of the keys.
#' @param name
#' A string indicating a subdirectory where model outputs will be saved when
#' `save_out = TRUE` and checked against when `check = TRUE`.
#' Defaults to 'cfa'.
#' Irrelevant if both `save_out = FALSE` and `check = FALSE`.
#' The name should be unique for each set of models, or outputs from calls with
#' the same name will be overwritten.
#' @param std.lv
#' Logical.
#' Sets the `std.lv` parameter, as per lavaan (see [lavaan::lavOptions()]).
#' `TRUE` indicates that factor variances should be fixed to 1.
#' `FALSE` indicates that loadings of the first items of factors should be fixed
#' to 1.
#' Defaults to `TRUE`.
#'
#' @return
#' Returns a list of length 2 (if `fit_save = FALSE`) or
#' 3 (if `fit_save = TRUE`).
#' The elements of the list are: a list of lavaan model output objects;
#' a list of parameter estimates from the models (standardized if `std = TRUE`);
#' and, if `fit_save = TRUE`, a matrix of fit measures for each model.
#'
#' @details
#' The model relies on [sem.check()] for the back-end of running the models.
#' This enables saving inputs and outputs from model runs
#' (with `save_out = TRUE`) and checking to see if anything has changed from
#' prior runs before running again (with `check = TRUE`).
#' The functionality was included for a number of very slow models or a lot of
#' faster models, such that time spent rerunning them would be onerous.
#' For further details on how this works, see the [sem.check()] function
#' documentation.
#'
#' The function does not provide any warnings for poor fit beyond those provided
#' by lavaan.
#' If any CFA models have poor fit, there is currently no capability to update
#' them beyond removing items by omitting them from the keys.
#' Any other changes to CFA models have to be made manually currently.
#' (Note that with `save_out = TRUE`,
#' model code is saved in `file.path(out_dir, name, paste0(name, _mod.rds)`,
#' which could help with manually updating models.)
#'
#' @seealso
#' [sem.check()], which `cfa.from.keys()` uses for all the back-end, and
#' [lavaan::sem()], which is used to estimate the models.
#'
#' @importFrom lavaan summary
#' @export
#'
#' @examples
#' # Create CFA keys
#' keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#' keys <- sapply(
#'   keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#' )
#' # Run models
#' cfa_fit <- cfa.from.keys(keys, BFIGritHope, check = FALSE, fit_save = TRUE)
#' # Examine some results
#' summary(cfa_fit$fit$grit_c)                # Standard lavaan summary
#' cfa_fit$fit_measures[, c("cfi", "rmsea")]  # Fit measures

cfa.from.keys <- function(
    keys, data, fit_save = TRUE, fit_measures = "all",
    std.lv = TRUE, miss = "ML", est = "default",
    name = "cfa", check = FALSE, save_out = FALSE
) {
  mods <- mapply(
    function(y, z) {
      if (length(z) > 2) {
        paste(y, "=~", paste(sapply(z, "["), collapse = " + "))
      } else {
        if (length(z) == 2) {
          warning(paste(y, "is only length 2. Model results may be poor."))
          paste0(y, " =~ 1 * ", z[[1]], "\n", y, " =~ ", z[[2]])
        } else {
          stop(paste(y, "is only length 1. Model cannot be run."))
        }
      }
    },
    y = names(keys), z = keys, SIMPLIFY = FALSE
  )
  sem.check(
    mods,
    data,
    name = name,
    keys_s = keys,
    keys_e = NULL,
    std = FALSE,  # For use in 2-stage procedure, must use non-standardised.
    fit_save = fit_save,
    fit_measures = fit_measures,
    std.lv = std.lv,
    miss = miss,
    est = est,
    check = check,
    save_out = save_out
  )
}

test_that(
  "Works with just fit_y, nagy = TRUE",
  {
    fit_y <- cfa_fit[1:2]
    cors <- sem.cor(BFIGritHope, fit_y)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), ncol(combn(names(fit_y), 2)))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), ncol(cors$cor_mat))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with just fit_y, nagy = FALSE, fit_save = TRUE",
  {
    cors <- sem.cor(BFIGritHope, cfa_fit, nagy = FALSE, fit_save = TRUE)
    expect_equal(length(cors), 5)
    expect_equal(length(cors$fit), ncol(combn(names(cfa_fit), 2)))
    expect_equal(ncol(cors$cor_mat), length(cfa_fit))
    expect_equal(nrow(cors$cor_mat), ncol(cors$cor_mat))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with fit_y and items, nagy = TRUE",
  {
    fit_y <- cfa_fit[1:2]
    items <- names(BFIGritHope)[grep("bfi.c\\d_1", names(BFIGritHope))]
    cors <- sem.cor(BFIGritHope, fit_y, items = items)
    expect_equal(length(cors), 4)
    expect_equal(
      length(cors$fit),
      ncol(combn(names(fit_y), 2)) + length(fit_y) * length(items)
    )
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_y) + length(items))
    expect_equal(ncol(cors$ci_lower), length(fit_y))
    expect_equal(nrow(cors$ci_lower), length(fit_y) + length(items))
    expect_equal(ncol(cors$ci_upper), length(fit_y))
    expect_equal(nrow(cors$ci_upper), length(fit_y) + length(items))
  }
)
test_that(
  "Works with fit_y and items, nagy = FALSE",
  {
    fit_y <- cfa_fit[1:2]
    items <- names(BFIGritHope)[grep("bfi.*1_1", names(BFIGritHope))]
    cors <- sem.cor(BFIGritHope, fit_y, items = items, nagy = FALSE)
    expect_equal(length(cors), 4)
    expect_equal(
      length(cors$fit),
      ncol(combn(names(fit_y), 2)) + length(fit_y) * length(items)
    )
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_y) + length(items))
    expect_equal(ncol(cors$ci_lower), length(fit_y))
    expect_equal(nrow(cors$ci_lower), length(fit_y) + length(items))
    expect_equal(ncol(cors$ci_upper), length(fit_y))
    expect_equal(nrow(cors$ci_upper), length(fit_y) + length(items))
  }
)
test_that(
  "Works with fit_y and fit_x, nagy = TRUE",
  {
    fit_y <- cfa_fit[1:2]
    fit_x <- cfa_fit[3:4]
    cors <- sem.cor(BFIGritHope, fit_y, fit_x)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), length(fit_y) * length(fit_x))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_x))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with fit_y and fit_x, nagy = FALSE",
  {
    fit_y <- cfa_fit[1:2]
    fit_x <- cfa_fit[3:4]
    cors <- sem.cor(BFIGritHope, fit_y, fit_x, nagy = FALSE)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), length(fit_y) * length(fit_x))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_x))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with fit_y, fit_x, and items, nagy = FALSE (TRUE not tested for time)",
  {
    fit_y <- cfa_fit[1:2]
    fit_x <- cfa_fit[3:4]
    items <- names(BFIGritHope)[grep("bfi.*1_1", names(BFIGritHope))]
    cors <- sem.cor(BFIGritHope, fit_y, fit_x, items = items, nagy = FALSE)
    expect_equal(length(cors), 4)
    # y * x + y * i = y(x + i)
    expect_equal(
      length(cors$fit), length(fit_y) * (length(fit_x) + length(items))
    )
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_x) + length(items))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with length 1 fit_y and items, nagy = TRUE",
  {
    items <- "bfi_e1_1"
    fit_y <- cfa_fit[1]
    cors <- sem.cor(BFIGritHope, fit_y, items = items)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), length(fit_y))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), ncol(cors$cor_mat))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with length 1 fit_y and items, nagy = FALSE",
  {
    items <- "bfi_e1_1"
    fit_y <- cfa_fit[1]
    cors <- sem.cor(BFIGritHope, fit_y, items = items, nagy = FALSE)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), length(fit_y))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), ncol(cors$cor_mat))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with length 1 fit_y and fit_x, nagy = TRUE",
  {
    fit_y <- cfa_fit[1]
    fit_x <- cfa_fit[3]
    cors <- sem.cor(BFIGritHope, fit_y, fit_x)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), length(fit_y) * length(fit_x))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_x))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Works with length 1 fit_y and fit_x, nagy = FALSE",
  {
    fit_y <- cfa_fit[1]
    fit_x <- cfa_fit[3]
    cors <- sem.cor(BFIGritHope, fit_y, fit_x, nagy = FALSE)
    expect_equal(length(cors), 4)
    expect_equal(length(cors$fit), length(fit_y) * length(fit_x))
    expect_equal(ncol(cors$cor_mat), length(fit_y))
    expect_equal(nrow(cors$cor_mat), length(fit_x))
    expect_equal(ncol(cors$ci_lower), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_lower), nrow(cors$cor_mat))
    expect_equal(ncol(cors$ci_upper), ncol(cors$cor_mat))
    expect_equal(nrow(cors$ci_upper), nrow(cors$cor_mat))
  }
)
test_that(
  "Element of 'fit_y' not a lavaan object",
  {
    expect_error(
      sem.cor(BFIGritHope, c(cfa_fit, "Hello")),
      "'fit_y' is not an object of class lavaan"
    )
  }
)
test_that(
  "Items cannot match a fit_y name",
  {
    items <- "bfi_e1_1"
    fit_y <- cfa_fit
    names(fit_y)[1] <- items
    expect_error(
      sem.cor(BFIGritHope, fit_y, items = "bfi_e1_1"),
      "'items' are included as a name of a fitted model in 'fit_y'"
    )
  }
)
test_that(
  "Element of 'fit_x' not a lavaan object",
  {
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[1:2], c(cfa_fit[3:4], "bfi_e1_1")),
      "'fit_x' is not an object of class lavaan"
    )
  }
)
test_that(
  "Items cannot match a fit_x name",
  {
    items <- "bfi_e1_1"
    fit_x <- cfa_fit[3:4]
    names(fit_x)[1] <- items
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[1:2], fit_x, items = "bfi_e1_1"),
      "'items' are included as a name of a fitted model in 'fit_x'"
    )
  }
)
test_that(
  "Item not in data",
  {
    expect_error(
      sem.cor(BFIGritHope, cfa_fit, items = "HelloWorld"),
      "'items' are not in 'data'"
    )
  }
)
test_that(
  "Length 1 fit_y with no fit_x or items",
  {
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[1]),
      "one measurement model and one item or two measurement models"
    )
  }
)
test_that(
  "Model with 2 factors in fit_y",
  {
    mod <- paste(
      "grit_c =~",
      paste(
        names(BFIGritHope)[grep("grit_c", names(BFIGritHope))], collapse = " + "
      ),
      "\ngrit_p =~",
      paste(
        names(BFIGritHope)[grep("grit_p", names(BFIGritHope))], collapse = " + "
      )
    )
    cfa_new <- sem(mod, BFIGritHope, std.lv = TRUE, missing = "ML")
    fit_y <- c(cfa_fit[3:4], list(new = cfa_new))
    expect_error(sem.cor(BFIGritHope, fit_y), "more than one latent variable")
  }
)
test_that(
  "Model with 2 factors in fit_x",
  {
    mod <- paste(
      "grit_c =~",
      paste(
        names(BFIGritHope)[grep("grit_c", names(BFIGritHope))], collapse = " + "
      ),
      "\ngrit_p =~",
      paste(
        names(BFIGritHope)[grep("grit_p", names(BFIGritHope))], collapse = " + "
      )
    )
    cfa_new <- sem(mod, BFIGritHope, std.lv = TRUE, missing = "ML")
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[3:4], list(new = cfa_new)),
      "more than one latent variable"
    )
  }
)
test_that(
  "Item from fit_y in items",
  {
    expect_error(
      sem.cor(BFIGritHope, cfa_fit, items = "grit_c_1"),
      "in both 'items' and contributes to the measurement of the"
    )
  }
)
test_that(
  "Item from fit_x in items (should work)",
  {
    expect_equal(
      length(
        sem.cor(
          BFIGritHope, cfa_fit[3:4], cfa_fit[1:2], items = "grit_c_1",
          nagy = FALSE
        )
      ),
      4
    )
  }
)
test_that(
  "All items shared in a different factor",
  {
    fit_x <- cfa.from.keys(
      list(grit = names(BFIGritHope)[grep("grit", names(BFIGritHope))]),
      BFIGritHope,
      fit_save = FALSE
    )$fit
    # Specifying fit_x, nagy = FALSE
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[1:2], fit_x, nagy = FALSE),
      "items in 'grit_c' are included in 'grit'"
    )
    # Only with fit_y, nagy = FALSE
    expect_error(
      sem.cor(BFIGritHope, c(cfa_fit[1:2], fit_x), nagy = FALSE),
      "items in 'grit_c' are included in 'grit'"
    )
    # Specifying fit_x, nagy = TRUE
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[1:2], fit_x, nagy = TRUE),
      "are in both the 'grit_c' and 'grit' factors"
    )
    # Only with fit_y, nagy = FALSE
    expect_error(
      sem.cor(BFIGritHope, c(cfa_fit[1:2], fit_x), nagy = TRUE),
      "are in both the 'grit_c' and 'grit' factors"
    )
  }
)
test_that(
  "Less than all items shared in different factors",
  {
    fit_x <- cfa.from.keys(
      list(
        grit_c = c(
          names(BFIGritHope)[grep("grit_c", names(BFIGritHope))], "grit_p_1"
        )
      ),
      BFIGritHope,
      fit_save = FALSE
    )$fit
    # Specifying fit_x, nagy = FALSE
    expect_warning(
      sem.cor(BFIGritHope, cfa_fit[2], fit_x, nagy = FALSE),
      "in both the 'grit_p' and 'grit_c' factors"
    )
    # Only with fit_y, nagy = FALSE
    expect_warning(
      sem.cor(BFIGritHope, c(cfa_fit[2], fit_x), nagy = FALSE),
      "in both the 'grit_p' and 'grit_c' factors"
    )
    # Specifying fit_x, nagy = FALSE
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[2], fit_x, nagy = TRUE),
      "are in both the 'grit_p' and 'grit_c' factors"
    )
    # Only with fit_y, nagy = FALSE
    expect_error(
      sem.cor(BFIGritHope, c(cfa_fit[2], fit_x), nagy = TRUE),
      "are in both the 'grit_p' and 'grit_c' factors"
    )
  }
)
test_that(
  "Non-positive definite matrix message",
  {
    b5 <- c("e", "c", "n", "o")
    b5f <- sapply(b5, function(x) paste0(x, 1:3))
    b5f_key <- sapply(
      b5f,
      function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))],
      simplify = FALSE
    )
    # Reduced set to minimise run time while still producing a nonPD matrix.
    b5f_key <- b5f_key[!grepl("(e|o)(1|2)|(c|n)1", names(b5f_key))]
    b5f_cfa <- cfa.from.keys(b5f_key, BFIGritHope, fit_save = FALSE)
    expect_warning(
      sem.cor(BFIGritHope, c(b5f_cfa$fit, cfa_fit), nagy = FALSE),
      "correlation matrix between 'fit_y' constructs has been adjusted"
    )
  }
)
test_that(
  "Works with 'item_loadings' set",
  {
    cors <- sem.cor(
      BFIGritHope, cfa_fit[1:2], items = paste0("bfi_e1_", 1:2),
      item_loadings = .8
    )
    parest <- lavaan::parameterestimates(cors$fit$grit_p.bfi_e1_1)
    expect_equal(
      parest$est[parest$lhs == "bfi_e1_1_l" & parest$rhs == "bfi_e1_1"], .8
    )
    cors <- sem.cor(
      BFIGritHope, cfa_fit[1:2], items = paste0("bfi_e1_", 1:2),
      item_loadings = c(.8, .7)
    )
    parest1 <- lavaan::parameterestimates(cors$fit$grit_p.bfi_e1_1)
    expect_equal(
      parest$est[parest$lhs == "bfi_e1_1_l" & parest$rhs == "bfi_e1_1"], .8
    )
    parest <- lavaan::parameterestimates(cors$fit$grit_p.bfi_e1_2)
    expect_equal(
      parest$est[parest$lhs == "bfi_e1_2_l" & parest$rhs == "bfi_e1_2"], .7
    )
  }
)
test_that(
  "Fitted objects rather than lists",
  {
    cors <- expect_no_error(
      sem.cor(BFIGritHope, cfa_fit[[1]], cfa_fit[[2]], nagy = FALSE)
    )
    expect_equal(colnames(cors$cor_mat), "grit_c")
    expect_equal(rownames(cors$cor_mat), "grit_p")
  }
)
test_that(
  "item_loadings specified and not equal to 1 or items.",
  {
    expect_error(
      sem.cor(
        BFIGritHope, cfa_fit,
        items = paste0("bfi_c1_", 1:4), item_loadings = 1:3 * .3
      ),
      "'item_loadings' must be either 'NULL', length 1, or length equal"
    )
  }
)
test_that(
  "Same CFA included in fit_y and fit_x",
  {
    expect_error(
      sem.cor(BFIGritHope, cfa_fit[1:2], cfa_fit[2:4], nagy = FALSE),
      "Have you included the same latent variable twice"
    )
  }
)

#' Runs ESEM based on CFA and EFA model outputs.
#'
#' `esem.from.keys` runs exploratory structural equation models (ESEM) in lavaan
#' where the exploratory factor analysis (EFA) factors predict confirmatory
#' factor analysis (CFA) factors and/or bifactor factors in separate models
#' for each CFA or bifactor model.
#' The function takes a fitted lavaan object from an EFA and lists of fitted
#' CFA and/or bifactor lavaan model objects as inputs so will typically
#' use outputs from [efa.from.keys()], and [cfa.from.keys()] or
#' [bifactor.from.keys()].
#'
#' @inheritParams sem.check
#' @param efa_fit A fitted lavaan object of an EFA model.
#' @param cfa_fit
#' A named list of fitted lavaan objects of CFA models.
#' Can be `NULL` if `bif_fit` is not `NULL`.
#' @param bif_fit
#' A named list of fitted lavaan objects of bifactor models.
#' Can be `NULL` if `cfa_fit` is not `NULL`.
#' @param name
#' A string indicating a subdirectory where model outputs will be saved when
#' `save_out = TRUE` and checked against when `check = TRUE`.
#' Defaults to "esem".
#' Irrelevant if both `save_out = FALSE` and `check = FALSE`.
#' The name should be unique for each set of models, or outputs from calls with
#' the same name will be overwritten.
#'
#' @return
#' Returns a list of length 4 (if `fit_save = FALSE`) or
#' 5 (if `fit_save = TRUE`).
#' The elements of the list are: a list of lavaan model output objects;
#' a list of parameter estimates from the models (standardized if `std = TRUE`);
#' if `fit_save = TRUE`, a matrix of fit measures for each model;
#' a list of regression beta parameters from each model;
#' and a dataframe of R-squared values from each model.
#'
#' @details
#' The function was designed to streamline running exploratory structural
#' equation models (ESEM) where EFA factors predict a series of latent variables
#' in separate models using Burt's (1976) 2-stage procedure to prevent
#' interpretational confounding. The function is designed to run analyses
#' equivalent to that of Bainbridge, Ludeke, and Smillie (2022).
#'
#' The function requires fitted lavaan objects as inputs in order to properly
#' employ the 2-stage procedure.
#' Using [efa.from.keys()], [cfa.from.keys()], and/or [bifactor.from.keys()]
#' should make this relatively straight-forward.
#'
#' Matching the philosophy of the package, the function is designed to run for
#' multiple models with a similar design. If you are using the function for a
#' single model, transform inputs into lists as appropriate.
#'
#' The model relies on [sem.check()] for the back-end of running the models.
#' This enables saving inputs and outputs from model runs
#' (with `save_out = TRUE`) and checking to see if anything has changed from
#' prior runs before running again (with `check = TRUE`).
#' The functionality was included for a number of very slow models or a lot of
#' faster models, such that time spent rerunning them would be onerous.
#' For further details on how this works, see the [sem.check()] function
#' documentation.
#'
#' Although the function will take any lavaan models that have successfully
#' produced standard lavaan outputs,
#' it will not complain if these have poor fit or other undesirable
#' characteristics (beyond warnings and errors produced by lavaan).
#' If this matters to you, it would be worth checking the outputs of upstream
#' functions prior to running `esem.from.mods()`.
#'
#' There is some disagreement about how to deal with interpretational
#' confounding.
#' The standard solution (other than ignoring it) is to create good fitting
#' measurement models first, then freely estimate the structural model with
#' checks to ensure adequate fit of the model and
#' that interpretational confounding is not an issue.
#' A problem with this is that changes to measurement models can require
#' entirely different factors (e.g., where original authors suggested a 4-factor
#' scale and subsequent analyses suggest a 2-factor scale) or items to be
#' removed.
#' How ought one to assess whether the scale measures the same thing(s) with
#' different factors or omitted items?
#' When a single scale is being assessed, these issues can be resolved;
#' however, when many scales are being assessed this solution is impractical.
#'
#' An alternative solution is to fix measurement model parameters in a model
#' estimating structural parameters.
#' This method means that less than ideal fit at the measurement level does not
#' propagate though the model as the measurement parameters are fixed.
#' It also solves interpretational confounding because the measurement of the
#' latent variables is fixed so their interpretation cannot change with the
#' inclusion of other variables in a model.
#' However, it is not a perfect solution because it underestimates uncertainty
#' in the measurement part of the structural model (e.g., Nagy et al., 2017),
#' which results in biased standard errors.
#'
#' A final option (among those noted here) was proposed by
#' Nagy and colleagues (2017).
#' It involves allowing item residuals to correlate with external variables
#' (or factors) and constrains those relationships such that the model is
#' identifiable.
#' If the sums of squares of correlations between all combinations of factors'
#' items and external factors are minimised,
#' measurement parameters in isolated measurement models are preserved in the
#' structural model without having to constrain them directly.
#' As a result, unbiased standard errors are preserved while simultaneously
#' eliminating interpretational confounding.
#' Unfortunately, estimating these models becomes increasingly slow with more
#' items and factors, such that it quickly becomes untenable,
#' especially for ESEM and when many models have to be run.
#'
#' As a result of these considerations, the 2-stage procedure of fixing
#' measurement parameters in the structural models was chosen for the current
#' function due to its simplicity and scalability.
#'
#' A further complication occurs for bifactor models.
#' Recall that bifactor models require orthogonal relationships between the
#' general and group factors.
#' When a factor is an outcome of a regression in a structural model,
#' it is not possible to include such a constraint because the instructions
#' normally used to do so will constrain residual variance instead.
#'
#' There are three solutions to this problem.
#' First, avoid using bifactor models.
#' This may be possible in some cases (e.g., if the scale can be modified into
#' a unidimentional scale), but frequently it is not.
#' Second, avoid using regressions in models with bifactor models and compute
#' regressions based on latent variable correlations.
#' (This can be done by, e.g., creating a correlation matrix from the factor
#' correlations and using the [psych::setCor()] function).
#' This method produces accurate point estimates in the regressions,
#' but standard errors are biased as the method cannot account for uncertainty
#' in the correlations.
#' Finally, if the 2-stage procedure is employed, then there is little room for
#' measurement models to change to allow factor correlations to change.
#' Therefore, the parameter can be relaxed and implied correlations between the
#' group and general factors should remain close to zero.
#' Given that the function already uses the 2-stage procedure,
#' this method is employed when bifactor models are used in `esem.from.mods()`.
#'
#' @seealso
#' [sem.check()], which this function uses for all the back-end;
#' [cfa.from.keys()], [efa.from.keys()], and [bifactor.from.keys()],
#' which are useful functions for creating inputs into `esem.from.mods()`; and
#' [lavaan::sem()], which is used to estimate the models.
#'
#' @references
#' Bainbridge, T. F., Ludeke, S. G., & Smillie, L. D. (2022).
#' Evaluating the Big Five as an organizing framework for commonly used
#' psychological trait scales.
#' Journal of Personality and Social Psychology, 122(4), 749-777.
#' https://doi.org/10.1037/pspp0000395.
#'
#' Burt, R. S. (1976).
#' Interpretational confounding of unobserved variables in Structural Equation
#' Models. Sociological Methods & Research, 5(1), 3-52.
#' https://doi.org/10.1177/004912417600500101.
#'
#' Eid, M., Geiser, C., Koch, T., & Heene, M. (2017).
#' Anomalous results in G-factor models: Explanations and alternatives.
#' Psychological Methods, 22(3), 541-562.
#' https://doi.org/10.1037/met0000083.
#'
#' Nagy, G., Brunner, M., Lüdtke, O., and Greiff, S. (2017).
#' Extension Procedures for Confirmatory Factor Analysis.
#' Journal of Experimental Education, 85(4).
#' https://doi.org/10.1080/00220973.2016.1260524.
#'
#' @importFrom lavaan summary
#' @export
#'
#' @examples
#' # Create CFA keys
#' keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#' keys <- sapply(
#'   keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#' )
#' # Create EFA keys
#' # Using only 3 factors and fewer items to save time for a simple example
#' # (This results in a less than ideal solution but it doesn't matter for an
#' # example)
#' keys_e0 <- paste0("bfi_", c("e", "a", "c"))
#' keys_e <- sapply(
#'   keys_e0,
#'   function(x) {
#'     names(BFIGritHope)[grep(paste0(x, "\\d_[1-2]"), names(BFIGritHope))]
#'   },
#'   simplify = FALSE
#' )
#' # Create fitted objects to use as inputs
#' cfa_fit <- cfa.from.keys(keys, BFIGritHope, fit_save = FALSE)
#' efa_fit <- efa.from.keys(keys_e, BFIGritHope, fit_save = FALSE)
#' # Run models
#' esem_fit <- esem.from.mods(
#'   efa_fit$fit$efa, cfa_fit$fit, data = BFIGritHope,
#'   fit_save = FALSE, check = FALSE
#' )
#' # Examine results
#' summary(esem_fit$fit$grit_c)  # Standard lavaan summary
#' esem_fit$r2                   # R-squareds
#' esem_fit$b                    # Betas

esem.from.mods <- function(
    data, efa_fit, cfa_fit = NULL, bif_fit = NULL,
    fit_save = FALSE, fit_measures = "all", miss = "ML", est = "default",
    name = "esem", check = FALSE, save_out = FALSE
) {
  if (is.null(cfa_fit) & is.null(bif_fit)) {
    stop("At least one of 'cfa_fit' and 'bif_fit' must be specified.")
  }
  if (!is.null(cfa_fit)) {
    # Single model instead of list.
    if (!is.list(cfa_fit) & inherits(cfa_fit, "lavaan")) {
      cfa_fit <- list(factor = cfa_fit)
    }
    if (sum(sapply(cfa_fit, function(x) !inherits(x, "lavaan"))) > 0) {
      paste0(
        names(cfa_fit)[sapply(cfa_fit, function(x) !inherits(x, "lavaan"))],
        stop("The above elements of 'cfa_fit' are not objects of type lavaan.")
      )
    }
  }
  if (!is.null(bif_fit)) {
    # Single model instead of list.
    if (!is.list(bif_fit) & inherits(bif_fit, "lavaan")) {
      bif_fit <- list(bifactor = bif_fit)
    }
    if (sum(sapply(bif_fit, function(x) !inherits(x, "lavaan"))) > 0) {
      paste0(
        names(bif_fit)[sapply(bif_fit, function(x) !inherits(x, "lavaan"))],
        stop("The above elements of 'bif_fit' are not objects of type lavaan.")
      )
    }
  }
  if (is.null(efa_fit)) {
    stop(
      paste(
        "'efa_fit' is NULL.",
        "'efa_fit' must be a fitted lavaan object of an EFA model."
      )
    )
  }
  if (!inherits(efa_fit, "lavaan")) {
    stop("'efa_fit' is not an object of type lavaan.")
  }
  if (!is.null(cfa_fit)) {
    cfa_par <- sapply(cfa_fit, parameterEstimates, simplify = FALSE)
    # Extract factor names
    cfa_names <- sapply(
      cfa_par,
      function(x) {
        x1 <- unique(x$lhs[x$op == "=~"])
        if (length(x1) > 1) {
          stop(
            paste(
              "A CFA containing more than one latent variable has been found.",
              "Currently, the function only supports CFAs included in separate",
              "models.",
              "Please either use 'bif_fit' and a model supported there,",
              "or separate the CFAs into separate measurement models.",
              "The offending factors are:\n",
              "    ",
              paste(x1, collapse = "\n    ")
            )
          )
        }
        return(x1)
      }
    )
    names(cfa_fit) <- names(cfa_par) <- cfa_names
    cfa_keys <- sapply(cfa_par, function(x) x$rhs[x$op == "=~"])
    names(cfa_keys) <- cfa_names
    if (sum(table(names(cfa_keys)) > 1) > 0) {
      stop(
        paste(
          "At least two different models in 'cfa_fit' have factors with the",
          "same name.",
          "Please ensure that all factor names are unique."
        )
      )
    }
  }
  if (!is.null(bif_fit)) {
    bif_par <- sapply(bif_fit, parameterEstimates, simplify = FALSE)
    bif_keys <- sapply(bif_par, function(x) unique(x$rhs[x$op == "=~"]))
    bif_names <- mapply(
      x = bif_par, y = bif_keys,
      FUN = function(x, y) {
        tmp <- table(x$lhs[x$op == "=~" & x$rhs %in% y])
        names(tmp)[tmp == max(tmp)]
      }
    )
    names(bif_par) <- bif_names
    if (!is.null(names(bif_fit))) {
      if (sum(names(bif_fit) != bif_names) > 0) {
        warning(
          paste(
            "The names of 'bif_fit' do not match the general factor names.",
            "Names of returned objects are based on factor names",
            "so they will not match the names of 'bif_fit'."
          )
        )
      }
    }
    names(bif_keys) <- bif_names
    if (sum(table(names(bif_keys)) > 1) > 0) {
      stop(
        paste(
          "At least two different models in 'bif_fit' have general factors",
          "with the same name.",
          "Please ensure that all factor names are unique."
        )
      )
    }
  }
  if (!is.null(cfa_fit) & !is.null(bif_fit)) {
    if (sum(names(cfa_keys) %in% names(bif_keys)) > 0) {
      stop(
        paste(
          "The following models in 'cfa_fit' have identically named",
          "factor(s) in 'bif_fit':\n    ",
          paste(
            names(cfa_fit)[names(cfa_fit) %in% names(bif_fit)], collapse = "\n"
          ),
          "\n\n  Please ensure that CFA factors and bifactor general factors",
          "have unique names."
        )
      )
    }
  }
  efa_par <- parameterEstimates(efa_fit)
  efa_par1 <-
    efa_par[efa_par$op %in% c("=~", "~~"), c("lhs", "op", "rhs", "est")]
  efa_mod <- paste(
    efa_par1$lhs, efa_par1$op, efa_par1$est, "*", efa_par1$rhs, collapse = "\n"
  )
  efa_facs <- unique(efa_par$lhs[efa_par$op == "=~"])
  efa_items <- unique(efa_par$rhs[efa_par$op == "=~"])
  efa_keys0 <- as.data.frame(
    t(
      sapply(
        efa_items,
        function(x) {
          tmp <- efa_par[efa_par$op == "=~" & efa_par$rhs == x, ]
          unlist(tmp[abs(tmp$est) == max(abs(tmp$est)), c("lhs", "rhs")])
        }
      )
    )
  )
  efa_keys <- sapply(
    efa_facs, function(x) efa_keys0$rhs[efa_keys0$lhs == x], simplify = FALSE
  )
  if (!is.null(cfa_fit)) {
    mods_cfa <- sapply(
      names(cfa_keys),
      function(x) {
        x0 <- cfa_par[[x]]
        x1 <- x0[x0$op %in% c("=~", "~~"), c("lhs", "op", "rhs", "est")]
        x2 <- x1[!(x == x1$lhs & x == x1$rhs), ]
        x3 <- paste0(paste(x2$lhs, x2$op, x2$est, "*", x2$rhs, collapse = "\n"))
        paste0(
          c(
            efa_mod,
            x3,
            paste(x, "~", paste0(names(efa_keys), collapse = " + "))
          ),
          collapse = "\n"
        )
      },
      simplify = FALSE
    )
  }
  if (!is.null(bif_fit)) {
    mods_bif <- sapply(
      names(bif_keys),
      function(x) {
        x0 <- bif_par[[x]]
        x1 <- x0[x0$op %in% c("=~", "~~"), c("lhs", "op", "rhs", "est")]
        x2 <- x1[!(x == x1$lhs & x1$op == "~~"), ]
        x3 <- paste0(paste(x2$lhs, x2$op, x2$est, "*", x2$rhs, collapse = "\n"))
        groupf <- unique(x0$lhs[x0$op == "=~" & !(x0$lhs %in% x)])
        paste0(
          c(
            efa_mod, x3,
            # Regression
            paste(x, "~", paste0(names(efa_keys), collapse = " + ")),
            # Correlations with group factors
            paste(
              paste(groupf, collapse = " + "),
              "~~",
              paste0(names(efa_keys), collapse = " + "),
              "+",
              x
            )
          ),
          collapse = "\n"
        )
      },
      simplify = FALSE
    )
  }
  if (!is.null(bif_fit) & !is.null(cfa_fit)) {
    mods <- c(mods_cfa, mods_bif)
    keys_s <- c(cfa_keys, bif_keys)
  } else if (!is.null(cfa_fit)) {
    mods <- mods_cfa
    keys_s <- cfa_keys
  } else {
    mods <- mods_bif
    keys_s <- bif_keys
  }
  mod_out <- sem.check(
    mods,
    data,
    name = name,
    keys_s = keys_s,
    keys_e = efa_keys,
    std = TRUE,  # For r2 calcs.
    fit_save = fit_save,
    fit_measures = fit_measures,
    miss = miss,
    est = est,
    std.lv = FALSE,  # Params are set from measurement models.
    check = check,
    save_out = save_out
  )
  r2 <- do.call(
    rbind,
    mapply(
      function(x, xn) {
        tmp <- x[x$op == "~~" & x$lhs == xn & x$rhs == xn, ]
        c(
          R2 = 1 - tmp$est.std,
          se = tmp$se,
          ci.lower = 1 - tmp$ci.upper,
          ci.upper = 1 - tmp$ci.lower
        )
      },
      x = mod_out$par_std, xn = names(mod_out$par_std), SIMPLIFY = FALSE
    )
  )
  b <- lapply(
    mod_out$par_std,
    function(x, xn) {
      tmp <- x[x$op == "~", ]
      tmp[-(1:2)]
    }
  )
  if (fit_save) {
    return(
      list(
        fit = mod_out$fit,
        par_std = mod_out$par_std,
        fit_measures = mod_out$fit_measures,
        b = b,
        r2 = r2
      )
    )
  } else {
    return(
      list(
        fit = mod_out$fit,
        par_std = mod_out$par_std,
        b = b,
        r2 = r2
      )
    )
  }
}

#' Runs bifactor models for multiple scales based on keys lists.
#'
#' `bifactor.from.keys` runs a series of bifactor model from three keys lists---
#' one for items on general factor; one for items on group factors;
#' and one for group factors on general factors.
#'
#' @inheritParams sem.check
#' @param keys_g
#' A named list of items in general factors.
#' Names must be the names of the general factors.
#' Each list element must be a vector of items that load on the general factors.
#' Must be the same length as `keys_b`.
#' @param keys_b
#' A named list of group factors in general factors.
#' Names must be the names of the general factors.
#' Each list element must be a vector of group factors that load on the general
#' factors.
#' Must be the same length as `keys_g`.
#' @param keys
#' A named list of items in group factors.
#' Names must be the names of the group factors.
#' Each list element must be a vector of items that load on the group factors.
#' Need not be the same length as `keys_g` and `keys_b`.
#' @param data
#' A dataframe or object coercible to a dataframe.
#' Data must include all observed variables in any of the keys.
#' @param name
#' A string indicating a subdirectory where model outputs will be saved when
#' `save_out = TRUE` and checked against when `check = TRUE`.
#' Defaults to 'bifactor'.
#' Irrelevant if both `save_out = FALSE` and `check = FALSE`.
#' The name should be unique for each set of models, or outputs from calls with
#' the same name will be overwritten.
#' @param std.lv
#' Logical.
#' Sets the `std.lv` parameter, as per lavaan (see [lavaan::lavOptions()]).
#' `TRUE` indicates that factor variances should be fixed to 1.
#' `FALSE` indicates that loadings of the first items of factors should be fixed
#' to 1.
#' Defaults to `TRUE`.
#'
#' @return
#' Returns a list of length 2 (if `fit_save = FALSE`) or
#' 3 (if `fit_save = TRUE`).
#' The elements of the list are: a list of lavaan model output objects;
#' a list of parameter estimates from the models (standardized if `std = TRUE`);
#' and, if `fit_save = TRUE`, a matrix of fit measures for each model.
#'
#' @details
#' The function is designed to streamline running measurement models for all
#' scales in a sample and to input model outputs into downstream functions.
#' Keys list must be named appropriately;
#' that is, `keys_g` and `keys_b` names must be the general factor names,
#' and `keys` must be the group factor names.
#'
#' The model relies on [sem.check()] for the back-end of running the models.
#' This enables saving inputs and outputs from model runs
#' (with `save_out = TRUE`) and checking to see if anything has changed from
#' prior runs before running again (with `check = TRUE`).
#' The functionality was included for a number of very slow models or a lot of
#' faster models, such that time spent rerunning them would be onerous.
#' For further details on how this works, see the [sem.check()] function
#' documentation.
#'
#' Please be careful with bifactor models.
#' In simulation studies, they can fit better than the models that generated the
#' data in the presence of unspecified complexity
#' (which is usually the case; Murray & Johnson, 2013).
#' They can also produce negative residual variances.
#' lavaan will warn you of this and you should deal with it before proceeding.
#' Items can also load in atheoretical ways on the general or group factors.
#' This could be some items loading negatively instead of positively or
#' vice versa or one or more items having insignificant loadings on a factor
#' that they should theoretically load on.
#'
#' These problems can often be solved by omitting a factor.
#' When items from one group factor dominate the general factor,
#' the best solution might be to remove the general factor and treat the group
#' factors as separate constructs.
#' Perhaps more frequently, these problems might be solved by omitting one
#' (or more) group factors (i.e., an S-1 model; Eid et al., 2017).
#' This can either be done a priori
#' (perhaps based on which one should theoretically be most 'central' to the
#' general factor)
#' or after examining the fully specified bifactor model and dropping the worst
#' performing factor (see the example).
#' When more than one group factor is poor, I am not aware of any specific
#' advice on which to remove, so use your judgment if previous research has not
#' got any advice for your case.
#'
#' @references
#' Eid, M., Geiser, C., Koch, T., & Heene, M. (2017).
#' Anomalous results in G-factor models: Explanations and alternatives.
#' Psychological Methods, 22(3), 541-562.
#' https://doi.org/10.1037/met0000083.
#'
#' Murray, A. L. & Johnson, W. (2013).
#' The limitations of model fit in comparing the bi-factor versus higher-order
#' models of human cognitive ability structure. Intelligence, 41(5), 407-422.
#' https://doi.org/10.1016/j.intell.2013.06.004.
#'
#' @seealso
#' [sem.check()], which `bifactor.from.keys()` uses for all the back-end, and
#' [lavaan::sem()], which is used to estimate the models.
#'
#' @importFrom lavaan summary
#' @export
#'
#' @examples
#' # Create keys
#' keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#' keys <- sapply(
#'   keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#' )
#' keys_g0 <- c("grit", "hope")
#' keys_g <- sapply(
#'   keys_g0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#' )
#' keys_b1 <- sapply(
#'   keys_g0, function(x) keys0[grep(x, keys0)], simplify = FALSE
#' )
#' # Including all factors produces a negative residual variance for hope.
#' \donttest{bif_fit0 <- bifactor.from.keys(
#'   keys_g, keys_b1, keys, BFIGritHope, check = FALSE, fit_save = TRUE
#' )
#' summary(bif_fit0$fit$hope)}
#' # Fix negative residual variance by removing the first group factor.
#' keys_b <- keys_b1
#' keys_b$hope <- keys_b1$hope[-1]
#' bif_fit <- bifactor.from.keys(
#'   keys_g, keys_b, keys, BFIGritHope, check = FALSE, fit_save = TRUE
#' )
#' # Examine some results
#' summary(bif_fit$fit$grit)                  # Standard lavaan summary
#' bif_fit$fit_measures[, c("cfi", "rmsea")]  # Fit measures

bifactor.from.keys <- function(
  keys_g, keys_b, keys, data, fit_save = TRUE, fit_measures = "all",
  std.lv = TRUE, miss = "ML", est = "default",
  name = "bifactor", check = FALSE, save_out = FALSE
) {
  if (!is.list(keys_g)) {
    stop("'keys_g' is not a list.")
  }
  if (!is.list(keys_b)) {
    stop("'keys_b' is not a list.")
  }
  if (!is.list(keys)) {
    stop("'keys' is not a list.")
  }
  if (length(keys_g) != length(keys_b)) {
    stop(
      paste(
        "'keys_g' is not the same length as 'keys_b'.",
        "Check that you have not mixed up 'keys' with 'keys_g' or 'keys_b',",
        "or otherwise misspecified one of these keys lists."
      )
    )
  }
  if (sum(names(keys_g) != names(keys_b)) > 0) {
    stop(
      paste(
        "Names of 'keys_g' do not match names of 'keys_b'.",
        "Check keys are correctly specified and that they have not been mixed",
        "up with 'keys'."
      )
    )
  }
  sapply(
    keys_b,
    function(x) {
      if (sum(!x %in% names(keys)) > 0) {
        grps <- x[!x %in% names(keys)]
        stop(
          paste0(
            "The following group factor(s) in 'keys_b' are not in 'keys':",
            "\n    ",
            paste(grps, collapse = "\n    "),
            "\n\nIf these are items, not group factors, ",
            "and you are using bifactor.from.keys, ",
            "check that keys_b only contains group factor names."
          )
        )
      }
    }
  )
  items <- with_options(
    list(warn = 1),
    mapply(
      function(x, xn, y) {
        sapply(
          y,
          function(z) {
            if (sum(!keys[[z]] %in% x) == length(keys[[z]])) {
              stop(
                paste0(
                  "All the items in the '",
                  z,
                  "' group factor are not in the '",
                  xn,
                  "' general factor."
                )
              )
            }
          }
        )
        if (sum(!unlist(keys[y]) %in% x) > 0) {
          warning(
            paste(
              "The following item(s) are in a group factor but not in the",
              "general factor:\n   ",
              paste(unlist(keys[y])[!unlist(keys[y]) %in% x], collapse = "\n    ")
            )
          )
          c(x, unlist(keys[y])[!unlist(keys[y]) %in% x])
        } else {
          x
        }
      },
      x = keys_g, xn = names(keys_g), y = keys_b
    )
  )
  mods <- mapply(
    function(x, y, z) {
      tmp <- paste(z, "=~", paste(x, collapse = " + "))
      paste0(
        c(
          tmp,
          mapply(
            function(i, ni) paste(ni, "=~", paste0(i, collapse = " + ")),
            i = keys[y], ni = names(keys[y]), SIMPLIFY = FALSE
          )
        ),
        collapse = "\n"
      )
    },
    x = keys_g, y = keys_b, z = names(keys_g), SIMPLIFY = FALSE
  )
  sem.check(
    mods,
    data,
    name = name,
    keys_s = items,
    keys_e = NULL,
    std = FALSE,  # For use in 2-stage procedure, must use non-standardised.
    fit_save = fit_save,
    fit_measures = fit_measures,
    std.lv = std.lv,
    miss = miss,
    est = est,
    orthogonal = TRUE,  # Must be TRUE for bifactor models.
    check = check,
    save_out = save_out
  )
}

test_that(
  "Test normal behaviour with 'fit_save = FALSE'",
  {
    bif_fit <- bifactor.from.keys(
      keys_g, keys_b, keys, BFIGritHope, fit_save = FALSE
    )
    expect_equal(length(bif_fit), 2)
    expect_equal(length(bif_fit$fit), length(keys_g))
    expect_equal(length(bif_fit$par), length(keys_g))
    expect_equal(sum(sapply(bif_fit$fit, function(x) class(x) != "lavaan")), 0)
  }
)
test_that(
  "Test normal behaviour with 'fit_save = TRUE'",
  {
    bif_fit <- bifactor.from.keys(
      keys_g, keys_b, keys, BFIGritHope, fit_save = TRUE
    )
    expect_equal(length(bif_fit), 3)
    expect_equal(length(bif_fit$fit), length(keys_g))
    expect_equal(length(bif_fit$par), length(keys_g))
    expect_equal(sum(sapply(bif_fit$fit, function(x) class(x) != "lavaan")), 0)
  }
)
test_that(
  "Mistakes with keys specification: Include CFA keys as keys",
  {
    expect_error(
      bifactor.from.keys(
        keys = keys_g, data = BFIGritHope, fit_save = FALSE
      ),
      'argument "keys_g" is missing'
    )
  }
)
test_that(
  "Mistakes with keys specification: Include CFA keys as keys_g",
  {
    expect_error(
      bifactor.from.keys(keys_g, data = BFIGritHope, fit_save = FALSE),
      'argument "keys_b" is missing'
    )
  }
)
test_that(
  "Mistakes with keys specification: Include data as keys_b",
  {
    expect_error(
      bifactor.from.keys(keys_g, BFIGritHope, fit_save = FALSE),
      '"keys" is missing'
    )
  }
)
test_that(
  "Mistakes with keys specification: Swap keys_g and keys_b",
  {
    expect_error(
      bifactor.from.keys(
        keys_b, keys_g, keys, BFIGritHope, fit_save = FALSE
      ),
      "group factor\\(s\\) in 'keys_b' are not in 'keys'"
    )
  }
)
test_that(
  "Mistakes with keys specification: Swap keys_g and keys",
  {
    expect_error(
      bifactor.from.keys(
        keys, keys_b, keys_g, BFIGritHope, fit_save = FALSE
      ),
      "'keys_g' is not the same length as 'keys_b'"
    )
  }
)
test_that(
  "Different order of 'keys_b' and 'keys_g' elements",
  {
    keys_b_swap <- keys_b[2:1]
    expect_error(
      bifactor.from.keys(
        keys_g, keys_b_swap, keys, BFIGritHope, fit_save = FALSE
      ),
      "Names of 'keys_g' do not match names of 'keys_b'"
    )
  }
)
test_that(
  "Mistakes in keys lists",
  {
    keys_mistake <- keys
    keys_mistake$grit_c[1] <- "mistake"
    keys_g_mistake <- keys_g
    keys_g_mistake$grit[1] <- "mistake"
    keys_b_mistake <- keys_b
    keys_b_mistake$grit[1] <- "mistake"
    expect_error(
      suppressWarnings(
        bifactor.from.keys(
          keys_g, keys_b, keys_mistake, BFIGritHope, fit_save = FALSE
        )
      ),
      "items are in 'keys_s' but they are not in 'data'"
    )
    expect_error(
      suppressWarnings(
        bifactor.from.keys(
          keys_g_mistake, keys_b, keys, BFIGritHope, fit_save = FALSE
        )
      ),
      "items are in 'keys_s' but they are not in 'data'"
    )
    expect_error(
      bifactor.from.keys(
        keys_g, keys_b_mistake, keys, BFIGritHope, fit_save = FALSE
      ),
      "group factor\\(s\\) in 'keys_b' are not in 'keys'"
    )
  }
)
test_that(
  "Non-list keys list",
  {
    expect_error(
      bifactor.from.keys(
        keys_g, keys_b, keys$grit_c, BFIGritHope, fit_save = FALSE
      ),
      "'keys' is not a list"
    )
    expect_error(
      bifactor.from.keys(
        keys_g$grit, keys_b, keys, BFIGritHope, fit_save = FALSE
      ),
      "'keys_g' is not a list"
    )
    expect_error(
      bifactor.from.keys(
        keys_g, keys_b$grit, keys, BFIGritHope, fit_save = FALSE
      ),
      "'keys_b' is not a list"
    )
  }
)
test_that(
  "No key for group factor",
  {
    keys_missing <- keys[-1]
    expect_error(
      bifactor.from.keys(
        keys_g, keys_b, keys_missing, BFIGritHope, fit_save = FALSE
      ),
      "group factor\\(s\\) in 'keys_b' are not in 'keys'"
    )
  }
)
test_that(
  "Group factors from a different scale",
  {
    keys_b_wrong <- keys_b
    keys_b_wrong$grit <- keys_b$hope
    expect_error(
      bifactor.from.keys(
        keys_g, keys_b_wrong, keys, BFIGritHope, fit_save = FALSE
      ),
      "items in the.*group factor are not in the.*general factor"
    )
  }
)
test_that(
  "Item in a group factor but not in the general factor",
  {
    keys_g2 <- keys_g
    keys_g2$grit <- keys_g2$grit[-1]
    expect_warning(
      bifactor.from.keys(
        keys_g2, keys_b, keys, BFIGritHope, fit_save = FALSE
      ),
      "item\\(s\\) are in a group factor but not in the general factor"
    )
  }
)

#' BFI, Grit, and Hope data from Bainbridge, Ludeke, & Smillie (2022),
#' Study 2b.
#'
#' The subset of the data reported in the paper include only the BFI-2, Grit,
#' and hope.
#'
#' @format ## `BFIGritHopeImp`
#' A data frame with 388 rows and 95 columns:
#' \describe{
#'   \item{bfi_e1_1}{BFI-2 Extraversion: Sociability, Item 1}
#'   \item{bfi_e1_2}{BFI-2 Extraversion: Sociability, Item 2}
#'   \item{bfi_e1_3}{BFI-2 Extraversion: Sociability, Item 3}
#'   \item{bfi_e1_4}{BFI-2 Extraversion: Sociability, Item 4}
#'   \item{bfi_e2_1}{BFI-2 Extraversion: Assertiveness, Item 1}
#'   \item{bfi_e2_2}{BFI-2 Extraversion: Assertiveness, Item 2}
#'   \item{bfi_e2_3}{BFI-2 Extraversion: Assertiveness, Item 3}
#'   \item{bfi_e2_4}{BFI-2 Extraversion: Assertiveness, Item 4}
#'   \item{bfi_e3_1}{BFI-2 Extraversion: Energy Level, Item 1}
#'   \item{bfi_e3_2}{BFI-2 Extraversion: Energy Level, Item 2}
#'   \item{bfi_e3_3}{BFI-2 Extraversion: Energy Level, Item 3}
#'   \item{bfi_e3_4}{BFI-2 Extraversion: Energy Level, Item 4}
#'   \item{bfi_a1_1}{BFI-2 Agreeableness: Compassion, Item 1}
#'   \item{bfi_a1_2}{BFI-2 Agreeableness: Compassion, Item 2}
#'   \item{bfi_a1_3}{BFI-2 Agreeableness: Compassion, Item 3}
#'   \item{bfi_a1_4}{BFI-2 Agreeableness: Compassion, Item 4}
#'   \item{bfi_a2_1}{BFI-2 Agreeableness: Respectfulness, Item 1}
#'   \item{bfi_a2_2}{BFI-2 Agreeableness: Respectfulness, Item 2}
#'   \item{bfi_a2_3}{BFI-2 Agreeableness: Respectfulness, Item 3}
#'   \item{bfi_a2_4}{BFI-2 Agreeableness: Respectfulness, Item 4}
#'   \item{bfi_a3_1}{BFI-2 Agreeableness: Trust, Item 1}
#'   \item{bfi_a3_2}{BFI-2 Agreeableness: Trust, Item 2}
#'   \item{bfi_a3_3}{BFI-2 Agreeableness: Trust, Item 3}
#'   \item{bfi_a3_4}{BFI-2 Agreeableness: Trust, Item 4}
#'   \item{bfi_c1_1}{BFI-2 Conscientiousness: Organization, Item 1}
#'   \item{bfi_c1_2}{BFI-2 Conscientiousness: Organization, Item 2}
#'   \item{bfi_c1_3}{BFI-2 Conscientiousness: Organization, Item 3}
#'   \item{bfi_c1_4}{BFI-2 Conscientiousness: Organization, Item 4}
#'   \item{bfi_c2_1}{BFI-2 Conscientiousness: Productiveness, Item 1}
#'   \item{bfi_c2_2}{BFI-2 Conscientiousness: Productiveness, Item 2}
#'   \item{bfi_c2_3}{BFI-2 Conscientiousness: Productiveness, Item 3}
#'   \item{bfi_c2_4}{BFI-2 Conscientiousness: Productiveness, Item 4}
#'   \item{bfi_c3_1}{BFI-2 Conscientiousness: Responsibility, Item 1}
#'   \item{bfi_c3_2}{BFI-2 Conscientiousness: Responsibility, Item 2}
#'   \item{bfi_c3_3}{BFI-2 Conscientiousness: Responsibility, Item 3}
#'   \item{bfi_c3_4}{BFI-2 Conscientiousness: Responsibility, Item 4}
#'   \item{bfi_n1_1}{BFI-2 Negative Emotionality: Anxiety, Item 1}
#'   \item{bfi_n1_2}{BFI-2 Negative Emotionality: Anxiety, Item 2}
#'   \item{bfi_n1_3}{BFI-2 Negative Emotionality: Anxiety, Item 3}
#'   \item{bfi_n1_4}{BFI-2 Negative Emotionality: Anxiety, Item 4}
#'   \item{bfi_n2_1}{BFI-2 Negative Emotionality: Depression, Item 1}
#'   \item{bfi_n2_2}{BFI-2 Negative Emotionality: Depression, Item 2}
#'   \item{bfi_n2_3}{BFI-2 Negative Emotionality: Depression, Item 3}
#'   \item{bfi_n2_4}{BFI-2 Negative Emotionality: Depression, Item 4}
#'   \item{bfi_n3_1}{BFI-2 Negative Emotionality: Emotional Volatility, Item 1}
#'   \item{bfi_n3_2}{BFI-2 Negative Emotionality: Emotional Volatility, Item 2}
#'   \item{bfi_n3_3}{BFI-2 Negative Emotionality: Emotional Volatility, Item 3}
#'   \item{bfi_n3_4}{BFI-2 Negative Emotionality: Emotional Volatility, Item 4}
#'   \item{bfi_o1_1}{BFI-2 Open-Mindedness: Intellectual Curiosity, Item 1}
#'   \item{bfi_o1_2}{BFI-2 Open-Mindedness: Intellectual Curiosity, Item 2}
#'   \item{bfi_o1_3}{BFI-2 Open-Mindedness: Intellectual Curiosity, Item 3}
#'   \item{bfi_o1_4}{BFI-2 Open-Mindedness: Intellectual Curiosity, Item 4}
#'   \item{bfi_o2_1}{BFI-2 Open-Mindedness: Aesthetic Sensitivity, Item 1}
#'   \item{bfi_o2_2}{BFI-2 Open-Mindedness: Aesthetic Sensitivity, Item 2}
#'   \item{bfi_o2_3}{BFI-2 Open-Mindedness: Aesthetic Sensitivity, Item 3}
#'   \item{bfi_o2_4}{BFI-2 Open-Mindedness: Aesthetic Sensitivity, Item 4}
#'   \item{bfi_o3_1}{BFI-2 Open-Mindedness: Creative Imagination, Item 1}
#'   \item{bfi_o3_2}{BFI-2 Open-Mindedness: Creative Imagination, Item 2}
#'   \item{bfi_o3_3}{BFI-2 Open-Mindedness: Creative Imagination, Item 3}
#'   \item{bfi_o3_4}{BFI-2 Open-Mindedness: Creative Imagination, Item 4}
#'   \item{grit_c_1}{Grit Consistency of Interests, Item 1}
#'   \item{grit_c_2}{Grit Consistency of Interests, Item 2}
#'   \item{grit_c_3}{Grit Consistency of Interests, Item 3}
#'   \item{grit_c_4}{Grit Consistency of Interests, Item 4}
#'   \item{grit_c_5}{Grit Consistency of Interests, Item 5}
#'   \item{grit_c_6}{Grit Consistency of Interests, Item 6}
#'   \item{grit_p_1}{Grit Persistence, Item 1}
#'   \item{grit_p_2}{Grit Persistence, Item 2}
#'   \item{grit_p_3}{Grit Persistence, Item 3}
#'   \item{grit_p_4}{Grit Persistence, Item 4}
#'   \item{grit_p_5}{Grit Persistence, Item 5}
#'   \item{grit_p_6}{Grit Persistence, Item 6}
#'   \item{hope_a_1}{Hope Agency, Item 1}
#'   \item{hope_a_2}{Hope Agency, Item 2}
#'   \item{hope_a_3}{Hope Agency, Item 3}
#'   \item{hope_a_4}{Hope Agency, Item 4}
#'   \item{hope_p_1}{Hope Pathways, Item 1}
#'   \item{hope_p_2}{Hope Pathways, Item 2}
#'   \item{hope_p_3}{Hope Pathways, Item 3}
#'   \item{hope_p_4}{Hope Pathways, Item 4}
#' }
#' @source <https://osf.io/f9hmg/files/eu8p9>
"BFIGritHope"

# Create CFA keys
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)

# Create bifactor keys
keys_g0 <- c("grit", "hope")
keys_g <- sapply(
  keys_g0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
keys_b1 <- sapply(keys_g0, function(x) keys0[grep(x, keys0)], simplify = FALSE)
keys_b <- keys_b1
# Fix negative residual variance
# (this is a lavaan warning, no need to test here)
keys_b$hope <- keys_b1$hope[-1]

# Create EFA keys
# Using only 3 factors to save time on checks
keys_e0 <- paste0("bfi_", c("e", "a", "c"))

# Using less than all items to save time on checks
# (This results in a less than ideal solution but it doesn't matter for the
# tests)
keys_e <- sapply(
  keys_e0,
  function(x) {
    names(BFIGritHope)[grep(paste0(x, "\\d_[1-2]"), names(BFIGritHope))]
  },
  simplify = FALSE
)

# Create fit objects to using in test-esem.from.mods
cfa_fit <- cfa.from.keys(keys, BFIGritHope, fit_save = FALSE)$fit
efa_fit <-
  efa.from.keys(keys_e, BFIGritHope, fit_save = FALSE)$fit$efa
bif_fit <- bifactor.from.keys(
  keys_g, keys_b, keys, BFIGritHope, fit_save = FALSE
)$fit

#' Creates a latent variable correlation matrix from fitted lavaan measurement
#' models
#'
#' sem.cor takes CFA outputs and produces a correlation matrix between latent
#' variables and/or single items.
#'
#' @inheritParams sem.check
#' @param fit_y A named list of CFA fitted objects.
#' @param fit_x
#' A named list of CFA fitted objects to be correlated with fit_y variables
#' or 'NULL'.
#' @param items
#' A vector of single-item variables to correlate with 'fit_y' latent variables.
#' Must not include any items contributing to the measurement of a `fit_y`
#' latent variable.
#' @param item_loadings
#' When single items are specified, items are included in models with single
#' item latent variables. `item_loadings` sets the loading of the item on the
#' factor.
#' It can be a single number to set all loadings equal or a vector of
#' length equal to the length of items.
#' Defaults allows the value to be free (which assumes perfect reliability).
#' Irrelevant if `items = NULL`.
#' @param name
#' A string indicating a subdirectory where model outputs will be saved when
#' `save_out = TRUE` and checked against when `check = TRUE`.
#' Defaults to "cors".
#' Irrelevant if both `save_out = FALSE` and `check = FALSE`.
#' The name should be unique for each set of models, or outputs from calls with
#' the same name will be overwritten.
#' @param nagy
#' Logical. Indicates whether to use Nagy and colleagues' (2017) extension
#' procedure instead of Burt's (1976) 2-stage procedure.
#'
#' @return
#' Returns a list of length 4.
#' The first element is the fitted lavaan models ('fit').
#' The second element is the correlation matrix ('cor_mat').
#' The third and fourth elements are the upper and lower confidence intervals
#' from the models ('ci_lower' and 'ci_upper', respectively).
#'
#' @details
#' The function computes correlations between latent variables from fitted CFA
#' models, including either all latent variables, or between two sets of latent
#' variables.
#' To compute a full correlation matrix include all CFA models in a list as
#' `fit_y`, alternatively,
#' to correlate a set of dependant variables with a set of independent
#' variables,
#' set `fit_y` as a list of fitted CFA models of dependant variables and
#' set `fit_x` as a list of fitted CFA models of independent variables.
#' Correlations between the `fit_y` latent variables and single items can also
#' be optionally included by setting `items` as a vector of item names.
#' In this case, items will be treated as single item latent variables with
#' reliability = `item_loadings`.
#'
#' Each correlation is calculated in a separate model.
#' This approach saves time for longer lists of variables compared to including
#' everything in one model and
#' it also means that excluding a variable cannot change correlations between
#' other variables (which is possible when all are included together).
#' The function uses either Burt's 2-stage procedure when `nagy = FALSE` or
#' Nagy and colleagues' (2017) extension procedure when `nagy = TRUE`
#' to control for interpretational confounding (Burt, 1976).
#'
#' Burt's method works by fixing measurement model parameters in the model
#' estimating structural parameters.
#' This method means that less than ideal fit at the measurement level does not
#' propagate though the model as the measurement parameters are fixed.
#' It also means that the latent variables interpretation cannot change with the
#' addition of different variables,
#' thereby solving interpretational confounding.
#' However, it is not a perfect solution because it underestimates uncertainty
#' in the measurement part of the structural model (e.g., Nagy et al., 2017),
#' which results in biased standard errors and fit statistics.
#'
#' Alternatively, Nagy's method involves allowing item residuals to correlate
#' with external variables (or factors) and constrains those relationships
#' such that the model is identifiable.
#' If the sums of squares of correlations between all combinations of factors'
#' items and external factors are minimised,
#' measurement parameters in the structural model match those of isolated
#' measurement models without having to constrain them directly.
#' As a result, unbiased standard errors are preserved while simultaneously
#' eliminating interpretational confounding.
#'
#' Burt's method is faster but artificially constrains parameters,
#' thereby biasing standard errors and model fit indices.
#' They should, however, give very similar point estimates for correlations.
#' Therefore, Nagy's method should be preferred whenever confidence intervals
#' or model fit matter, except for extremely large sets of variables.
#'
#' It is possible for latent variable correlations to produce a non-positive
#' definite correlation matrix between variables included in `fit_y`,
#' especially when closely related factors are included.
#' If the matrix of latent variables is not positive definite,
#' then the matrix will be adjusted to the nearest positive definite
#' matrix using the [Matrix::nearPD()] function,
#' which employs the method developed by Higham (2002),
#' and a message will state that the matrix was adjusted and the maximum
#' adjustment to any cell.
#' Confidence intervals will be adjusted by the same amount.
#'
#' The model relies on [sem.check()] for the back-end of running the models,
#' which enables saving inputs and outputs from model runs
#' (with `save_out = TRUE`) and checking to see if anything has changed from
#' prior runs before running again (with `check = TRUE`).
#' The functionality was included for a number of very slow models or a lot of
#' faster models, such that time spent rerunning them would be onerous.
#' In the case of `sem.cor()`, the number of correlations can add up quickly,
#' so the functionality may be useful
#' (e.g., with 20 scales, there are 19 + 18 + 17 + ... + 1 = 190 correlations).
#' For further details on how this works, see the [sem.check()] function
#' documentation.
#'
#' @seealso
#' [sem.check()], which `sem.cor()` uses for all the back-end;
#' [lavaan::sem()], which is used to estimate the models;
#' [matrixcalc::is.positive.definite], which is used to assess whether the
#' correlation matrix between `fit_y` constructs is positive definite; and
#' [Matrix::nearPD()] for the function used to transform non-positive definite
#' matrices into positive definite matrices.
#'
#' @importFrom matrixcalc is.positive.definite
#' @importFrom Matrix nearPD
#' @export
#'
#' @references
#' Burt, R. S. (1976).
#' Interpretational confounding of unobserved variables in Structural Equation
#' Models. Sociological Methods & Research, 5(1), 3-52.
#' https://doi.org/10.1177/004912417600500101.
#'
#' Higham, N. J. (2002).
#' Computing the nearest correlation matrix—a problem from finance.
#' IMA Journal of Numerical Analysis, 22(3), 329-343.
#' https://doi.org/10.1093/imanum/22.3.329.
#'
#' Nagy, G., Brunner, M., Lüdtke, O., and Greiff, S. (2017).
#' Extension Procedures for Confirmatory Factor Analysis.
#' Journal of Experimental Education, 85(4), 574-596.
#' https://doi.org/10.1080/00220973.2016.1260524.
#'
#' @examples
#' # Create CFA keys
#' keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
#' keys <- sapply(
#'   keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
#' )
#' # Run CFA models
#' cfa_fit <- cfa.from.keys(keys, BFIGritHope, check = FALSE, fit_save = FALSE)
#' # Find correlations between all cfa_fit constructs.
#' cors <- sem.cor(BFIGritHope, cfa_fit$fit, nagy = FALSE)
#' # View the correlation matrix
#' cors$cor_mat
#'
#' # Correlations of grit facets with hope facets and the first item from each
#' # Big Five factor.
#' items <- names(BFIGritHope)[grep("bfi_.*1_1", names(BFIGritHope))]
#' cors2 <- sem.cor(
#'   BFIGritHope, cfa_fit$fit[1:2], cfa_fit$fit[3:4], items, nagy = FALSE
#' )
#' # View correlations
#' cors2$cor_mat

sem.cor <- function(
    data, fit_y, fit_x = NULL, items = NULL,  item_loadings = NULL, nagy = TRUE,
    fit_save = FALSE, fit_measures = "all", miss = "ML", est = "default",
    name = "cors", check = FALSE, save_out = FALSE
) {
  # Single model instead of list.
  if (!is.list(fit_y) & inherits(fit_y, "lavaan")) {
    fit_y <- list(factor = fit_y)
  }
  if (sum(sapply(fit_y, function(x) !inherits(x, "lavaan"))) > 0) {
    stop(
      paste(
        "At least one of the elements of 'fit_y' is not an object of class",
        "lavaan."
      )
    )
  }
  if (!is.null(items)) {
    if (sum(!items %in% names(data)) > 0) {
      item_miss <- items[!items %in% names(data)]
      stop(
        paste0(
          "The following item(s) in 'items' are not in 'data':\n  ",
          paste(item_miss, collapse = "\n  ")
        )
      )
    }
    if (sum(items %in% names(fit_y)) > 0) {
      item_overlap <- items[items %in% names(fit_y)]
      stop(
        paste0(
          "The following item(s) in 'items' are included as a name of a fitted",
          " model in 'fit_y'. ",
          "Please ensure item names do not conflict with latent variable ",
          "names.\n    ",
          paste(item_overlap, collapse = "\n    ")
        )
      )
    }
    if (!is.null(item_loadings)) {
      if (length(item_loadings) != 1 & length(items) != length(item_loadings)) {
        stop(
          paste(
            "'item_loadings' must be either 'NULL', length 1, or length equal",
            "to the lenght or items."
          )
        )
      }
    }
  }
  if (!is.null(fit_x)) {
    # Single model instead of list.
    if (!is.list(fit_x) & inherits(fit_x, "lavaan")) {
      fit_x <- list(factor = fit_x)
    }
    if (sum(sapply(fit_x, function(x) !inherits(x, "lavaan"))) > 0) {
      stop(
        paste(
          "At least one of the elements of 'fit_x' is not an object of class",
          "lavaan."
        )
      )
    }
    if (sum(items %in% names(fit_x)) > 0) {
      item_overlap <- items[items %in% names(fit_x)]
      stop(
        paste0(
          "The following item(s) in 'items' are included as a name of a fitted",
          " model in 'fit_x'. ",
          "Please ensure item names do not conflict with latent variable ",
          "names.\n    ",
          paste(item_overlap, collapse = "\n    ")
        )
      )
    }
  }
  if (length(fit_y) <= 1) {
    if (is.null(items) & is.null(fit_x)) {
      stop(
        paste(
          "'fit_y' is not at least length 2, and 'fit_x' and items are not",
          "specified.",
          "At least one measurement model and one item or two measurement",
          "models must be specified to calculate correlations with or between",
          "latent variables."
        )
      )
    }
  }
  par1 <- lapply(fit_y, parameterEstimates)
  # Rename y objects to match factors
  names(fit_y) <- names(par1) <- sapply(
    par1,
    function(y) {
      y1 <- y[y$op %in% c("~~", "=~"), ]
      yn <- unique(y$lhs[y$op == "=~"])
      if (length(yn) > 1) {
        stop(
          paste0(
            "The 'fit_y' model including '", paste(yn, collapse = "' and '"),
            "' includes more than one latent variable. ",
            "These models are not currently supported by 'sem.cor()'."
          )
        )
      } else {
        yn
      }
    }
  )
  if (is.null(fit_x) & length(fit_y) >= 2) {
    pars <- lapply(
      stats::setNames(
        seq_along(par1[-length(par1)]),
        names(par1)[seq_along(par1[-length(par1)])]
      ),
      function(y) {
        lapply(
          stats::setNames(
            (y + 1):length(par1), names(par1)[(y + 1):length(par1)]
          ),
          function(x) {
            x0 <- par1[[x]]
            y0 <- par1[[y]]
            list(x0 = x0, y0 = y0)
          }
        )
      }
    )
  } else if (!is.null(fit_x)) {
    # Correlations between constructs in fit_y and constructs in fit_x
    par2 <- lapply(fit_x, parameterEstimates)
    names(fit_x) <- names(par2) <- sapply(
      par2,
      function(x) {
        x1 <- x[x$op %in% c("~~", "=~"), ]
        xn <- unique(x$lhs[x$op == "=~"])
        if (length(xn) > 1) {
          stop(
            paste0(
              "The 'fit_x' model including '", paste(xn, collapse = "' and '"),
              "' includes more than one latent variable. ",
              "These models are not currently supported by 'sem.cor()'. ",
              "Please include them separately."
            )
          )
        } else {
          xn
        }
      }
    )
    pars <- sapply(
      par1,
      function(y) {
        sapply(par2, function(x) list(x0 = x, y0 = y), simplify = FALSE)
      },
      simplify = FALSE
    )
  }
  if ((is.null(fit_x) & length(fit_y) >= 2) | !is.null(fit_x)) {
    mod_key <- lapply(
      pars,
      function(k) {
        tmp <- lapply(
          k,
          function(j) {
            x <- j[["x0"]]
            y <- j[["y0"]]
            x1 <- x[x$op == "~~" | x$op == "=~", ]
            y1 <- y[y$op == "~~" | y$op == "=~", ]
            key_x <- unique(x1$rhs[x1$op == "=~"])
            key_y <- unique(y1$rhs[y1$op == "=~"])
            key0 <- unique(c(key_x, key_y))
            xn <- unique(x1$lhs[x1$op == "=~"])
            yn <- unique(y1$lhs[y1$op == "=~"])
            # Any shared items. Need to unfix residual variance for these.
            i <- x1$lhs[x1$lhs %in% y1$lhs]
            if (length(i) != 0) {
              if (nagy) {
                stop(
                  paste0(
                    "The following item(s) are in both the '", yn, "' and '",
                    xn, "' factors. ",
                    "This is not supported when 'nagy = TRUE'.\n    ",
                    paste(i, collapse = "\n    ")
                  )
                )
              }
              if (length(i) >= min(length(key_x), length(key_y))) {
                if (length(key_x) < length(key_y)) {
                  shorter_f <- xn
                  longer_f <- yn
                }
                if (length(key_x) > length(key_y)) {
                  shorter_f <- yn
                  longer_f <- xn
                }
                if (length(key_x) == length(key_y)) {
                  stop(
                    paste0(
                      "All items between '", yn, "' and '", xn, "' are shared.",
                      "\nHave you included the same latent variable twice by ",
                      "mistake?"
                    )
                  )
                }
                stop(
                  paste0(
                    "All the items in '", shorter_f, "' are included in '",
                    longer_f, "'.\n  ",
                    "Such a relationship should be specified as a bifactor ",
                    "model with a correlation of 0 between the group and ",
                    "general factors."
                  )
                )
              }
              warning(
                paste0(
                  "The following item(s) are in both the '", yn, "' and '", xn,
                  "' factors. ",
                  "If this is not intended, please correct it and disregard ",
                  "the correlation between these factors.\n    ",
                  paste(i, collapse = "\n    ")
                )
              )
              for (j in i) {
                x1 <- x1[!(x1$lhs == j & x1$op == "~~" & x1$rhs == j), ]
                y1 <- y1[!(y1$lhs == j & y1$op == "~~" & y1$rhs == j), ]
              }
            }
            if (!nagy) {
              mod0 <- paste0(
                # CFA1
                paste(x1$lhs, x1$op, x1$est, "*", x1$rhs, collapse = "\n"),
                "\n",
                # CFA2
                paste(y1$lhs, y1$op, y1$est, "*", y1$rhs, collapse = "\n"),
                collapse = "\n"
              )
              return(list(mod = mod0, key = key0))
            } else {
              x1l <- x1[x1$op == "=~", ]
              y1l <- y1[y1$op == "=~", ]
              x1u <- x1[x1$op == "~~" & x1$lhs != xn, ]
              y1u <- y1[y1$op == "~~" & y1$lhs != yn, ]
              x1v <- x1[x1$lhs == x1$rhs & x1$lhs == xn, ]
              y1v <- y1[y1$lhs == y1$rhs & y1$lhs == yn, ]
              mod0 <- paste0(
                # CFA1
                paste0(
                  x1l$lhs, x1l$op, "lx", seq_along(key_x), "*", x1l$rhs,
                  collapse = "\n"
                ),
                "\n",
                paste0(
                  x1u$lhs, x1u$op, "dx", seq_along(key_x), "*", x1u$rhs,
                  collapse = "\n"
                ),
                "\n",
                paste0(x1v$lhs, x1v$op, x1v$est, "*", x1v$rhs),
                "\n",
                # CFA2
                paste0(
                  y1l$lhs, y1l$op, "ly", seq_along(key_y), "*", y1l$rhs,
                  collapse = "\n"
                ),
                "\n",
                paste0(
                  y1u$lhs, y1u$op, "dy", seq_along(key_y), "*", y1u$rhs,
                  collapse = "\n"
                ),
                "\n",
                paste0(y1v$lhs, y1v$op, y1v$est, "*", y1v$rhs),
                "\n",
                # Extension parameters
                paste0(
                  mapply(
                    ye = key_y, yes = seq_along(key_y),
                    FUN = function(ye, yes) paste0(ye, "~~pxy", yes, "*", xn)
                  ),
                  collapse = "\n"
                ),
                "\n",
                paste0(
                  mapply(
                    xe = key_x, xes = seq_along(key_x),
                    FUN = function(xe, xes) paste0(xe, "~~pyx", xes, "*", yn)
                  ),
                  collapse = "\n"
                ),
                "\n",
                # Model constraints
                paste0(
                  "0==",
                  paste0(
                    sapply(
                      seq_along(key_y),
                      function(ys) {
                        paste0("ly", ys, "*pxy", ys, "/dy", ys)
                      }
                    ),
                    collapse = "+"
                  )
                ),
                "\n",
                paste0(
                  "0==",
                  paste0(
                    sapply(
                      seq_along(key_x),
                      function(xs) paste0("lx", xs, "*pyx", xs, "/dx", xs)
                    ),
                    collapse = "+"
                  )
                )
              )
              return(list(mod = mod0, key = key0))
            }
          }
        )
        mod1 <- lapply(tmp, function(x) x$mod)
        key1 <- lapply(tmp, function(x) x$key)
        return(list(mod = mod1, key = key1))
      }
    )
    mods <- unlist(lapply(mod_key, function(x) x$mod), recursive = FALSE)
    key <- unlist(lapply(mod_key, function(x) x$key), recursive = FALSE)
  } else {
    mods <- NULL
    key <- NULL
  }
  if (!is.null(items)) {
    # Correlations with single items
    mod_key_i <- lapply(
      par1,
      function(y) {
        y1 <- y[y$op %in% c("~~", "=~"), ]
        yn <- unique(y$lhs[y$op == "=~"])
        if (nagy) {
          y1l <- y1[y1$op == "=~", ]
          y1u <- y1[y1$op == "~~" & y1$lhs != yn, ]
          y1v <- y1[y1$lhs == y1$rhs & y1$lhs == yn, ]
          key_y <- unique(y1$rhs[y1$op == "=~"])
        }
        item_overlap <- items[items %in% y1$rhs]
        if (length(item_overlap) > 0) {
          stop(
            paste0(
              "The following item(s) are in both 'items' and contributes ",
              "to the measurement of the '", yn, "' latent variable.",
              "\n  This is not supported.\n  ",
              "Either remove the item(s) from 'items' or ",
              "(if appropriate) remove the item(s) from the latent ",
              "measurement of '", yn, "'.\n    ",
              paste(item_overlap, collapse = "\n    ")
            )
          )
        }
        if (length(item_loadings) == length(items)) {
          names(item_loadings) <- items
        }
        tmp <- lapply(
          stats::setNames(nm = items),
          function(i) {
            if (!is.null(item_loadings)) {
              if (length(item_loadings) == length(items)) {
                i_r <- paste(item_loadings[i], " * ")
              } else {
                i_r <- paste(item_loadings, " * ")
              }
            } else {
              i_r <- ""
            }
            i_l <- paste0(i, "_l")
            key0 <- c(y$rhs[y$op == "=~"], i)
            if (!nagy) {
              mod0 <- paste0(
                # CFA
                paste(y1$lhs, y1$op, y1$est, "*", y1$rhs, collapse = "\n"),
                "\n",
                paste0(i_l, " =~ ", i_r, i),
                collapse = "\n"
              )
            } else {
              mod0 <- paste0(
                # CFA
                paste0(
                  y1l$lhs, y1l$op, "ly", seq_along(key_y), "*", y1l$rhs,
                  collapse = "\n"
                ),
                "\n",
                paste0(
                  y1u$lhs, y1u$op, "dy", seq_along(key_y), "*", y1u$rhs,
                  collapse = "\n"
                ),
                "\n",
                paste0(y1v$lhs, y1v$op, y1v$est, "*", y1v$rhs),
                "\n",
                # Item latent variable
                paste0(i_l, "=~", i_r, i),
                "\n",
                # Correlation
                paste0(yn, "~~", paste0(i_l, collapse = "+")),
                "\n",
                # Extension parameters
                paste0(
                  mapply(
                    ye = key_y, yes = seq_along(key_y),
                    FUN = function(ye, yes) paste0(ye, "~~piy", yes, "*", i_l)
                  ),
                  collapse = "\n"
                ),
                "\n",
                # Model constraints
                paste0(
                  "0==",
                  paste0(
                    sapply(
                      seq_along(key_y),
                      function(ys) paste0("ly", ys, "*piy", ys, "/dy", ys)
                    ),
                    collapse = "+"
                  )
                ),
                collapse = "\n"
              )
            }
            return(list(mod = mod0, key = key0))
          }
        )
        mod1 <- lapply(tmp, function(x) x$mod)
        key1 <- lapply(tmp, function(x) x$key)
        return(list(mod = mod1, key = key1))
      }
    )
    mods_i <- unlist(lapply(mod_key_i, function(x) x$mod), recursive = FALSE)
    key_i <- unlist(lapply(mod_key_i, function(x) x$key), recursive = FALSE)
    if (is.null(mods)) {
      mods <- mods_i
      key <- key_i
    } else {
      mods <- c(mods, mods_i)
      key <- c(key, key_i)
    }
  }
  fit <- sem.check(
    mods,
    data,
    name = name,
    keys_s = key,
    fit_save = fit_save,
    fit_measures = fit_measures,
    miss = miss,
    est = est,
    check = check,
    save_out = save_out,
    std.lv = TRUE
  )
  if (!is.null(fit_x) | length(fit_y) > 1) {
    xn0 <- sub(".*\\.", "", names(mods))
    xn <- xn0[!xn0 %in% items]
    yn <- sub("\\..*", "", names(mods))[!xn0 %in% items]
    cors_y <- mapply(
      x = fit$par_std[
        !grepl(paste0("\\.", items, "$", collapse = "|"), names(fit$par_std))
      ],
      xn0 = xn,
      yn0 = yn,
      FUN = function(x, xn0, yn0) x$est.std[x$lhs == xn0 & x$rhs == yn0]
    )
    ci_lower_y0 <- mapply(
      x = fit$par_std[
        !grepl(paste0("\\.", items, "$", collapse = "|"), names(fit$par_std))
      ],
      xn0 = xn,
      yn0 = yn,
      FUN = function(x, xn0, yn0) x$ci.lower[x$lhs == xn0 & x$rhs == yn0]
    )
    ci_upper_y0 <- mapply(
      x = fit$par_std[
        !grepl(paste0("\\.", items, "$", collapse = "|"), names(fit$par_std))
      ],
      xn0 = xn,
      yn0 = yn,
      FUN = function(x, xn0, yn0) x$ci.upper[x$lhs == xn0 & x$rhs == yn0]
    )
    if (is.null(fit_x)) {
      cor_mat_y <- sapply(
        names(fit_y),
        function(x) {
          sapply(
            names(fit_y),
            function(y) {
              ptn <- paste0(x, "\\.", y, "|", y, "\\.", x)
              tmp <- cors_y[grep(ptn, names(cors_y))]
              if (length(tmp) == 0) {
                tmp <- 1
              }
              tmp
            },
            USE.NAMES = FALSE
          )
        }
      )
      rownames(cor_mat_y) <- names(fit_y)
      ci_lower_y <- sapply(
        names(fit_y),
        function(x) {
          sapply(
            names(fit_y),
            function(y) {
              ptn <- paste0(x, "\\.", y, "|", y, "\\.", x)
              tmp <- ci_lower_y0[grep(ptn, names(ci_lower_y0))]
              if (length(tmp) == 0) {
                tmp <- 1
              }
              tmp
            },
            USE.NAMES = FALSE
          )
        }
      )
      rownames(ci_lower_y) <- names(fit_y)
      ci_upper_y <- sapply(
        names(fit_y),
        function(x) {
          sapply(
            names(fit_y),
            function(y) {
              ptn <- paste0(x, "\\.", y, "|", y, "\\.", x)
              tmp <- ci_upper_y0[grep(ptn, names(ci_upper_y0))]
              if (length(tmp) == 0) {
                tmp <- 1
              }
              tmp
            },
            USE.NAMES = FALSE
          )
        }
      )
      rownames(ci_upper_y) <- names(fit_y)
    } else {
      cor_mat_y <- sapply(
        names(fit_y),
        function(y) {
          sapply(
            names(fit_x),
            function(x) {
              ptn <- c(paste0(x, ".", y), paste0(y, ".", x))
              cors_y[names(cors_y) %in% ptn]
            },
            USE.NAMES = FALSE
          )
        }
      )
      ci_lower_y <- sapply(
        names(fit_y),
        function(y) {
          sapply(
            names(fit_x),
            function(x) {
              ptn <- c(paste0(x, ".", y), paste0(y, ".", x))
              ci_lower_y0[names(ci_lower_y0) %in% ptn]
            },
            USE.NAMES = FALSE
          )
        }
      )
      ci_upper_y <- sapply(
        names(fit_y),
        function(y) {
          sapply(
            names(fit_x),
            function(x) {
              ptn <- c(paste0(x, ".", y), paste0(y, ".", x))
              ci_upper_y0[names(ci_upper_y0) %in% ptn]
            },
            USE.NAMES = FALSE
          )
        }
      )
      if (is.vector(cor_mat_y) & length(fit_x) == 1) {
        cor_mat_y <- matrix(cor_mat_y, nrow = 1)
        colnames(cor_mat_y) <- names(fit_y)
        ci_lower_y <- matrix(ci_lower_y, nrow = 1)
        colnames(ci_lower_y) <- names(fit_y)
        ci_upper_y <- matrix(ci_upper_y, nrow = 1)
        colnames(ci_upper_y) <- names(fit_y)
      }
      rownames(cor_mat_y) <- names(fit_x)
      rownames(ci_lower_y) <- names(fit_x)
      rownames(ci_upper_y) <- names(fit_x)
    }
  }
  if (!is.null(items)) {
    cors_yi <- do.call(
      rbind,
      mapply(
        x =
          fit$par_std[grepl(paste0(items, collapse = "|"), names(fit$par_std))],
        y = rep(names(fit_y), each = length(items)),
        SIMPLIFY = FALSE,
        FUN = function(x, y) {
          x[
            x$lhs != x$rhs & x$op == "~~" & grepl(y, x$lhs),
            c("lhs", "rhs", "est.std", "ci.lower", "ci.upper")
          ]
        }
      )
    )
    items_l <- paste0(items, "_l")
    cor_mat_yi <- sapply(
      names(fit_y),
      function(x) {
        sapply(
          items_l,
          function(y) cors_yi$est.std[cors_yi$lhs == x & cors_yi$rhs == y]
        )
      }
    )
    ci_lower_yi <- sapply(
      names(fit_y),
      function(x) {
        sapply(
          items_l,
          function(y) cors_yi$ci.lower[cors_yi$lhs == x & cors_yi$rhs == y]
        )
      }
    )
    ci_upper_yi <- sapply(
      names(fit_y),
      function(x) {
        sapply(
          items_l,
          function(y) cors_yi$ci.upper[cors_yi$lhs == x & cors_yi$rhs == y]
        )
      }
    )
    if (is.vector(cor_mat_yi) & length(items) == 1) {
      cor_mat_yi <- matrix(cor_mat_yi, nrow = 1)
      colnames(cor_mat_yi) <- names(fit_y)
      ci_lower_yi <- matrix(ci_lower_yi, nrow = 1)
      colnames(ci_lower_yi) <- names(fit_y)
      ci_upper_yi <- matrix(ci_upper_yi, nrow = 1)
      colnames(ci_upper_yi) <- names(fit_y)
    }
    rownames(cor_mat_yi) <- items
    rownames(ci_lower_yi) <- items
    rownames(ci_upper_yi) <- items
  }
  if (is.null(fit_x) & length(fit_y) > 1) {
    if (!is.positive.definite(cor_mat_y)) {
      cor_mat_y0 <- as.matrix(nearPD(cor_mat_y, corr = TRUE)$mat)
      dif_mat <- cor_mat_y - cor_mat_y0
      cor_mat_y <- cor_mat_y0
      ci_lower_y <- ci_lower_y - dif_mat
      ci_upper_y <- ci_upper_y - dif_mat
      max_adj <- round(max(abs(dif_mat)), 3)
      max_adj <- ifelse(
        max_adj == 0, "< .001", format(max_adj, scientific = FALSE)
      )
      warning(
        paste0(
          "The correlation matrix between 'fit_y' constructs has been adjusted",
          " from initial estimates with the 'Matrix::nearPD()' function ",
          "to ensure it is positive definite.\n  ",
          "The maximum adjustment to any cell was ", max_adj, ".\n  ",
          "'ci_lower' and 'ci_upper' were adjusted by the same absolute amount",
          " as the primary correlation matrix."
        )
      )
    }
  }
  if (is.null(items)) {
    cor_mat <- cor_mat_y
    ci_lower <- ci_lower_y
    ci_upper <- ci_upper_y
  } else if (is.null(fit_x) & length(fit_y) == 1) {
    cor_mat <- cor_mat_yi
    ci_lower <- ci_lower_yi
    ci_upper <- ci_upper_yi
  } else {
    cor_mat <- rbind(cor_mat_y, cor_mat_yi)
    ci_lower <- rbind(ci_lower_y, ci_lower_yi)
    ci_upper <- rbind(ci_upper_y, ci_upper_yi)
  }
  if (fit_save) {
    return(
      list(
        fit = fit$fit, fit_measures = fit$fit_measures,
        cor_mat = cor_mat, ci_lower = ci_lower, ci_upper = ci_upper
      )
    )
  } else {
    return(
      list(
        fit = fit$fit,
        cor_mat = cor_mat, ci_lower = ci_lower, ci_upper = ci_upper
      )
    )
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sem.cor.R
\name{sem.cor}
\alias{sem.cor}
\title{Creates a latent variable correlation matrix from fitted lavaan measurement
models}
\usage{
sem.cor(
  data,
  fit_y,
  fit_x = NULL,
  items = NULL,
  item_loadings = NULL,
  nagy = TRUE,
  fit_save = FALSE,
  fit_measures = "all",
  miss = "ML",
  est = "default",
  name = "cors",
  check = FALSE,
  save_out = FALSE
)
}
\arguments{
\item{data}{A dataframe or object coercible to a dataframe.
Data must include all observed variables used in any of the models.}

\item{fit_y}{A named list of CFA fitted objects.}

\item{fit_x}{A named list of CFA fitted objects to be correlated with fit_y variables
or 'NULL'.}

\item{items}{A vector of single-item variables to correlate with 'fit_y' latent variables.
Must not include any items contributing to the measurement of a \code{fit_y}
latent variable.}

\item{item_loadings}{When single items are specified, items are included in models with single
item latent variables. \code{item_loadings} sets the loading of the item on the
factor.
It can be a single number to set all loadings equal or a vector of
length equal to the length of items.
Defaults allows the value to be free (which assumes perfect reliability).
Irrelevant if \code{items = NULL}.}

\item{nagy}{Logical. Indicates whether to use Nagy and colleagues' (2017) extension
procedure instead of Burt's (1976) 2-stage procedure.}

\item{fit_save}{Logical.
\code{TRUE} indicates model fit measures should be included in the output;
\code{FALSE} indicates model fit measures should not be included in the output.
\code{FALSE} may be desirable when fit measures are of little interest and when
they may take a long time to estimate.
Fit can still be examined for individual models with,
\verb{lavaan::fitMeasures(fit$fit$[model name])}.}

\item{fit_measures}{A vector of fit measures to save or 'all' to select all fit measures,
as per the \code{fit.measures} parameter from lavaan's \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}
function.
Defaults to 'all'. Irrelevant if \code{fit_save = FALSE}.}

\item{miss}{A string.
Sets the \code{missing} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to 'ML'.}

\item{est}{A string.
Sets the \code{estimator} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
The default ('default') uses the lavaan default for the model being run.}

\item{name}{A string indicating a subdirectory where model outputs will be saved when
\code{save_out = TRUE} and checked against when \code{check = TRUE}.
Defaults to "cors".
Irrelevant if both \code{save_out = FALSE} and \code{check = FALSE}.
The name should be unique for each set of models, or outputs from calls with
the same name will be overwritten.}

\item{check}{Logical.
\code{TRUE} indicates that current inputs should be compared to previous inputs
if they exist and that the model should not be rerun if nothing has changed;
\code{FALSE} indicates that these checks should not be made and the model should
be run regardless of the existence of previous inputs.}

\item{save_out}{Logical.
\code{TRUE} indicates that model code, a hash of the data, important input
parameter values, and output will be saved;
\code{FALSE} indicates that nothing will be saved.
Selecting \code{save_out = TRUE} enables the function to not rerun models next
time if \code{check = TRUE} the next time the code is run and nothing has changed
in the meantime.}
}
\value{
Returns a list of length 4.
The first element is the fitted lavaan models ('fit').
The second element is the correlation matrix ('cor_mat').
The third and fourth elements are the upper and lower confidence intervals
from the models ('ci_lower' and 'ci_upper', respectively).
}
\description{
sem.cor takes CFA outputs and produces a correlation matrix between latent
variables and/or single items.
}
\details{
The function computes correlations between latent variables from fitted CFA
models, including either all latent variables, or between two sets of latent
variables.
To compute a full correlation matrix include all CFA models in a list as
\code{fit_y}, alternatively,
to correlate a set of dependant variables with a set of independent
variables,
set \code{fit_y} as a list of fitted CFA models of dependant variables and
set \code{fit_x} as a list of fitted CFA models of independent variables.
Correlations between the \code{fit_y} latent variables and single items can also
be optionally included by setting \code{items} as a vector of item names.
In this case, items will be treated as single item latent variables with
reliability = \code{item_loadings}.

Each correlation is calculated in a separate model.
This approach saves time for longer lists of variables compared to including
everything in one model and
it also means that excluding a variable cannot change correlations between
other variables (which is possible when all are included together).
The function uses either Burt's 2-stage procedure when \code{nagy = FALSE} or
Nagy and colleagues' (2017) extension procedure when \code{nagy = TRUE}
to control for interpretational confounding (Burt, 1976).

Burt's method works by fixing measurement model parameters in the model
estimating structural parameters.
This method means that less than ideal fit at the measurement level does not
propagate though the model as the measurement parameters are fixed.
It also means that the latent variables interpretation cannot change with the
addition of different variables,
thereby solving interpretational confounding.
However, it is not a perfect solution because it underestimates uncertainty
in the measurement part of the structural model (e.g., Nagy et al., 2017),
which results in biased standard errors and fit statistics.

Alternatively, Nagy's method involves allowing item residuals to correlate
with external variables (or factors) and constrains those relationships
such that the model is identifiable.
If the sums of squares of correlations between all combinations of factors'
items and external factors are minimised,
measurement parameters in the structural model match those of isolated
measurement models without having to constrain them directly.
As a result, unbiased standard errors are preserved while simultaneously
eliminating interpretational confounding.

Burt's method is faster but artificially constrains parameters,
thereby biasing standard errors and model fit indices.
They should, however, give very similar point estimates for correlations.
Therefore, Nagy's method should be preferred whenever confidence intervals
or model fit matter, except for extremely large sets of variables.

It is possible for latent variable correlations to produce a non-positive
definite correlation matrix between variables included in \code{fit_y},
especially when closely related factors are included.
If the matrix of latent variables is not positive definite,
then the matrix will be adjusted to the nearest positive definite
matrix using the \code{\link[Matrix:nearPD]{Matrix::nearPD()}} function,
which employs the method developed by Higham (2002),
and a message will state that the matrix was adjusted and the maximum
adjustment to any cell.
Confidence intervals will be adjusted by the same amount.

The model relies on \code{\link[=sem.check]{sem.check()}} for the back-end of running the models,
which enables saving inputs and outputs from model runs
(with \code{save_out = TRUE}) and checking to see if anything has changed from
prior runs before running again (with \code{check = TRUE}).
The functionality was included for a number of very slow models or a lot of
faster models, such that time spent rerunning them would be onerous.
In the case of \code{sem.cor()}, the number of correlations can add up quickly,
so the functionality may be useful
(e.g., with 20 scales, there are 19 + 18 + 17 + ... + 1 = 190 correlations).
For further details on how this works, see the \code{\link[=sem.check]{sem.check()}} function
documentation.
}
\examples{
# Create CFA keys
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
# Run CFA models
cfa_fit <- cfa.from.keys(keys, BFIGritHope, check = FALSE, fit_save = FALSE)
# Find correlations between all cfa_fit constructs.
cors <- sem.cor(BFIGritHope, cfa_fit$fit, nagy = FALSE)
# View the correlation matrix
cors$cor_mat

# Correlations of grit facets with hope facets and the first item from each
# Big Five factor.
items <- names(BFIGritHope)[grep("bfi_.*1_1", names(BFIGritHope))]
cors2 <- sem.cor(
  BFIGritHope, cfa_fit$fit[1:2], cfa_fit$fit[3:4], items, nagy = FALSE
)
# View correlations
cors2$cor_mat
}
\references{
Burt, R. S. (1976).
Interpretational confounding of unobserved variables in Structural Equation
Models. Sociological Methods & Research, 5(1), 3-52.
https://doi.org/10.1177/004912417600500101.

Higham, N. J. (2002).
Computing the nearest correlation matrix—a problem from finance.
IMA Journal of Numerical Analysis, 22(3), 329-343.
https://doi.org/10.1093/imanum/22.3.329.

Nagy, G., Brunner, M., Lüdtke, O., and Greiff, S. (2017).
Extension Procedures for Confirmatory Factor Analysis.
Journal of Experimental Education, 85(4).
https://doi.org/10.1080/00220973.2016.1260524.
}
\seealso{
\code{\link[=sem.check]{sem.check()}}, which \code{sem.cor()} uses for all the back-end;
\code{\link[lavaan:sem]{lavaan::sem()}}, which is used to estimate the models;
\link[matrixcalc:is.positive.definite]{matrixcalc::is.positive.definite}, which is used to assess whether the
correlation matrix between \code{fit_y} constructs is positive definite; and
\code{\link[Matrix:nearPD]{Matrix::nearPD()}} for the function used to transform non-positive definite
matrices into positive definite matrices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/esem.from.mods.R
\name{esem.from.mods}
\alias{esem.from.mods}
\title{Runs ESEM based on CFA and EFA model outputs.}
\usage{
esem.from.mods(
  data,
  efa_fit,
  cfa_fit = NULL,
  bif_fit = NULL,
  fit_save = FALSE,
  fit_measures = "all",
  miss = "ML",
  est = "default",
  name = "esem",
  check = FALSE,
  save_out = FALSE
)
}
\arguments{
\item{data}{A dataframe or object coercible to a dataframe.
Data must include all observed variables used in any of the models.}

\item{efa_fit}{A fitted lavaan object of an EFA model.}

\item{cfa_fit}{A named list of fitted lavaan objects of CFA models.
Can be \code{NULL} if \code{bif_fit} is not \code{NULL}.}

\item{bif_fit}{A named list of fitted lavaan objects of bifactor models.
Can be \code{NULL} if \code{cfa_fit} is not \code{NULL}.}

\item{fit_save}{Logical.
\code{TRUE} indicates model fit measures should be included in the output;
\code{FALSE} indicates model fit measures should not be included in the output.
\code{FALSE} may be desirable when fit measures are of little interest and when
they may take a long time to estimate.
Fit can still be examined for individual models with,
\verb{lavaan::fitMeasures(fit$fit$[model name])}.}

\item{fit_measures}{A vector of fit measures to save or 'all' to select all fit measures,
as per the \code{fit.measures} parameter from lavaan's \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}
function.
Defaults to 'all'. Irrelevant if \code{fit_save = FALSE}.}

\item{miss}{A string.
Sets the \code{missing} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to 'ML'.}

\item{est}{A string.
Sets the \code{estimator} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
The default ('default') uses the lavaan default for the model being run.}

\item{name}{A string indicating a subdirectory where model outputs will be saved when
\code{save_out = TRUE} and checked against when \code{check = TRUE}.
Defaults to "esem".
Irrelevant if both \code{save_out = FALSE} and \code{check = FALSE}.
The name should be unique for each set of models, or outputs from calls with
the same name will be overwritten.}

\item{check}{Logical.
\code{TRUE} indicates that current inputs should be compared to previous inputs
if they exist and that the model should not be rerun if nothing has changed;
\code{FALSE} indicates that these checks should not be made and the model should
be run regardless of the existence of previous inputs.}

\item{save_out}{Logical.
\code{TRUE} indicates that model code, a hash of the data, important input
parameter values, and output will be saved;
\code{FALSE} indicates that nothing will be saved.
Selecting \code{save_out = TRUE} enables the function to not rerun models next
time if \code{check = TRUE} the next time the code is run and nothing has changed
in the meantime.}
}
\value{
Returns a list of length 4 (if \code{fit_save = FALSE}) or
5 (if \code{fit_save = TRUE}).
The elements of the list are: a list of lavaan model output objects;
a list of parameter estimates from the models (standardized if \code{std = TRUE});
if \code{fit_save = TRUE}, a matrix of fit measures for each model;
a list of regression beta parameters from each model;
and a dataframe of R-squared values from each model.
}
\description{
\code{esem.from.keys} runs exploratory structural equation models (ESEM) in lavaan
where the exploratory factor analysis (EFA) factors predict confirmatory
factor analysis (CFA) factors and/or bifactor factors in separate models
for each CFA or bifactor model.
The function takes a fitted lavaan object from an EFA and lists of fitted
CFA and/or bifactor lavaan model objects as inputs so will typically
use outputs from \code{\link[=efa.from.keys]{efa.from.keys()}}, and \code{\link[=cfa.from.keys]{cfa.from.keys()}} or
\code{\link[=bifactor.from.keys]{bifactor.from.keys()}}.
}
\details{
The function was designed to streamline running exploratory structural
equation models (ESEM) where EFA factors predict a series of latent variables
in separate models using Burt's (1976) 2-stage procedure to prevent
interpretational confounding. The function is designed to run analyses
equivalent to that of Bainbridge, Ludeke, and Smillie (2022).

The function requires fitted lavaan objects as inputs in order to properly
employ the 2-stage procedure.
Using \code{\link[=efa.from.keys]{efa.from.keys()}}, \code{\link[=cfa.from.keys]{cfa.from.keys()}}, and/or \code{\link[=bifactor.from.keys]{bifactor.from.keys()}}
should make this relatively straight-forward.

Matching the philosophy of the package, the function is designed to run for
multiple models with a similar design. If you are using the function for a
single model, transform inputs into lists as appropriate.

The model relies on \code{\link[=sem.check]{sem.check()}} for the back-end of running the models.
This enables saving inputs and outputs from model runs
(with \code{save_out = TRUE}) and checking to see if anything has changed from
prior runs before running again (with \code{check = TRUE}).
The functionality was included for a number of very slow models or a lot of
faster models, such that time spent rerunning them would be onerous.
For further details on how this works, see the \code{\link[=sem.check]{sem.check()}} function
documentation.

Although the function will take any lavaan models that have successfully
produced standard lavaan outputs,
it will not complain if these have poor fit or other undesirable
characteristics (beyond warnings and errors produced by lavaan).
If this matters to you, it would be worth checking the outputs of upstream
functions prior to running \code{esem.from.mods()}.

There is some disagreement about how to deal with interpretational
confounding.
The standard solution (other than ignoring it) is to create good fitting
measurement models first, then freely estimate the structural model with
checks to ensure adequate fit of the model and
that interpretational confounding is not an issue.
A problem with this is that changes to measurement models can require
entirely different factors (e.g., where original authors suggested a 4-factor
scale and subsequent analyses suggest a 2-factor scale) or items to be
removed.
How ought one to assess whether the scale measures the same thing(s) with
different factors or omitted items?
When a single scale is being assessed, these issues can be resolved;
however, when many scales are being assessed this solution is impractical.

An alternative solution is to fix measurement model parameters in a model
estimating structural parameters.
This method means that less than ideal fit at the measurement level does not
propagate though the model as the measurement parameters are fixed.
It also solves interpretational confounding because the measurement of the
latent variables is fixed so their interpretation cannot change with the
inclusion of other variables in a model.
However, it is not a perfect solution because it underestimates uncertainty
in the measurement part of the structural model (e.g., Nagy et al., 2017),
which results in biased standard errors.

A final option (among those noted here) was proposed by
Nagy and colleagues (2017).
It involves allowing item residuals to correlate with external variables
(or factors) and constrains those relationships such that the model is
identifiable.
If the sums of squares of correlations between all combinations of factors'
items and external factors are minimised,
measurement parameters in isolated measurement models are preserved in the
structural model without having to constrain them directly.
As a result, unbiased standard errors are preserved while simultaneously
eliminating interpretational confounding.
Unfortunately, estimating these models becomes increasingly slow with more
items and factors, such that it quickly becomes untenable,
especially for ESEM and when many models have to be run.

As a result of these considerations, the 2-stage procedure of fixing
measurement parameters in the structural models was chosen for the current
function due to its simplicity and scalability.

A further complication occurs for bifactor models.
Recall that bifactor models require orthogonal relationships between the
general and group factors.
When a factor is an outcome of a regression in a structural model,
it is not possible to include such a constraint because the instructions
normally used to do so will constrain residual variance instead.

There are three solutions to this problem.
First, avoid using bifactor models.
This may be possible in some cases (e.g., if the scale can be modified into
a unidimentional scale), but frequently it is not.
Second, avoid using regressions in models with bifactor models and compute
regressions based on latent variable correlations.
(This can be done by, e.g., creating a correlation matrix from the factor
correlations and using the \code{\link[psych:setCor]{psych::setCor()}} function).
This method produces accurate point estimates in the regressions,
but standard errors are biased as the method cannot account for uncertainty
in the correlations.
Finally, if the 2-stage procedure is employed, then there is little room for
measurement models to change to allow factor correlations to change.
Therefore, the parameter can be relaxed and implied correlations between the
group and general factors should remain close to zero.
Given that the function already uses the 2-stage procedure,
this method is employed when bifactor models are used in \code{esem.from.mods()}.
}
\examples{
# Create CFA keys
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
# Create EFA keys
# Using only 3 factors and fewer items to save time for a simple example
# (This results in a less than ideal solution but it doesn't matter for an
# example)
keys_e0 <- paste0("bfi_", c("e", "a", "c"))
keys_e <- sapply(
  keys_e0,
  function(x) {
    names(BFIGritHope)[grep(paste0(x, "\\\\d_[1-2]"), names(BFIGritHope))]
  },
  simplify = FALSE
)
# Create fitted objects to use as inputs
cfa_fit <- cfa.from.keys(keys, BFIGritHope, fit_save = FALSE)
efa_fit <- efa.from.keys(keys_e, BFIGritHope, fit_save = FALSE)
# Run models
esem_fit <- esem.from.mods(
  efa_fit$fit$efa, cfa_fit$fit, data = BFIGritHope,
  fit_save = FALSE, check = FALSE
)
# Examine results
summary(esem_fit$fit$grit_c)  # Standard lavaan summary
esem_fit$r2                   # R-squareds
esem_fit$b                    # Betas
}
\references{
Bainbridge, T. F., Ludeke, S. G., & Smillie, L. D. (2022).
Evaluating the Big Five as an organizing framework for commonly used
psychological trait scales.
Journal of Personality and Social Psychology, 122(4), 749-777.
https://doi.org/10.1037/pspp0000395.

Burt, R. S. (1976).
Interpretational confounding of unobserved variables in Structural Equation
Models. Sociological Methods & Research, 5(1), 3-52.
https://doi.org/10.1177/004912417600500101.

Eid, M., Geiser, C., Koch, T., & Heene, M. (2017).
Anomalous results in G-factor models: Explanations and alternatives.
Psychological Methods, 22(3), 541-562.
https://doi.org/10.1037/met0000083.

Nagy, G., Brunner, M., Lüdtke, O., and Greiff, S. (2017).
Extension Procedures for Confirmatory Factor Analysis.
Journal of Experimental Education, 85(4).
https://doi.org/10.1080/00220973.2016.1260524.
}
\seealso{
\code{\link[=sem.check]{sem.check()}}, which this function uses for all the back-end;
\code{\link[=cfa.from.keys]{cfa.from.keys()}}, \code{\link[=efa.from.keys]{efa.from.keys()}}, and \code{\link[=bifactor.from.keys]{bifactor.from.keys()}},
which are useful functions for creating inputs into \code{esem.from.mods()}; and
\code{\link[lavaan:sem]{lavaan::sem()}}, which is used to estimate the models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sem.check.R
\name{sem.check}
\alias{sem.check}
\title{Runs lavaan models after checking for code or data changes.}
\usage{
sem.check(
  mods,
  data,
  keys_s = NULL,
  keys_e = NULL,
  fit_save = FALSE,
  fit_measures = "all",
  miss = "ML",
  est = "default",
  std.lv = FALSE,
  std = TRUE,
  orthogonal = FALSE,
  target = NULL,
  name = "sem",
  check = FALSE,
  save_out = FALSE
)
}
\arguments{
\item{mods}{A named list of lavaan models to run.}

\item{data}{A dataframe or object coercible to a dataframe.
Data must include all observed variables used in any of the models.}

\item{keys_s}{A named keys list matching the names and length of mod.}

\item{keys_e}{A named keys list of the factors in an ESEM to be included.}

\item{fit_save}{Logical.
\code{TRUE} indicates model fit measures should be included in the output;
\code{FALSE} indicates model fit measures should not be included in the output.
\code{FALSE} may be desirable when fit measures are of little interest and when
they may take a long time to estimate.
Fit can still be examined for individual models with,
\verb{lavaan::fitMeasures(fit$fit$[model name])}.}

\item{fit_measures}{A vector of fit measures to save or 'all' to select all fit measures,
as per the \code{fit.measures} parameter from lavaan's \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}
function.
Defaults to 'all'. Irrelevant if \code{fit_save = FALSE}.}

\item{miss}{A string.
Sets the \code{missing} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to 'ML'.}

\item{est}{A string.
Sets the \code{estimator} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
The default ('default') uses the lavaan default for the model being run.}

\item{std.lv}{Logical.
Sets the \code{std.lv} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
\code{TRUE} indicates that factor variances should be fixed to 1.
\code{FALSE} indicates that loadings of the first items of factors should be fixed
to 1.
Defaults to \code{FALSE}.}

\item{std}{Logical.
\code{TRUE} indicates that standardised parameter estimates should be saved;
\code{FALSE} indicates that unstandardised parameters estimates should be saved.}

\item{orthogonal}{Logical.
Sets the \code{orthogonal} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
\code{TRUE} indicates that unspecified latent variable correlations should be
fixed at 0;
\code{FALSE} indicates that unspecified latent variable correlations should be
freely estimated.
Defaults to \code{FALSE}.}

\item{target}{A matrix indicating a rotation target,
as used in the \code{rotation.args} argument in lavaan (see \code{\link[lavaan:efa]{lavaan::efa()}}.
If \code{NULL} (default),
target is not specified and lavaan uses default behaviour.
Irrelevant when the model does not include an EFA or ESEM.}

\item{name}{A string indicating a subdirectory where model outputs will be saved when
\code{save_out = TRUE} and checked against when \code{check = TRUE}.
Defaults to "sem".
The name should be unique for each set of models, or outputs from calls with
the same name will be overwritten.}

\item{check}{Logical.
\code{TRUE} indicates that current inputs should be compared to previous inputs
if they exist and that the model should not be rerun if nothing has changed;
\code{FALSE} indicates that these checks should not be made and the model should
be run regardless of the existence of previous inputs.}

\item{save_out}{Logical.
\code{TRUE} indicates that model code, a hash of the data, important input
parameter values, and output will be saved;
\code{FALSE} indicates that nothing will be saved.
Selecting \code{save_out = TRUE} enables the function to not rerun models next
time if \code{check = TRUE} the next time the code is run and nothing has changed
in the meantime.}
}
\value{
Returns a list of length 2 (if \code{fit_save = FALSE}) or
3 (if \code{fit_save = TRUE}).
The elements of the list are: a list of lavaan model output objects;
a list of parameter estimates from the models (standardized if \code{std = TRUE});
and, if \code{fit_save = TRUE}, a matrix of fit measures for each model.
}
\description{
sem.check takes a model list, keys lists, and data, and
produces outputs from a series of lavaan models.
The function is primarily used as a function to be called by other function
within the package but it can also be run independently.

For each model, the code optionally checks for previously saved model code,
a hash of data, and important parameter inputs.
If they exist and match values for the current code and data,
the model is not run, the previous output is loaded instead.
If either the code has changed, the hash of the data has changed, or
important parameter inputs have change, or any of these do not exist,
then the models are run as normal.
}
\details{
The function is largely intended to be used as a helper function to upstream
functions,
including \code{\link[=cfa.from.keys]{cfa.from.keys()}}, \code{\link[=bifactor.from.keys]{bifactor.from.keys()}}, \code{\link[=efa.from.keys]{efa.from.keys()}}, and
\code{\link[=esem.from.mods]{esem.from.mods()}}.
Although it is recommended to use the appropriate upstream function whenever
possible,
there are not (currently) options to do so when customised lavaan models are
required;
for example, when allowing two items' residuals to correlate in a CFA.
\code{sem.check()} can be used in these cases (see example).

Matching the philosophy of the package, the function is designed to run for
multiple models with a similar design. If you are using the function for a
single model, transform inputs into lists as appropriate or simply use lavaan
without the assistance of semFromKeys.

The function includes functionality designed to save time re-running code
when lots of slow models are included.
To do this, when \code{save_out = TRUE} and a cache directory has been set,
the model will save various inputs and outputs from the function call, and,
when \code{check = TRUE}, the model will look for any previously saved outputs
from earlier model runs in the same cache directory and
only run again if nothing has changed.
In cases where something has changed,
then the function will re-run the models where something has changed
(but it will not for those where nothing has changed).
Changes to arguments that influence all lavaan model runs (e.g., miss or est)
will trigger all models to be re-run.

For either \code{save_out = TRUE} or \code{check = TRUE},
the function will look for a cache directory set and created by the
\code{\link[=cache.setup]{cache.setup()}} function.
If a cache directory has not been set for the current session,
then the function will exit with an error suggesting that either
\code{\link[=cache.setup]{cache.setup()}} be run or \code{save_out} and \code{check} set to \code{FALSE}.

When the cache directory is found and output from previous runs are detected,
the comparisons performed are for:
\itemize{
\item model code;
\item hashes of the data (using \code{\link[openssl:md5]{openssl::md5()}});
\item values of the \code{miss}, \code{est}, \code{std}, \code{std.lv}, and \code{orthogonal}parameters;
\item the class of model objects (i.e., class lavaan); and,
\item the class of parameter estimates (i.e., class lavaan.data.frame).
}

For most applications, the checking feature can be safely ignored by not
saving outputs (i.e., \code{fit_save = FALSE})
and not checking for past saves (i.e., \code{check = FALSE}, both default),
but may be beneficial in cases with lots of models or very slow models.
However, the functionality can be safely used for faster runs too.
}
\examples{
# Create CFA keys
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
# Create model code
mods <- mapply(
  x = keys, y = names(keys), SIMPLIFY = FALSE,
  FUN = function(x, y) paste(y, "=~", paste(x, collapse = " + "))
)
# Edit model code to add correlated residuals
mods[[1]] <- paste0(mods[[1]], "\ngrit_c_1 ~~ grit_c_2")
# Estimate the models with sem.check()
cfa_fit <- sem.check(mods, BFIGritHope, keys, name = "cfa", check = FALSE)
}
\seealso{
\code{\link[=cfa.from.keys]{cfa.from.keys()}}, \code{\link[=efa.from.keys]{efa.from.keys()}}, \code{\link[=bifactor.from.keys]{bifactor.from.keys()}}, and
\code{\link[=esem.from.mods]{esem.from.mods()}}---all these function depend upon \code{sem.check()} to work;
\code{\link[lavaan:sem]{lavaan::sem()}}, which is used to estimate the models;
\code{\link[lavaan:parameterEstimates]{lavaan::parameterEstimates()}}, which is used to estimate parameter values;
and \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}, which is used to estimate fit statistics when
\code{fit_save = TRUE}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cfa.from.keys.R
\name{cfa.from.keys}
\alias{cfa.from.keys}
\title{Runs CFA models for multiple scales based on items in a keys list.}
\usage{
cfa.from.keys(
  keys,
  data,
  fit_save = TRUE,
  fit_measures = "all",
  std.lv = TRUE,
  miss = "ML",
  est = "default",
  name = "cfa",
  check = FALSE,
  save_out = FALSE
)
}
\arguments{
\item{keys}{A named list of keys.
Names should be scale names,
elements should a list of items included in each scale.}

\item{data}{A dataframe or object coercible to a dataframe.
Data must include all observed variables in any of the keys.}

\item{fit_save}{Logical.
\code{TRUE} indicates model fit measures should be included in the output;
\code{FALSE} indicates model fit measures should not be included in the output.
\code{FALSE} may be desirable when fit measures are of little interest and when
they may take a long time to estimate.
Fit can still be examined for individual models with,
\verb{lavaan::fitMeasures(fit$fit$[model name])}.}

\item{fit_measures}{A vector of fit measures to save or 'all' to select all fit measures,
as per the \code{fit.measures} parameter from lavaan's \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}
function.
Defaults to 'all'. Irrelevant if \code{fit_save = FALSE}.}

\item{std.lv}{Logical.
Sets the \code{std.lv} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
\code{TRUE} indicates that factor variances should be fixed to 1.
\code{FALSE} indicates that loadings of the first items of factors should be fixed
to 1.
Defaults to \code{TRUE}.}

\item{miss}{A string.
Sets the \code{missing} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to 'ML'.}

\item{est}{A string.
Sets the \code{estimator} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
The default ('default') uses the lavaan default for the model being run.}

\item{name}{A string indicating a subdirectory where model outputs will be saved when
\code{save_out = TRUE} and checked against when \code{check = TRUE}.
Defaults to 'cfa'.
Irrelevant if both \code{save_out = FALSE} and \code{check = FALSE}.
The name should be unique for each set of models, or outputs from calls with
the same name will be overwritten.}

\item{check}{Logical.
\code{TRUE} indicates that current inputs should be compared to previous inputs
if they exist and that the model should not be rerun if nothing has changed;
\code{FALSE} indicates that these checks should not be made and the model should
be run regardless of the existence of previous inputs.}

\item{save_out}{Logical.
\code{TRUE} indicates that model code, a hash of the data, important input
parameter values, and output will be saved;
\code{FALSE} indicates that nothing will be saved.
Selecting \code{save_out = TRUE} enables the function to not rerun models next
time if \code{check = TRUE} the next time the code is run and nothing has changed
in the meantime.}
}
\value{
Returns a list of length 2 (if \code{fit_save = FALSE}) or
3 (if \code{fit_save = TRUE}).
The elements of the list are: a list of lavaan model output objects;
a list of parameter estimates from the models (standardized if \code{std = TRUE});
and, if \code{fit_save = TRUE}, a matrix of fit measures for each model.
}
\description{
\code{cfa.from.keys} runs a confirmatory factor analysis (CFA) model for each
element of a keys list. The keys list must be a named list of scales, where
each element is an item from the corresponding scale. The function is
designed to streamline running CFA models for all scales in a sample and to
input model outputs into downstream functions.
}
\details{
The model relies on \code{\link[=sem.check]{sem.check()}} for the back-end of running the models.
This enables saving inputs and outputs from model runs
(with \code{save_out = TRUE}) and checking to see if anything has changed from
prior runs before running again (with \code{check = TRUE}).
The functionality was included for a number of very slow models or a lot of
faster models, such that time spent rerunning them would be onerous.
For further details on how this works, see the \code{\link[=sem.check]{sem.check()}} function
documentation.

The function does not provide any warnings for poor fit beyond those provided
by lavaan.
If any CFA models have poor fit, there is currently no capability to update
them beyond removing items by omitting them from the keys.
Any other changes to CFA models have to be made manually currently.
(Note that with \code{save_out = TRUE},
model code is saved in \verb{file.path(out_dir, name, paste0(name, _mod.rds)},
which could help with manually updating models.)
}
\examples{
# Create CFA keys
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
# Run models
cfa_fit <- cfa.from.keys(keys, BFIGritHope, check = FALSE, fit_save = TRUE)
# Examine some results
summary(cfa_fit$fit$grit_c)                # Standard lavaan summary
cfa_fit$fit_measures[, c("cfi", "rmsea")]  # Fit measures
}
\seealso{
\code{\link[=sem.check]{sem.check()}}, which \code{cfa.from.keys()} uses for all the back-end, and
\code{\link[lavaan:sem]{lavaan::sem()}}, which is used to estimate the models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.clean.R
\name{cache.clean}
\alias{cache.clean}
\title{Cleans selected files from the current cache directory}
\usage{
cache.clean(older_than = NULL, interactive = TRUE)
}
\arguments{
\item{older_than}{A positive number indicating number of days (fractions allowed).
Files with older modification times than this will be deleted
after confirmation if \code{interactive = TRUE}.}

\item{interactive}{Logical.
\code{TRUE} indicates that confirmation will be required before files are deleted.
\code{FALSE} indicates that files will be deleted without requiring confirmation.
It is recommended to use \code{TRUE} except for carefully checked automated
processes.
Defaults to \code{TRUE}.}
}
\value{
\code{NULL} (invisibly). Primarily called to clean up the cache directory.
}
\description{
Functions in the package include options to write files to a cache directory,
set up with the \code{\link[=cache.setup]{cache.setup()}} function.
Obsoletely files might be created if the \code{name} parameter in a function call
is changed or if a function call is no longer being used.
The \code{cache.clean()} function provides a way to find files that have not be
modified in the last \code{older_than} days and lists the files for the user to
agree to delete or not.
}
\details{
The function is interactive if run in an interactive session with
\code{interctive = TRUE}.
In addition to deleting old files, the function will also optionally delete
empty directories and, if the top level cache directory is also empty,
then it will also optionally delete the cache directory and unset it.
The latter will only be done in interactive sessions with
\code{interactive = TRUE} to avoid breaking non-interactive code.

One way to clean only unused files is to run all current code and then
run \code{cache.clean()} with \code{older_than} set to something greater than the
number of days it takes for the code to run and less than when the code was
previously run. For example, if your code takes 5 minutes to run and you
previously ran it 2 days ago, you could run all your code and, when it has
finished, use \code{cache.clean(older_than = 1)}.

\code{semFromKeys} has no way to know what directories you have specified as the
cache in the past and cannot clean up unknown former cache directories.
Therefore, it is recommended that you use either the default location or
the same location for all models within the same project.
Either of these will enable easy detection, so that unneeded cached files
can be found and deleted.
}
\examples{
\dontrun{
  # Setup a cache directory
  cache.setup()
  # Now code with check = TRUE or save_out = TRUE will work, e.g.,
  # Create CFA keys
  keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
  keys <- sapply(
    keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
  )
  # Run models
  cfa_fit <- cfa.from.keys(
    keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
  )
  # Check that they are not estimated again.
  cfa_fit <- cfa.from.keys(
    keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
  )

  # The cache.clean lines are commented out so you do not inadvertently
  # delete any models created with your own code in the same cache directory.

  # cache.clean(30)  # Delete files not modified in the last 30 days.
  # cache.clean(60/86400)  # Delete files not modified in the last minute.
  # Delete all files matching those created by the package in the current
  # cache directory and subdirectories without confirmation.
  # cache.clean(0, interactive = FALSE)
}
}
\seealso{
\code{\link[=cache.setup]{cache.setup()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa.from.keys.R
\name{efa.from.keys}
\alias{efa.from.keys}
\title{Runs an EFA model based on items in a keys list.}
\usage{
efa.from.keys(
  keys,
  data,
  orthogonal = FALSE,
  fit_save = TRUE,
  fit_measures = "all",
  std.lv = TRUE,
  miss = "ML",
  est = "default",
  name = "efa",
  check = FALSE,
  save_out = FALSE
)
}
\arguments{
\item{keys}{A named list of keys. Names must be factor names, elements must be
vectors of items that should be targeted to load on the factor.}

\item{data}{A dataframe or object coercible to a dataframe.
Data must include all observed variables in any of the keys.}

\item{orthogonal}{Logical.
Sets the \code{orthogonal} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
\code{TRUE} indicates that unspecified latent variable correlations should be
fixed at 0;
\code{FALSE} indicates that unspecified latent variable correlations should be
freely estimated.
Defaults to \code{FALSE}.}

\item{fit_save}{Logical.
\code{TRUE} indicates model fit measures should be included in the output;
\code{FALSE} indicates model fit measures should not be included in the output.
\code{FALSE} may be desirable when fit measures are of little interest and when
they may take a long time to estimate.
Fit can still be examined for individual models with,
\verb{lavaan::fitMeasures(fit$fit$[model name])}.}

\item{fit_measures}{A vector of fit measures to save or 'all' to select all fit measures,
as per the \code{fit.measures} parameter from lavaan's \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}
function.
Defaults to 'all'. Irrelevant if \code{fit_save = FALSE}.}

\item{std.lv}{Sets the \code{std.lv} param, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to \code{TRUE}.}

\item{miss}{A string.
Sets the \code{missing} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to 'ML'.}

\item{est}{A string.
Sets the \code{estimator} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
The default ('default') uses the lavaan default for the model being run.}

\item{name}{A string indicating a subdirectory where model outputs will be saved when
\code{save_out = TRUE} and checked against when \code{check = TRUE}.
Defaults to 'efa'.
Irrelevant if both \code{save_out = FALSE} and \code{check = FALSE}.
The name should be unique for each set of models, or outputs from calls with
the same name will be overwritten.}

\item{check}{Logical.
\code{TRUE} indicates that current inputs should be compared to previous inputs
if they exist and that the model should not be rerun if nothing has changed;
\code{FALSE} indicates that these checks should not be made and the model should
be run regardless of the existence of previous inputs.}

\item{save_out}{Logical.
\code{TRUE} indicates that model code, a hash of the data, important input
parameter values, and output will be saved;
\code{FALSE} indicates that nothing will be saved.
Selecting \code{save_out = TRUE} enables the function to not rerun models next
time if \code{check = TRUE} the next time the code is run and nothing has changed
in the meantime.}
}
\value{
Returns a list of lists.
The elements are a list of lavaan bifactor model output objects;
a list of parameter estimates from the models (standardized if \code{std = TRUE});
and, if \code{fit_save = TRUE}, a matrix of fit measures for each model.
}
\description{
\code{efa.from.keys} runs a exploratory factor analysis (EFA) in lavaan with a
rotation targeted based on a keys list.
}
\details{
The function was designed to streamline running exploratory structural
equation models (ESEM) using Burt's (1976) 2-stage procedure to prevent
interpretational confounding in the context of ESEM. However, it can also be
used to easily run a targeted EFA with only a keys list to avoid having to
manually specify the target and model.

The function was designed for use with established multidimensional scales,
such that a target is always reasonable.
The function does not currently support untargeted rotations.

The model relies on \code{\link[=sem.check]{sem.check()}} for the back-end of running the models.
This enables saving inputs and outputs from model runs
(with \code{save_out = TRUE}) and checking to see if anything has changed from
prior runs before running again (with \code{check = TRUE}).
The functionality was included for a number of very slow models or a lot of
faster models, such that time spent rerunning them would be onerous.
For further details on how this works, see the \code{\link[=sem.check]{sem.check()}} function
documentation.
}
\examples{
# Create EFA keys
# Using only 3 factors to save time
keys_e0 <- paste0("bfi_", c("e", "a", "c"))
# Using less than all items to save time
# (This results in a less than ideal solution but it shouldn't matter for an
# example)
keys_e <- sapply(
  keys_e0,
  function(x) {
    names(BFIGritHope)[grep(paste0(x, "\\\\d_[1-2]"), names(BFIGritHope))]
  },
  simplify = FALSE
)
# Run model
efa_fit <- efa.from.keys(keys_e, BFIGritHope, check = FALSE, fit_save = TRUE)
# Examine results
summary(efa_fit$fit$efa)                   # Standard lavaan summary
efa_fit$fit_measures[, c("cfi", "rmsea")]  # Fit measures
}
\references{
Burt, R. S. (1976).
Interpretational confounding of unobserved variables in Structural Equation
Models. Sociological Methods & Research, 5(1), 3-52.
https://doi.org/10.1177/004912417600500101.
}
\seealso{
\code{\link[=sem.check]{sem.check()}}, which \code{efa.from.keys()} uses for all the back-end, and
\code{\link[lavaan:sem]{lavaan::sem()}}, which is used to estimate the models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bifactor.from.keys.R
\name{bifactor.from.keys}
\alias{bifactor.from.keys}
\title{Runs bifactor models for multiple scales based on keys lists.}
\usage{
bifactor.from.keys(
  keys_g,
  keys_b,
  keys,
  data,
  fit_save = TRUE,
  fit_measures = "all",
  std.lv = TRUE,
  miss = "ML",
  est = "default",
  name = "bifactor",
  check = FALSE,
  save_out = FALSE
)
}
\arguments{
\item{keys_g}{A named list of items in general factors.
Names must be the names of the general factors.
Each list element must be a vector of items that load on the general factors.
Must be the same length as \code{keys_b}.}

\item{keys_b}{A named list of group factors in general factors.
Names must be the names of the general factors.
Each list element must be a vector of group factors that load on the general
factors.
Must be the same length as \code{keys_g}.}

\item{keys}{A named list of items in group factors.
Names must be the names of the group factors.
Each list element must be a vector of items that load on the group factors.
Need not be the same length as \code{keys_g} and \code{keys_b}.}

\item{data}{A dataframe or object coercible to a dataframe.
Data must include all observed variables in any of the keys.}

\item{fit_save}{Logical.
\code{TRUE} indicates model fit measures should be included in the output;
\code{FALSE} indicates model fit measures should not be included in the output.
\code{FALSE} may be desirable when fit measures are of little interest and when
they may take a long time to estimate.
Fit can still be examined for individual models with,
\verb{lavaan::fitMeasures(fit$fit$[model name])}.}

\item{fit_measures}{A vector of fit measures to save or 'all' to select all fit measures,
as per the \code{fit.measures} parameter from lavaan's \code{\link[lavaan:fitMeasures]{lavaan::fitMeasures()}}
function.
Defaults to 'all'. Irrelevant if \code{fit_save = FALSE}.}

\item{std.lv}{Logical.
Sets the \code{std.lv} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
\code{TRUE} indicates that factor variances should be fixed to 1.
\code{FALSE} indicates that loadings of the first items of factors should be fixed
to 1.
Defaults to \code{TRUE}.}

\item{miss}{A string.
Sets the \code{missing} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
Defaults to 'ML'.}

\item{est}{A string.
Sets the \code{estimator} parameter, as per lavaan (see \code{\link[lavaan:lavOptions]{lavaan::lavOptions()}}).
The default ('default') uses the lavaan default for the model being run.}

\item{name}{A string indicating a subdirectory where model outputs will be saved when
\code{save_out = TRUE} and checked against when \code{check = TRUE}.
Defaults to 'bifactor'.
Irrelevant if both \code{save_out = FALSE} and \code{check = FALSE}.
The name should be unique for each set of models, or outputs from calls with
the same name will be overwritten.}

\item{check}{Logical.
\code{TRUE} indicates that current inputs should be compared to previous inputs
if they exist and that the model should not be rerun if nothing has changed;
\code{FALSE} indicates that these checks should not be made and the model should
be run regardless of the existence of previous inputs.}

\item{save_out}{Logical.
\code{TRUE} indicates that model code, a hash of the data, important input
parameter values, and output will be saved;
\code{FALSE} indicates that nothing will be saved.
Selecting \code{save_out = TRUE} enables the function to not rerun models next
time if \code{check = TRUE} the next time the code is run and nothing has changed
in the meantime.}
}
\value{
Returns a list of length 2 (if \code{fit_save = FALSE}) or
3 (if \code{fit_save = TRUE}).
The elements of the list are: a list of lavaan model output objects;
a list of parameter estimates from the models (standardized if \code{std = TRUE});
and, if \code{fit_save = TRUE}, a matrix of fit measures for each model.
}
\description{
\code{bifactor.from.keys} runs a series of bifactor model from three keys lists---
one for items on general factor; one for items on group factors;
and one for group factors on general factors.
}
\details{
The function is designed to streamline running measurement models for all
scales in a sample and to input model outputs into downstream functions.
Keys list must be named appropriately;
that is, \code{keys_g} and \code{keys_b} names must be the general factor names,
and \code{keys} must be the group factor names.

The model relies on \code{\link[=sem.check]{sem.check()}} for the back-end of running the models.
This enables saving inputs and outputs from model runs
(with \code{save_out = TRUE}) and checking to see if anything has changed from
prior runs before running again (with \code{check = TRUE}).
The functionality was included for a number of very slow models or a lot of
faster models, such that time spent rerunning them would be onerous.
For further details on how this works, see the \code{\link[=sem.check]{sem.check()}} function
documentation.

Please be careful with bifactor models.
In simulation studies, they can fit better than the models that generated the
data in the presence of unspecified complexity
(which is usually the case; Murray & Johnson, 2013).
They can also produce negative residual variances.
lavaan will warn you of this and you should deal with it before proceeding.
Items can also load in atheoretical ways on the general or group factors.
This could be some items loading negatively instead of positively or
vice versa or one or more items having insignificant loadings on a factor
that they should theoretically load on.

These problems can often be solved by omitting a factor.
When items from one group factor dominate the general factor,
the best solution might be to remove the general factor and treat the group
factors as separate constructs.
Perhaps more frequently, these problems might be solved by omitting one
(or more) group factors (i.e., an S-1 model; Eid et al., 2017).
This can either be done a priori
(perhaps based on which one should theoretically be most 'central' to the
general factor)
or after examining the fully specified bifactor model and dropping the worst
performing factor (see the example).
When more than one group factor is poor, I am not aware of any specific
advice on which to remove, so use your judgment if previous research has not
got any advice for your case.
}
\examples{
# Create keys
keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
keys <- sapply(
  keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
keys_g0 <- c("grit", "hope")
keys_g <- sapply(
  keys_g0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
)
keys_b1 <- sapply(
  keys_g0, function(x) keys0[grep(x, keys0)], simplify = FALSE
)
# Including all factors produces a negative residual variance for hope.
\donttest{bif_fit0 <- bifactor.from.keys(
  keys_g, keys_b1, keys, BFIGritHope, check = FALSE, fit_save = TRUE
)
summary(bif_fit0$fit$hope)}
# Fix negative residual variance by removing the first group factor.
keys_b <- keys_b1
keys_b$hope <- keys_b1$hope[-1]
bif_fit <- bifactor.from.keys(
  keys_g, keys_b, keys, BFIGritHope, check = FALSE, fit_save = TRUE
)
# Examine some results
summary(bif_fit$fit$grit)                  # Standard lavaan summary
bif_fit$fit_measures[, c("cfi", "rmsea")]  # Fit measures
}
\references{
Eid, M., Geiser, C., Koch, T., & Heene, M. (2017).
Anomalous results in G-factor models: Explanations and alternatives.
Psychological Methods, 22(3), 541-562.
https://doi.org/10.1037/met0000083.

Murray, A. L. & Johnson, W. (2013).
The limitations of model fit in comparing the bi-factor versus higher-order
models of human cognitive ability structure. Intelligence, 41(5), 407-422.
https://doi.org/10.1016/j.intell.2013.06.004.
}
\seealso{
\code{\link[=sem.check]{sem.check()}}, which \code{bifactor.from.keys()} uses for all the back-end, and
\code{\link[lavaan:sem]{lavaan::sem()}}, which is used to estimate the models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.setup.R
\name{cache.setup}
\alias{cache.setup}
\title{Sets up a cache directory for storing model inputs and outputs}
\usage{
cache.setup(location = "user", interactive = TRUE)
}
\arguments{
\item{location}{The cache location to save model objects to to enable checking.
If \code{location = "user"} (default, recommended if unsure),
then the system's default cache location is used.
If any other string is used a folder will be created within either the
\code{\link[=getwd]{getwd()}} directory or,
if available, within the directory identified by \code{\link[here:here]{here::here()}}.
If using an RStudio project, these will create the directory in the
project directory.}

\item{interactive}{Logical.
\code{TRUE} indicates that confirmation will be required before a cache directory
is set \emph{if} the default location is not used.
\code{FALSE} indicates that a cache directory will be set without confirmation.
Irrelevant if \code{location = "user"}.
It is recommended to use \code{TRUE}.
Defaults to \code{TRUE}.}
}
\value{
The cache directory path (invisibly).
Primarily called to set up the cache configuration.
}
\description{
Using \code{save_out = TRUE} or \code{check = TRUE} from
\code{\link[=cfa.from.keys]{cfa.from.keys()}}, \code{\link[=bifactor.from.keys]{bifactor.from.keys()}}, \code{\link[=efa.from.keys]{efa.from.keys()}}, or
\code{\link[=esem.from.mods]{esem.from.mods()}} requires a cache directory to save model objects to or
to check against.
By default, \code{cache.setup} uses the system's default cache location,
otherwise a location within the current working directory or project can be
selected.
The function needs to be run once after the environment is cleared whenever
a cache directory is required.
}
\details{
Saving outputs to save time running identical models more than once requires
files to be saved on your computer.
To avoid writing to your computer without your permission,
you are required to run this function first to ensure that you know \emph{that}
files are being saved and \emph{where} files are being saved.
The function temporarily sets a hidden environment variable '.cache_env'
(with, if empty,
\code{assign(".cache_env", new.env(parent = emptyenv()), envir = parent.frame(1))}
and with
\code{assign("cache_dir", cache_dir, envir = get(".cache_env", envir = parent.frame(1)))},
if not), which will be removed whenever the environment is cleared.

By default, the function creates the cache directory in the standard place
for the operating system being used, appended by \verb{semFromKeys/[projectname]},
if a project is being used and the \code{rstudioapi} is available.
If a project is not being used or the \code{rstudioapi} package is not available,
then the relative path with be \code{semFromKeys}.

In the later case, if two function calls include the same \code{name} assignment,
then outputs from one will overwrite the other. Therefore, it is recommended
to either set the cache directory to something other than the default or use
the default within a project with the \code{rstudioapi} package installed.
Obviously this latter option will likely not work outside of RStudio,
so, in that case, it is recommended to use a custom location.

To ensure that R knows what the cache directory is,
a hidden environment variable containing the information---\code{.cache_env}---is
saved within the working environment.
Other functions from the package will look for and use \code{.cache_env} to
identify the cache directory.
As a result, whenever the working environment is cleared
(e.g., upon closing RStudio) or whenever the \code{.cache_env} is otherwise
removed,
the \code{cache.setup} function will need to be re-run before the cache
functionality will work,
regardless of the status of the any recently used cache directories.

The function relies on code that was unavailable in versions of R prior to
version 4.0,
so anyone using an earlier version of R will either have to update R or forgo
the caching functionality.

\code{semFromKeys} has no way to know what directories you have specified as the
cache in the past and cannot clean up unknown former cache directories.
Therefore, it is recommended that you use either the default location or
the same location for all models within the same project.
Either of these will enable easy detection, so that unneeded cached files
can be found and deleted.

Functions that directly or indirectly might require a cache directory are:
\code{\link[=cfa.from.keys]{cfa.from.keys()}}, \code{\link[=bifactor.from.keys]{bifactor.from.keys()}}, \code{\link[=efa.from.keys]{efa.from.keys()}},
\code{\link[=esem.from.mods]{esem.from.mods()}}, and \code{\link[=sem.check]{sem.check()}}.
}
\examples{
\dontrun{
  # Setup a cache directory
  cache.setup()
  # Now code with check = TRUE or save_out = TRUE will work, e.g.,
  # Create CFA keys
  keys0 <- c("grit_c", "grit_p", "hope_a", "hope_p")
  keys <- sapply(
    keys0, function(x) names(BFIGritHope)[grep(x, names(BFIGritHope))]
  )
  # Run models
  cfa_fit <- cfa.from.keys(
    keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
  )
  # Check that models are not run again.
  cfa_fit <- cfa.from.keys(
    keys, BFIGritHope, fit_save = TRUE, check = TRUE, save_out = TRUE
  )
}
}
\seealso{
\code{\link[=cfa.from.keys]{cfa.from.keys()}}, \code{\link[=bifactor.from.keys]{bifactor.from.keys()}}, \code{\link[=efa.from.keys]{efa.from.keys()}},
\code{\link[=esem.from.mods]{esem.from.mods()}}, and \code{\link[=sem.check]{sem.check()}} for dependent functions;
\code{\link[tools:R_user_dir]{tools::R_user_dir()}} for default cache directories;
\code{\link[here:here]{here::here()}} for project-based cache directory setting; and
\code{\link[=cache.clean]{cache.clean()}} for a function to clean cache.
}
